/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#pragma once
extern "C" {
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
}
#include <string>
#include <vector>
#include <iostream>

#include <bmruntime_interface.h>
#include <op_code.h>
#include <bmdnn_api.h>
#include <bmcv_api.h>
#include "tensor.hpp"

namespace sail {

class Frame {
 public:
  Frame() {
    frame_ = av_frame_alloc();
  }
  ~Frame() {
    av_frame_free(&frame_);
  }
  AVFrame* get() {
    return frame_;
  }
  bool is_null() {
    return (frame_ == NULL);
  }
  int get_height() {
    return frame_->height;
  }
  int get_width() {
    return frame_->width;
  }
 private:
  AVFrame* frame_;
};

class VideoDecoder {
 public:
  explicit VideoDecoder(const std::string& file_path);
  ~VideoDecoder();
  bool is_opened();
  std::vector<int> get_frame_shape();
  int read(Frame& frame);

 private:
  bool open_codec_context();
  int decode_packet(AVFrame* frame);
  void (*av_log_callback)(void*, int, const char*, va_list) = NULL;
  int open_input_file();
  std::string file_path_;
  AVFormatContext* fmt_ctx_;
  AVCodecContext* dec_ctx_;
  AVPixelFormat pix_fmt_;
  AVPacket* packet_;
  AVStream* stream_;
  int pcie_board_id_;
  int refcount_;
  int stream_index_;
  int height_;
  int width_;
  bool status_;
  int total_frames_;
  int seeking_type_;
  bool read_end_;
  bool inside_decode_;
  bool start_;
};

class Bmcv {
 public:
  explicit Bmcv(BmRuntime bmrt) :
      handle_(static_cast<bm_handle_t>(bmrt_get_bm_handle(bmrt.get_ptr()))),
      p_bmrt_(bmrt.get_ptr()) {}
  void* get_bmrt() {
    return p_bmrt_;
  }
  ~Bmcv();

  void get_bgr_from_frame(Frame& frame, Tensor<float>& tensor);

  // interpolation=INTER_NEAREST
  int resize(
      Tensor<float>&             input,
      Tensor<float>&             output);

  int crop(
      Tensor<float>&             input,
      Tensor<float>&             output,
      int                        top,
      int                        left);

  int yuv2bgr(
      Tensor<float>&             input_y,
      Tensor<float>&             input_uv,
      Tensor<float>&             output);

  int crop_resize_norm(
      Tensor<float>&             input,
      Tensor<float>&             output,
      int                        top,
      int                        left,
      int                        crop_height,
      int                        crop_width,
      float                      scale_b,
      float                      bias_b,
      float                      scale_g,
      float                      bias_g,
      float                      scale_r,
      float                      bias_r);

  int resize_norm(
      Tensor<float>&             input,
      Tensor<float>&             output,
      float                      scale_b,
      float                      bias_b,
      float                      scale_g,
      float                      bias_g,
      float                      scale_r,
      float                      bias_r);

  int transpose(
      Tensor<float>&             input,
      Tensor<float>&             output);

  int bgrsplit(
      Tensor<float>&             input,
      Tensor<float>&             output);

 private:
  bmcv_image get_bmcv_image(Tensor<float>& tensor);

  bm_handle_t handle_;
  void* p_bmrt_;
};

}  // namespace sail
