/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @file     graph.hpp
 *  @brief    Header file of graph
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-21
 */

#pragma once
#include <bmruntime_interface.h>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <memory>
#include "tensor.hpp"

#ifdef CROSS_COMPILE
#define NAMESPACE_USED  std
#else
#define NAMESPACE_USED  boost
#endif

/// Namespace containing all symbols from the sail library.
namespace sail {

/** @brief Program state at current batch when using asynchronous mode. */
enum AsyncStatus {
  /// First batch, only copy input tensors of this batch from system to device.
  BEGIN,
  /// Middle batches, Copy input tensors of this batch from system to device,
  /// and copy output tensors of last batch from device to system.
  PROCESS,
  /// Last batch, only copy output tensors of last batch from device to system.
  END
};

/**
 * @brief A class handles one model on one TPU.
 *
 * It's the real worker to do inference of a model on one TPU. And it copys
 * input and output tensors between system memory and device memory.
 */
template <typename Dtype>
class Graph {
 public:
  /**
   * @brief The only constructor of Graph.
   *
   * @param p_bmrt         Pointer of bmruntime
   * @param name           Name of the graph
   * @param mode           Specify the input/output tensors are in system
   *                       memory or device memory
   * @param dynamic        True for dynamic models and false for static ones
   * @param async          True for asynchronous mode and false for synchronous
   * @param input_shapes   Shapes of all input tensors
   * @param output_shapes  Shapes of all output tensors
   * @param input_tensors  Data pointers in system memory of all input tensors
   * @param output_tensors Data pointers in system memory of all output tensors
   */
  explicit Graph(
      void*                                    p_bmrt,
      const std::string&                       name,
      IOMode                                   mode,
      bool                                     dynamic,
      bool                                     async,
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, std::vector<int>>& output_shapes,
      std::map<std::string, Dtype*>&           input_tensors,
      std::map<std::string, Dtype*>&           output_tensors);
  virtual ~Graph();
  /**
   * @brief Reshape input tensors for dynamic models with synchronous mode.
   *
   * @param input_shapes  Shapes of all input tensors
   * @param input_tensors Data pointers in system memory of all input tensors
   * @force_reshape       If ture, force to reshape
   */
  void reshape(
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, Dtype*>&           input_tensors,
      bool                                     force_reshape);
  /**
   * @brief Reshape input tensors for dynamic models with asynchronous mode.
   *
   * @param input_shapes  Shapes of all input tensors
   * @param input_tensors Data pointers in system memory of all input tensors
   * @force_reshape       If ture, force to reshape
   * @param mode          TensorMode used at current batch
   */
  void reshape(
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, Dtype*>&           input_tensors,
      bool                                     force_reshape,
      TensorMode                               mode);
  /**
   * @brief Do inference with asynchronous mode.
   *
   * At first batch, mode is BEGIN. At middle batches, mode is process. At last
   * batch, mode is END. Must call the function (batch number + 1) times to get
   * all the output.
   *
   * @param mode       TensorMode used at current batch
   * @param status     AsyncStatus at current batch
   */
  void inference(TensorMode mode, AsyncStatus status);
  /**
   * @brief Do inference with synchronous mode.
   */
  void inference();
  /**
   * @brief Get the data pointer in the system memory of the tensor.
   *
   * @param tensor_name Specified tensor name
   * @return Data pointer in system memory of the tensor
   */
  Dtype* get_tensor_sys(const std::string& name) const;
  /**
   * @brief Get the device memory of thetensor with asynchronous mode.
   *
   * @param tensor_name Specified tensor name
   * @param mode        TensorMode used at current batch
   * @return Device memory of the tensor
   */
  bm_device_mem_t* get_tensor(const std::string& name, TensorMode mode);
  /**
   * @brief Get the device memory of thetensor with synchronous mode.
   *
   * @param tensor_name Specified tensor name
   * @return Device memory of the tensor
   */
  bm_device_mem_t* get_tensor(const std::string& name);
  /**
   * @brief Clear information of tensors and free device memory structs.
   */
  void free_tensors();
  /**
   *  Set pointer of output tensor shapes when using dynamic models.
   *
   *  @param Shapes of output tensors
   */
  void set_dynamic_output_shapes(
      std::map<std::string, std::vector<int>>* p_output_shapes);

 private:
  /**
   * @brief Forbidden copy constructor.
   */
  Graph(const Graph &other);
  /**
   * @brief Forbidden assignment function.
   */
  Graph& operator=(const Graph &other);
  /**
   * @brief Get indexes of the input tensors.
   */
  void get_input_tensor_idx();
  /**
   * @brief Get indexes of the output tensors.
   */
  void get_output_tensor_idx();
  /**
   * @brief Copy all input tensors from system memory to device memory when
   *        using synchronous mode.
   */
  void input_s2d();
  /**
   * @brief Copy all input tensors from system memory to device memory when
   *        using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   */
  void input_s2d(TensorMode mode);
  /**
   * @brief Copy all output tensors from device memory to system memory when
   *        using synchronous mode.
   */
  void output_d2s();
  /**
   * @brief Copy all output tensors from device memory to system memory with
   *        specified size when using synchronous mode.
   *
   * @param size Byte size of all output tensors to be copied
   */
  void output_d2s(const std::vector<int>& size);
  /**
   * @brief Copy all output tensors from device memory to system memory when
   *        using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   */
  void output_d2s(TensorMode mode);
  /**
   * @brief Copy all output tensors from device memory to system memory with
   *        specified size when using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   * @param size Byte size of all output tensors to be copied
   */
  void output_d2s(TensorMode mode, const std::vector<int>& size);
  /**
   * @brief Get pointer to the device memory structs of all input tensors when
   *        using synchronous mode.
   */
  bm_device_mem_t* get_input_mems();
  /**
   * @brief Get pointer to the device memory structs of all input tensors when
   *        using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   */
  bm_device_mem_t* get_input_mems(TensorMode mode);
  /**
   * @brief Get pointer to the device memory structs of all output tensors when
   *        using synchronous mode.
   */
  bm_device_mem_t* get_output_mems();
  /**
   * @brief Get pointer to the device memory structs of all output tensors when
   *        using synchronous mode.
   *
   * @param mode TensorMode used at current batch
   */
  bm_device_mem_t* get_output_mems(TensorMode mode);

  /// Graph name
  std::string name_;
  /// Indicator of where to store input and output tensors.
  /// SYSI: Input tensors are in system memory while output tensors are
  ///       in device memory.
  /// SYSO: Input tensors are in device memory while output tensors are
  ///       in system memory.
  /// SYSIO: Both input and output tensors are in system memory.
  /// DEVIO: Both input and output tensors are in device memory.
  IOMode io_mode_;
  /// Indicator of model property.
  /// True for dynamic, and the shapes of input/output tensor can change.
  /// False for static, and the shapes of input/output tensor cannot change.
  bool dynamic_;
  /// Indicator of processing manner.
  /// True for asynchronization, and PING/PONG buffers are used.
  /// False for synchronization, and Only PING buffer is used.
  bool async_;
  /// Pointer to bmruntime.
  void* p_bmrt_;
  /// Map of input tensor names and indexes
  std::map<std::string, size_t> input_map_;
  /// Map of output tensor names and indexes
  std::map<std::string, size_t> output_map_;
  /// Map of tensor name and tensor index
  std::map<std::string, size_t> tensor_map_;
  /// Pointers to all input and output tensors
  std::vector<std::shared_ptr<Tensor<Dtype>>> tensors_;
  /// Pointers to all input tensors
  std::vector<Tensor<Dtype>*> input_tensors_;
  /// Pointers to all output tensors
  std::vector<Tensor<Dtype>*> output_tensors_;
  /// Pointer to the device memory structs of all input tensors of PING.
  bm_device_mem_t* input_mem_ping_;
  /// Pointer to the device memory structs of all input tensors of PONG.
  /// It is used when using asynchronous mode.
  bm_device_mem_t* input_mem_pong_;
  /// Pointer to the device memory structs of all output tensors of PING.
  bm_device_mem_t* output_mem_ping_;
  /// Pointer to the device memory structs of all output tensors of PONG.
  /// It is used when using asynchronous mode.
  bm_device_mem_t* output_mem_pong_;
  /// Output shapes are calculated after bmrt_launch_nhw when using dynamic
  /// models.
  std::map<std::string, std::vector<int>>* dynamic_output_shapes_;
};

}  // namespace sail
