/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @file     inference.hpp
 *  @brief    Header file of Inference
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-21
 */

#pragma once
/* POSIX Linux */
#include <unistd.h>
#include <bmruntime_interface.h>
#include <bmlib_runtime.h>
#include <op_code.h>
#include <iostream>
#include <algorithm>
#include <numeric>
#include <cmath>
#include <memory>
#include <cstring>
#include <map>
#include <set>
#include <string>
#include <vector>
#include "graph.hpp"
#include "tensor.hpp"
#include "utils.hpp"
#ifdef PYTHON
#include <boost/python.hpp>
#include <boost/python/numpy.hpp>
#include <boost/python/list.hpp>
#include <boost/python/dict.hpp>
namespace bp = boost::python;
namespace bpn = boost::python::numpy;
#endif

/// Namespace containing all symbols from the sail library.
namespace sail {

#define ASSERT(_cond)                            \
  do {                                           \
    if (!(_cond)) {                              \
      printf("ASSERT %s: %s: %d: %s\n",          \
          __FILE__, __func__, __LINE__, #_cond); \
    }                                            \
  } while (0)

template <typename Dtype> class Inference;

/**
 * @brief A struct contains the parameters when using multi-cards with
 *        synchronous mode.
 *
 * Each TPU owns a thread to run the model. And the parameters needed by
 * by the thread are stored in this struct.
 */
template <typename Dtype>
struct MultiSync {
  pthread_t thread_id;
  std::shared_ptr<Graph<Dtype>> graph;
};

/**
 * @brief The main class of running deep learning inference on TPU.
 *
 * It's a high level encapsulation of BMRuntime. It automatically manage
 * memory of input and output tensors for both static and dynamic models.
 * Synchronous or asynchronous mode is optional according to the application.
 * It can load more than one model and runs on more than one TPU.
 */
template <typename Dtype>
class Inference {
 public:
  /**
   * @brief Constructor using default input shapes.
   *
   * @param context_dir Context directory path generated by BMCompiler
   * @param chip_name   Name of the hardware. Option: "BM1682_PCIE",
   *                    "BM1682_SOC", "BM1684_PCIE", "BM1684_SOC"
   * @param tpus        Ids of TPUS to be used, split by comma. Eg.: "0,1" for
   *                    TPU 0 and TPU 1, "all" for all available TPUs
   * @param mode        Specify the input/output tensors are in system memory
   *                    or device memory
   * @param dynamic     True for dynamic models and false for static ones
   * @param async       True for asynchronous mode and false for synchronous
   */
  Inference(
      const std::string& context_dir,
      const std::string& chip_name,
      const std::string& tpus,
      IOMode             mode,
      bool               dynamic,
      bool               async);
  /**
   * @brief Copy constructor.
   *
   * @param other An other Inference instance.
   */
  Inference(const Inference& other);
  /**
   * @brief Assignment function.
   *
   * @param other An other Inference instance.
   * @return Reference of a Inference instance.
   */
  Inference<Dtype>& operator=(const Inference& other);
  ~Inference();
  /**
   * @brief Get all graph names in the loaded context.
   *
   * @return All graph names
   */
  std::set<std::string> get_all_graph_names();
  /**
   * @brief Get all input tensor names of the specified graph.
   *
   * @param graph_name The specified graph name
   * @return All the input tensor names of the graph
   */
  std::set<std::string> get_input_tensor_names(const std::string& graph_name);
  /**
   * @brief Get all output tensor names of the specified graph.
   *
   * @param graph_name The specified graph name
   * @return All the output tensor names of the graph
   */
  std::set<std::string> get_output_tensor_names(const std::string& graph_name);
  /**
   * @brief Judge if batch size of all input tensors are equal.
   *
   * @param input_shapes All input tensor shapes
   * @batch_size         Batch size to return
   *
   * @return True if batch size of all input tensors are equal
   */
  bool is_input_batch_size_equal(
      std::map<std::string, std::vector<int>>& input_shapes,
      int&                                     batch_size);
  /**
   * @brief Set input tensor shapes when running dynamic models with
   *        synchronous mode.
   *
   * The input tensor shapes may change between batches when running dynamic
   * models. New input shapes should be set before inference. Note that this
   * function is for synchronous mode.
   *
   * @param graph_name   The specified graph name
   * @param input_shapes Specified shapes of all input tensors of the graph
   */
  void set_input_shape(
      const std::string&                       graph_name,
      std::map<std::string, std::vector<int>>& input_shapes,
      TensorMode                               mode);
  /**
   * @brief Set input tensor shapes when running dynamic models with
   *        synchronous mode.
   *
   * The input tensor shapes may change between batches when running dynamic
   * models. New input shapes should be set before inference. Note that this
   * function is for asynchronous mode.
   *
   * @param graph_name   The specified graph name
   * @param input_shapes Specified shapes of all input tensors of the graph
   * @param mode         TensorMode used at current batch
   */
  void set_input_shape(
      const std::string&                       graph_name,
      std::map<std::string, std::vector<int>>& input_shapes);
  /**
   * @brief Get bmruntime pointers created by constructor.
   *
   * @return bmruntime pointers
   */
  BmRuntime get_bmrt();

  /**
   * @brief Get max shapes of input tensors in a graph.
   *
   * For static models, the max shape is the static shape and it should not be
   * changed. While for dynamic models, the tensor shape should be smaller than
   * or equal to the max shape.
   *
   * @param graph_name The specified graph name
   * @param is_total   True for total specified tpus; False for a single one
   * @return The max shape of the tensor
   */
  std::map<std::string, std::vector<int>> get_max_input_shapes(
      const std::string& graph_name,
      bool               is_total);
  /**
   * @brief Get the shape of an input tensor in a graph.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return The shape of the tensor
   */
  std::vector<int> get_input_shape(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the shape of an output tensor in a graph.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return The shape of the tensor
   */
  std::vector<int> get_output_shape(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the data pointer in the system memory of the input tensor.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return The data pointer in the system memory of the input tensor
   */
  Dtype* get_input_tensor_sys(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the data pointer in the system memory of the output tensor.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return The data pointer in the system memory of the output tensor
   * @note It's static models or synchronous mode
   */
  Dtype* get_output_tensor_sys(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the device memory of the input tensor with synchronous mode.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return The device memory of the input tensor
   */
  bm_device_mem_t* get_input_tensor_dev_mem(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the input tensor which only owns the device memory with
   *        synchronous mode.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return A Tensor instance which only owns the device memory of the tensor.
   */
  Tensor<Dtype> get_input_tensor_dev(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the device memory of the input tensor with asynchronous mode.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @param mode        TensorMode used at current batch
   * @return The device memory of the input tensor
   */
  bm_device_mem_t* get_input_tensor_dev(
      const std::string& graph_name,
      const std::string& tensor_name,
      TensorMode         mode);
  /**
   * @brief Get the device memory of the output tensor with synchronous mode.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return The device memory of the output tensor
   */
  bm_device_mem_t* get_output_tensor_dev_mem(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the output tensor which only owns the device memory with
   *        synchronous mode.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @return A Tensor instance which only owns the device memory of the tensor.
   */
  Tensor<Dtype> get_output_tensor_dev(
      const std::string& graph_name,
      const std::string& tensor_name);
  /**
   * @brief Get the device memory of the output tensor with asynchronous mode.
   *
   * @param graph_name  The specified graph name
   * @param tensor_name The specified tensor name
   * @param mode        TensorMode used at current batch
   * @return The device memory of the output tensor
   */
  bm_device_mem_t* get_output_tensor_dev(
      const std::string& graph_name,
      const std::string& tensor_name,
      TensorMode         mode);
  /**
   * @brief Do inference with asynchronous mode.
   *
   * At first batch, mode is BEGIN. At middle batches, mode is process. At last
   * batch, mode is END. Must call the function (batch number + 1) times to get
   * all the output.
   *
   * @param graph_name The specified graph name
   * @param mode       TensorMode used at current batch
   * @param status     AsyncStatus at current batch
   */
  void process(
      const std::string& graph_name,
      TensorMode         mode,
      AsyncStatus        status);
  /**
   * @brief Do inference with asynchronous mode.
   *
   * At first batch, mode is BEGIN. At middle batches, mode is process. At last
   * batch, mode is END. Must call the function (batch number + 1) times to get
   * all the output.
   *
   * @param graph_name The specified graph name
   * @param mode       TensorMode used at current batch
   * @param status     AsyncStatus at current batch
   */
  void process(
      const std::string&                       graph_name,
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, Dtype*>&           input_tensors,
      TensorMode                               mode,
      AsyncStatus                              status);
  /**
   * @brief Do inference with synchronous mode.
   *
   * @param graph_name The specified graph name
   */
  void process(const std::string& graph_name);
  /**
   * @brief Do inference with synchronous mode with provided input tensors
   *
   * @param graph_name The specified graph name
   * @input_shapes     Shapes of all input tensors
   * @input_tensors    Data pointers of all input tensors in system meory
   */
  void process(
      const std::string&                       graph_name,
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, Dtype*>&           input_tensors);
  /**
   * @brief Reset data pointers of input tensors in system memory.
   *
   * @param graph_name    The specified graph name
   * @param input_shapes  Shapes of all input tensors
   * @param input_tensors Data pointers in system meory for all input tensors
   */
  void reset_input_tensors(
      const std::string&                       graph_name,
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, Dtype*>&           input_tensors);
  /**
   * @brief Reset data pointers of input tensors in system memory.
   *
   * @param graph_name    The specified graph name
   * @param input_shapes  Shapes of all input tensors
   * @param input_tensors Data pointers in system meory for all input tensors
   * @param mode          TensorMode used at current batch
   */
  void reset_input_tensors(
      const std::string&                       graph_name,
      std::map<std::string, std::vector<int>>& input_shapes,
      std::map<std::string, Dtype*>&           input_tensors,
      TensorMode                               mode);

#ifdef PYTHON
  bp::list get_all_graph_names_py();

  bp::list get_input_tensor_names_py(
      const std::string &graph_name);

  bp::list get_output_tensor_names_py(
      const std::string &graph_name);

  void set_input_shape_sync_py(
      const std::string& graph_name,
      bp::dict&          input_shapes);

  void set_input_shape_async_py(
      const std::string& graph_name,
      bp::dict&          input_shapes,
      TensorMode         mode);

  bp::dict get_max_input_shapes_py(
      const std::string& graph_name);

  bpn::ndarray get_input_shape_py(
      const std::string& graph_name,
      const std::string& tensor_name);

  bpn::ndarray get_output_shape_py(
      const std::string& graph_name,
      const std::string& tensor_name);

  void set_input_tensor_sys_py(
      const std::string& graph_name,
      const std::string& tensor_name,
      bpn::ndarray&      data);

  bpn::ndarray get_input_tensor_sys_py(
      const std::string& graph_name,
      const std::string& tensor_name);

  bpn::ndarray get_output_tensor_sys_py(
      const std::string& graph_name,
      const std::string& tensor_name);

  Tensor<Dtype> get_input_tensor_dev_py(
      const std::string& graph_name,
      const std::string& tensor_name);

  Tensor<Dtype> get_input_tensor_dev_py(
      const std::string& graph_name,
      const std::string& tensor_name,
      TensorMode         mode);

  Tensor<Dtype> get_output_tensor_dev_py(
      const std::string& graph_name,
      const std::string& tensor_name);

  Tensor<Dtype> get_output_tensor_dev_py(
      const std::string& graph_name,
      const std::string& tensor_name,
      TensorMode mode);

  bp::dict process_sync(
      const std::string& graph_name,
      bp::dict&          input_tensors);

  bp::dict process_async(
      const std::string& graph_name,
      bp::dict&          input_tensors,
      TensorMode         mode,
      AsyncStatus        status);
#endif

 private:
  /**
   * @brief Convert string parameters of tpus to an int vector.
   *
   * @param input_tpus String parameters of TPUs, split by comma
   * @return An int vector contains the TPU ids
   */
  std::vector<int> get_tpus(const std::string& input_tpus);
  /**
   * @brief Initialize bmruntimes.
   *
   * It's one-to-one relationships between TPU and bmruntime. A bmruntime hold
   * the information of running the models in the context.
   *
   * @param context_dir Context directory path generated by BMCompiler
   * @param chip_name   Name of the hardware. Option: "BM1682_PCIE",
   *                    "BM1682_SOC", "BM1684_PCIE", "BM1684_SOC"
   * @param tpus        Ids of TPUS to be used, split by comma. Eg.: "0,1" for
   *                    TPU 0 and TPU 1, "all" for all available TPUs
   * @return Program state
   *     @retval O      Success
   *     @retval others Failure
   */
  int init_bmrts(
      const std::string& context_dir,
      const std::string& chip_name,
      const std::string& tpus);
  /**
   * @brief Judge if the shape of input tensor is valid.
   *
   * @param graph_name   The specified graph name
   * @param input_shapes Specified shapes of all input tensors of the graph
   * @return Program state
   *     @retval O      Success
   *     @retval others Failure
   */
  int is_input_shape_valid(
      const std::string&                             graph_name,
      const std::map<std::string, std::vector<int>>& input_shapes);
  /**
   * @brief Reset shapes of input and output tensors for dynamic models.
   *
   * @param graph_name     The specified graph name
   * @param input_shapes   Shapes of input tensors
   * @param input_flags    Indicators of whether shapes of input tensors are
   *                       changed
   * @param input_shapes_v Shapes of input tensors on each TPU
   * @return Program state
   *     @retval O      Success
   *     @retval others Failure
   */
  int reshape(
      const std::string&                                    graph_name,
      std::map<std::string, std::vector<int>>&              input_shapes,
      std::map<std::string, bool>&                          input_flags,
      std::map<std::string, std::vector<std::vector<int>>>& input_shapes_v);
  /**
   * @brief Allocate system memroy for input and output tensors.
   *
   * @param graph_name   The specified graph name
   * @param input_flags  Indicators of whether shapes of input tensors are
   *                     changed
   */
  void alloc_sys_data(
      const std::string&                       graph_name,
      std::map<std::string, bool>&             input_flags);
  /**
   * @brief Get data pointers of tensors in system meory which bound to a tpu.
   *
   * @param tensors Data pointers in system meory for the whole tensors
   * @param shapes  Shapes of tensors that bound to a tpu
   * @param offsets Offsets of data pointers
   * @return Data pointers of tensors in system meory which bound to the
   *         specified tpu
   */
  std::map<std::string, Dtype*> split_sys_data(
      std::map<std::string, Dtype*>&           tensors,
      std::map<std::string, std::vector<int>>& shapes,
      std::map<std::string, size_t>&           offsets);
  /**
   * @brief Allocate device memroy for input and output tensors.
   *
   * @param graph_name     The specified graph name
   * @param input_shapes_v Shapes of input tensors on each TPU
   */
  void alloc_dev_data(
      const std::string&                                    graph_name,
      std::map<std::string, std::vector<std::vector<int>>>& input_shapes_v);
  /**
   * @brief Allocate device memroy for input tensors with asynchronous mode.
   *
   * @param graph_name     The specified graph name
   * @param input_shapes_v Shapes of input tensors on each TPU
   * @param mode           TensorMode used at current batch
   */
  void alloc_dev_data(
      const std::string&                                    graph_name,
      std::map<std::string, std::vector<std::vector<int>>>& input_shapes_v,
      TensorMode                                            mode);
  /**
   * @brief Initialize Graphs.
   *
   * It's one-to-one relationships between TPU and bmruntime. While it's
   * one-to-many relationships between bmruntime and Graph.
   *
   * @return Program state
   *     @retval O      Success
   *     @retval others Failure
   */
  int init_graphs();
  /**
   * @brief Free all tensors on both system and device memory.
   */
  void free_pointers();
  /**
   * @brief Do inference with synchronous mode using multi-threads
   *        on multi-TPUs.
   *
   * @param Poniter of the thread parameters
   */
  static void* multi_process_sync(void* parameter);
  /**
   * @brief Do inference with asynchronous mode using multi-threads
   *        on multi-TPUs.
   *
   * @param Poniter of the thread parameters
   */
  std::map<std::string, std::vector<int>> get_max_output_shapes(
      const std::string& graph_name,
      bool               is_total);
  /**
   * @brief Calculate output shapes and handle output tensors when using
   *        dynamic models.

   * @param graph_name The specified graph name
   */
  void prepare_dynamic_outputs(const std::string& graph_name);

  /// Indicator of where to store input and output tensors.
  /// SYSI: Input tensors are in system memory while output tensors are
  ///       in device memory.
  /// SYSO: Input tensors are in device memory while output tensors are
  ///       in system memory.
  /// SYSIO: Both input and output tensors are in system memory.
  /// DEVIO: Both input and output tensors are in device memory.
  IOMode io_mode_;
  /// Indicator of model property.
  /// True for dynamic, and the shapes of input/output tensor can change.
  /// False for static, and the shapes of input/output tensor cannot change.
  bool dynamic_;
  /// Indicator of processing manner.
  /// True for asynchronization, and PING/PONG buffers are used.
  /// False for synchronization, and Only PING buffer is used.
  bool async_;
  /// An int vector contains the TPU ids.
  std::vector<int> tpus_;
  /// A pointer vector contains all the bmruntimes. One TPU one bmruntime.
  std::vector<void*> p_bmrts_;
  /// Number of working TPUs.
  int worker_num_;
  /// Graph instance for each model on each TPU.
  std::map<std::string, std::vector<std::shared_ptr<Graph<Dtype>>>> graphs_;
  /// Shapes of all input tensors.
  std::map<std::string, std::map<std::string, std::vector<int>>> input_shapes_;
  /// Shapes of all output tensors.
  std::map<std::string, std::map<std::string, std::vector<int>>> output_shapes_;
  /// Data pointers in the system memory of all input tensors.
  std::map<std::string, std::map<std::string, Dtype*>> input_tensors_;
  /// Data pointers in the system memory of all output tensors.
  std::map<std::string, std::map<std::string, Dtype*>> output_tensors_;
  /// Data pointers in the system memory of all dynamic output tensors. Only
  /// uesd when using dynamic models.
  std::map<std::string, std::map<std::string, Dtype*>> dynamic_output_tensors_;
  /// Max shapes of input tensors on a single TPU.
  /// Only uesd when using dynamic models.
  std::map<std::string, std::map<std::string, std::vector<int>>>
      max_input_shapes_single_;
  /// Max shapes of output tensors on a single TPU.
  /// Only uesd when using dynamic models.
  std::map<std::string, std::map<std::string, std::vector<int>>>
      max_output_shapes_single_;
  /// Shapes of all output tensors on each TPU.
  /// Only uesd when using dynamic models.
  std::map<std::string, std::vector<std::map<std::string, std::vector<int>>>>
      dynamic_output_shapes_;
};

}  // namespace sail
