/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @file     statistics.hpp
 *  @brief    Header file of Statistics
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-21
 */

#pragma once
#include <ctime>
#include <map>
#include <string>
#include <vector>

/// Namespace containing all symbols from the sail library.
namespace sail {

/** @brief A struct records the start time and end time of a time slice. */
struct TimeSlice {
  /// Start time
  struct timespec start;
  /// End time
  struct timespec end;
};

/** @brief A timer records the processing time of items. */
class Statistics {
 public:
  Statistics() {}
  ~Statistics() {}
  /**
   * @brief Get item names.
   *
   * @return Item names
   */
  std::vector<std::string> get_items();
  /**
   * @brief Records start time of a item in a time slice.
   *
   * @param item Item name
   */
  void record_start(const std::string& item);
  /**
   * @brief Records end time of a item in a time slice.
   *
   * @param item Item name
   */
  void record_end(const std::string& item);
  /**
   * @brief Get the total processing time of a item.
   *
   * @param item Item name
   * @return Total processing time in ms
   */
  long int get_time(const std::string& item);
  /**
   * @brief Print the time of all items in ms.
   */
  void print_time();

 private:
  /**
   * @brief Convert time to ms.
   *
   * @param ts Time of timespec
   * @return Time in ms
   */
  long int timespec2ms(const struct timespec& ts);
  /// Records of the start and end time of all items.
  std::map<std::string, std::vector<TimeSlice> > records_;
};

}  // namespace sail
