/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @file     tensor.hpp
 *  @brief    Header file of Tensor
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-21
 */

#pragma once
#include <bmruntime_interface.h>
#include <vector>
#ifdef PYTHON
#include <boost/python.hpp>
#include <boost/python/numpy.hpp>
namespace bp = boost::python;
namespace bpn = boost::python::numpy;
#endif

/// Namespace containing all symbols from the sail library.
namespace sail {

class BmRuntime {
 public:
  BmRuntime(void* p_bmrt) : p_bmrt_(p_bmrt) {}
  bool is_null() {
    return (p_bmrt_ == NULL);
  }
  void* get_ptr() {
    return p_bmrt_;
  }
 private:
  void* p_bmrt_;
};

/** @brief Indicator of where to store input and output tensors. */
enum IOMode {
  /// Input tensors are in system memory while output tensors are
  /// in device memory.
  SYSI,
  /// Input tensors are in device memory while output tensors are
  /// in system memory.
  SYSO,
  /// Both input and output tensors are in system memory.
  SYSIO,
  /// Both input and output tensors are in device memory.
  DEVIO
};
/** @brief Indicator of which devicem memory to use in asynchronous mode. */
enum TensorMode {
  /// Using device memory PING of the tensor
  PING,
  /// Using device memory PONG of the tensor
  PONG
};

/**
 * @brief A class manages the system and device memory of a tensor.
 *
 * A tensor may only stores in sytem memory, or only stores in device memory,
 * or stores in both system memory and device memory. This class handles all
 * the conditions.
 */
template <typename Dtype>
class Tensor {
 public:
  /**
   * @brief Constructor only allocates system memory of the tensor.
   *
   * @param shape Shape of the tensor
   */
  explicit Tensor(const std::vector<int>& shape);
  /**
   * @brief Constructor only allocates device memory for PING of the tensor.
   *
   * @param p_bmrt Pointer to bmruntime
   * @param shape  Shape of the tensor
   */
  explicit Tensor(
      void*                   p_bmrt,
      const std::vector<int>& shape);
  /**
   * @brief Constructor only allocates device memory for PING of the tensor.
   *
   * @param p_bmrt Pointer to bmruntime
   * @param shape  Shape of the tensor
   * @param bytes  Bytes of data
   */
  explicit Tensor(
      void*                   p_bmrt,
      const std::vector<int>& shape,
      int                     bytes);
  /**
   * @brief Constructor wrapper existied device memory.
   *
   * @param p_bmrt Pointer to bmruntime
   * @param shape  Shape of the tensor
   * @param data   Pointer of device memory struct for PING.
   */
  explicit Tensor(
      void*                   p_bmrt,
      const std::vector<int>& shape,
      bm_device_mem_t*        data);

  /**
   * @brief Common constructor.\n
   * @detail
   *  case 0: only allocate system memory
   *          (NULL, shape, true, false, false, NULL, NULL, NULL) \n
   *  case 1: allocate system memory and use outer dev_data of PING
   *          (p_bmrt, shape, true, false, false, NULL, dev_mem_ping, NULL) \n
   *  case 2: allocate system memory and use outer dev_data of PING and PONG
   *          (p_bmrt, shape, true, false, false, NULL, dev_mem_ping, NULL) \n
   *  case 3: only allocate device memory of PING
   *          (p_bmrt, shape, false, true, true, NULL, NULL, NULL) \n
   *  case 4: only allocate device memory of PING and PONG
   *          (p_bmrt, shape, false, true, false, NULL, NULL, NULL) \n
   *  case 5: allocate system memroy and device memory of PING
   *          (p_bmrt, shape, true, true, false, NULL, NULL, NULL) \n
   *  case 6: allocate system memroy and device memory of PING and PONG
   *          (p_bmrt, shape, true, true, true, NULL, NULL, NULL) \n
   *  case 7: use outer sys_data and dev_struct of PING
   *          (p_bmrt, shape, false, false, true, false,
   *           data, dev_mem_ping, NULL) \n
   *  case 8: use outer sys_data and dev_struct of PING and PONG
   *          (p_bmrt, shape, false, false, true, true,
   *           data, dev_mem_ping, dev_mem_pong) \n
   *
   * @param p_bmrt       Pointer to bmruntime
   * @param shape        Shape of the tensor
   * @param own_sys_data Indicator of whether own the data pointer in system
   *                     memory. Allocate and free the pointer if it is true
   * @param own_dev_data Indicator of whether own the device memory struct.
   *                     Allocate and free the struct if it is true. The data
   *                     in device memory will be Allocated and freed if not
   *                     only allocate system memory for the tensor
   * @param async        True for asynchronous mode and false for synchronous
   * @param data         Data pointer in system memory of the tensor. Set NULL
   *                     if own_sys_data is false
   * @param dev_mem_ping Pointer of device memory struct for PING. Set NULL if
   *                     own_dev_data is false
   * @param dev_mem_pong Pointer of device memory struct for PING. Set NULL if
   *                     own_dev_data is false or async is false
   */
  explicit Tensor(
      void*                   p_bmrt,
      const std::vector<int>& shape,
      bool                    own_sys_data,
      bool                    own_dev_data,
      bool                    async,
      Dtype*                  data,
      bm_device_mem_t*        dev_mem_ping,
      bm_device_mem_t*        dev_mem_pong);

#ifdef PYTHON
  explicit Tensor(
      bpn::ndarray& shape);

  explicit Tensor(
      BmRuntime     bmrt,
      bpn::ndarray& shape);

  bpn::ndarray shape_py();

  bpn::ndarray sys_data_py();
#endif
  /**
   * @brief Copy constructor.
   *
   * @param tensor A Tensor instance
   */
  Tensor(const Tensor& tensor);
  /**
   * @brief Assignment function.
   *
   * @param tensor A Tensor instance
   * @return A Tensor instance
   */
  Tensor<Dtype>& operator=(const Tensor& tensor);
  virtual ~Tensor();
  /**
   * @brief Get shape of the tensor.
   *
   * @return Shape of the tensor
   */
  const std::vector<int>& shape() const;
  /**
   * @brief Reset shape of the tensor.
   *
   * @param shpae Shape of the tensor
   */
  void reshape(const std::vector<int>& shape);
  /**
   * @brief Reset shape of the tensor when using dynmaic models
   *        with asynchronous mode.
   *
   * @param shpae Shape of the tensor
   * @param mode  TensorMode used at current batch
   */
  void reshape(const std::vector<int>& shape, TensorMode mode);
  /**
   * @brief Judge if the tensor owns data pointer in system memory.
   *
   * @return True for owns data pointer in system memory.
   */
  bool own_sys_data();
  /**
   * @brief Judge if the tensor owns data pointer in device memory.
   *
   * @return True for owns data pointer in device memory.
   */
  bool own_dev_data();
  /**
   * @brief Get data pointer in system memory of the tensor.
   *
   * @return Data pointer in system memory of the tensor
   */
  Dtype* sys_data();
  /**
   * @brief Get pointer to device memory struct(PING) of the tensor.
   *
   * @return Pointer to device memory struct(PING) of the tensor
   */
  bm_device_mem_t* dev_data();
  /**
   * @brief Get pointer to device memory struct of the tensor when
   *        using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   * @return Pointer to device memory struct(PING) of the tensor
   */
  bm_device_mem_t* dev_data(TensorMode mode);
  /**
   * @brief Reset data pointer in system memory of the tensor.
   *
   * @param data Data pointer in system memory of the tensor
   */
  void reset_sys_data(Dtype* data);
  /**
   * @brief Reset pointer to device memory struct(PING) of the tensor.
   *
   * @param data Pointer to device memory struct(PING)
   */
  void reset_dev_data(bm_device_mem_t* data);
  /**
   * @brief Reset pointer to device memory struct of the tensor for both
   *        PING and PONG.
   *
   * @param data_ping Pointer to device memory struct(PING)
   * @param data_pong Pointer to device memory struct(PONG)
   */
  void reset_dev_data(
      bm_device_mem_t* data_ping,
      bm_device_mem_t* data_pong);
  /**
   * @brief Copy the tensor from system memory to device memory when
   *        using synchronous mode.
   */
  void sync_s2d();
  /**
   * @brief Copy the tensor from device memory to system memory when
   *        using synchronous mode.
   */
  void sync_d2s();
  /**
   * @brief Copy the tensor from device memory to system memory of specified
   *        size when using synchronous mode.
   *
   * @param size Byte size to be copied
   */
  void sync_d2s(int size);
  /**
   * @brief Copy the tensor from system memory to device memory when
   *        using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   */
  void sync_s2d(TensorMode mode);
  /**
   * @brief Copy the tensor from device memory to system memory when
   *        using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   */
  void sync_d2s(TensorMode mode);
  /**
   * @brief Copy the tensor from device memory to system memory of specified
   *        size when using asynchronous mode.
   *
   * @param mode TensorMode used at current batch
   * @param size Byte size to be copied
   */
  void sync_d2s(TensorMode mode, int size);
  /**
   * @brief Copy data from this tensor to another.
   *
   * @param dst Destination tensor to copy data to.
   */
  bool copy_to(Tensor& dst);
  /**
   * @brief Copy data from a source tensor to this one.
   *
   * @param src Source tensor to copy data from.
   */
  bool copy_from(Tensor& src);

 private:
  /**
   * @brief Allocate device memory for the tensor.
   *
   * @param mode      TensorMode used at current batch
   * @param data_size Bytes of data size.
   */
  void malloc_dev_data(TensorMode mode, int data_size);
  /**
   * @brief Free device memory of the tensor.
   *
   * @param mode TensorMode used at current batch
   */
  void free_dev_data(TensorMode mode);
  /**
   * @brief Free device memory PING and PONG of the tensor.
   */
  void free_dev_data();
  /**
   * @brief Free device memory PING, PONG and the respective structs.
   */
  void free_dev_data_and_struct();
  /**
   * @brief Free system memory of the tensor.
   */
  void free_sys_data();

  /// Pointer to bmruntime.
  void* p_bmrt_;
  /// Shape of the tensor.
  std::vector<int> shape_;
  /// Indicator of whether own the data pointer in system memory.
  bool own_sys_data_;
  /// Indicator of whether own the device memory struct.
  bool own_dev_data_;
  /// Indicator of processing manner.
  /// True for asynchronization, and PING/PONG buffers are used.
  /// False for synchronization, and Only PING buffer is used.
  bool async_;
  /// Data pointer in system memory of the tensor.
  Dtype* sys_data_;
  ///  Pointer to device memory struct(PING).
  bm_device_mem_t* dev_data_ping_;
  ///  Pointer to device memory struct(PONG), only used when async_ is true.
  bm_device_mem_t* dev_data_pong_;
};

}  // namespace sail
