/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @file     utils.hpp
 *  @brief    Header file of some common functinosn and class Ciper
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-21
 */

#pragma once
#include <vector>
#include <set>
#include <map>
#include <string>
#ifdef PYTHON
#include <boost/python.hpp>
#include <boost/python/numpy.hpp>
namespace bp = boost::python;
namespace bpn = boost::python::numpy;
#endif

/// Namespace containing all symbols from the sail library.
namespace sail {

/**
 * @brief Judge if the shape of a tensor is valid.
 *
 * @param shape Shape of a tensor
 * @return Whether the shape of a tensor is valid
 *     @retval True for valid
 *     @retval False for invalid
 */
bool shape_is_valid(const std::vector<int>& shape);
/**
 * @brief Judge if shapes of two tensors are equal.
 *
 * @param shape0 Shape of a tensor
 * @param shape1 Shape of another tensor
 * @return Whether shapes of the two tensors are equal
 *     @retval True for equal
 *     @retval False for not equal
 */
bool shapes_are_equal(
    const std::vector<int>& shape0,
    const std::vector<int>& shape1);
/**
 * @brief Get the data size of a tensor
 *
 * @param shape Shape of a tensor
 * @return Data size of the tensor
 */
int get_size_of_shape(const std::vector<int>& shape);
/**
 * @brief Get all file names in a directory
 *
 * @param dir Path of a directory
 * @return All file names in the directory
 */
std::vector<std::string> get_files_in_dir(const std::string& dir);

#ifdef PYTHON
std::vector<int> extract_shape(bpn::ndarray& shape);

bpn::ndarray convert_shape(const std::vector<int>& shape);

std::map<std::string, std::vector<int>> extract_shapes(bp::dict& shape);

template <typename Dtype>
bp::list vec_to_list(const std::vector<Dtype>& data) {
  bp::list ret;
  for (typename std::vector<Dtype>::const_iterator it = data.begin();
      it != data.end(); ++it) {
    ret.append(*it);
  }
  return ret;
}

template <typename Dtype>
bp::list set_to_list(const std::set<Dtype>& data) {
  bp::list ret;
  for (typename std::set<Dtype>::const_iterator it = data.begin();
      it != data.end(); ++it) {
    ret.append(*it);
  }
  return ret;
}

template <typename Dtype>
bpn::ndarray convert_to_ndarray(Dtype* data, const std::vector<int>& shape) {
  bp::list array_shape;
  bp::list array_stride;
  int max_stride = sizeof(Dtype);
  for (size_t i = 0; i < shape.size(); ++i) {
    max_stride *= shape[i];
  }
  for (size_t i = 0; i < shape.size(); ++i) {
    array_shape.append(shape[i]);
    max_stride /= shape[i];
    array_stride.append(max_stride);
  }
  bpn::ndarray ret = bpn::from_data(data,
                                    bpn::dtype::get_builtin<Dtype>(),
                                    array_shape,
                                    array_stride,
                                    bp::object());
  return ret;
}
#endif

}  // namespace sail
