/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @brief    Sail Postprocess Header File
 *  @file     postprocess.hpp
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-25
 */

#pragma once
#include <pthread.h>
#include <assert.h>
#include <vector>
#include <string>
#include "process/preprocess.hpp"
#include "core/tensor.hpp"
#include "core/inference.hpp"

/** @brief this namespace is designed for sail
 *
 * Sophon Artificial Intelligent Library namespace
 */

namespace sail {

// basic struct for detection
struct Box {
  float x_min;
  float y_min;
  float x_max;
  float y_max;
};

struct DetectRet {
  float x_min_ratio;
  float y_min_ratio;
  float x_max_ratio;
  float y_max_ratio;
  int class_id;
  float score;

  bool operator==(const DetectRet &t) const {
    return (t.x_min_ratio == this->x_min_ratio &&
        t.y_min_ratio == this->y_min_ratio &&
        t.x_max_ratio == this->x_max_ratio &&
        t.y_max_ratio == this->y_max_ratio &&
        t.class_id == this->class_id &&
        t.score == this->score);
  }

  bool operator!=(const DetectRet &t) const {
    return (t.x_min_ratio != this->x_min_ratio ||
        t.y_min_ratio != this->y_min_ratio ||
        t.x_max_ratio != this->x_max_ratio ||
        t.y_max_ratio != this->y_max_ratio ||
        t.class_id != this->class_id ||
        t.score != this->score);
  }
};

/****************
for ssd tf multithread
*****************/
struct TraversalArgs {
  float threshold;
  int num_input_prior;
  int num_class;
  float *input_conf;
  float *input_loc;
  int num_output_prior;
  int *output_index;
  float *output_conf;
  float *output_box;
  std::vector<std::vector<float>> *anchor_pointer;
  int anchor_offset;
};

void *traversal_confs(void *arg);

struct TraversalArgsV2 {
  float threshold;
  int num_thread;
  int id_this_thread;
  int num_input_prior;
  int num_class;
  float *input_conf;
  float *input_loc;
  int num_output_prior;
  int *output_index;
  float *output_conf;
  float *output_box;
  std::vector<std::vector<float>> *anchor_pointer;
};

void *traversal_confs_v2(void *arg);

/**
 * @brief A wrapper that performs a type of postprocess.
 *
 * An algorithm wrapper to complete the classification network
 * postprocessing and other algorithms with the same postprocessing
 */

class ImageNetPostprocess {
 public:
  /**
   * @brief Construct an imagenet postprocess object
   *
   * @param batch_size The batch size of network inference
   * @param class_num  The class_num of classification nerwork
   * @param top_k      The number of the classification result
   */
  ImageNetPostprocess(size_t batch_size, size_t class_num, size_t top_k);
  virtual ~ImageNetPostprocess() {}
  /**
   * @brief The execution function of postprocessing
   *
   * @param input The output data of inference
   * @return The classification result of each input
   *     @retval Ex:[[20,3,4,5,7],[6,30,1,2,3]] the index of classes
   */
  std::vector<std::vector<size_t>> process(float* input);

 private:
  size_t batch_size_;
  size_t class_num_;
  size_t top_k_;
};

/**
 * @brief A wrapper that performs the caffe-ssd postprocess.
 *
 * An algorithm wrapper to complete the postprocessing of caffe-ssd
 */

class SsdPostprocess {
 public:
  /**
  * @brief Construct a caffe-ssd postprocess object
  *
  * @param class_num  The class number of ssd object detection network
  * @param prior_num  The number of ssd prior bbox
  * @param threshold_conf  The threshold of object confidence
  * @param threshold_nms  The threshold of bbox nms
  */
  SsdPostprocess(
      size_t class_num,
      size_t prior_num,
      float  threshold_conf,
      float  threshold_nms);
  /**
   * @brief The execution function of ssd postprocessing
   *
   * @param conf_data  The confidence data of inference
   * @param loc_data  The regression data of inference
   * @param prior_data  The prior box data of inference
   * @return The detected bbox
   *     @retval Ex:[x_min,y_min,x_max,y_max,class_index,score]
   *             the info of bbox
   */
  std::vector<DetectRet> process(
      float* conf_data,
      float* loc_data,
      float* prior_data);
  virtual ~SsdPostprocess() {}

 private:
  float lap(float x1_min, float x1_max, float x2_min, float x2_max);
  float jaccard(std::vector<float> x1, std::vector<float> x2);
  size_t class_num_;
  size_t prior_num_;
  float threshold_conf_;
  float threshold_nms_;
};

/**
 * @brief A wrapper that performs the tf-ssd postprocess.
 *
 * An algorithm wrapper to complete the postprocessing of tf-ssd
 */

class SsdTfPostprocess {
 public:
  /**
  * @brief Construct a tf-ssd postprocess object
  *
  * @param class_num  The class number of ssd object detection network
  * @param prior_num  The number of ssd prior bbox
  * @param threshold_conf  The threshold of object confidence
  * @param threshold_nms  The threshold of bbox nms
  * @param max_box_num  The limited number of bbox
  */
  SsdTfPostprocess(
      size_t class_num,
      size_t prior_num,
      float  threshold_conf,
      float  threshold_nms,
      size_t max_box_num);
  /**
   * @brief The execution function of ssd postprocessing
   *
   * @param conf_data  The confidence data of inference
   * @param loc_data  The regression data of inference
   * @return The detected bbox
   *     @retval Ex:[x_min,y_min,x_max,y_max,class_index,score]
   *             the info of bbox
   */
  std::vector<DetectRet> process(float* conf_data, float* loc_data);
  /**
   * @brief The multithreading execution function v1.0
   *    of ssd postprocessing
   *
   * @param conf_data  The confidence data of inference
   * @param loc_data  The regression data of inference
   * @param numthread  The number of thread
   * @return The detected bbox
   *     @retval Ex:[x_min,y_min,x_max,y_max,class_index,score]
   *             the info of bbox
   */
  std::vector<DetectRet> process_mt(
      float* conf_data,
      float* loc_data,
      int    num_thread = 4);
  /**
   * @brief The multithreading execution function v2.0
   *    of ssd postprocessing
   *
   * @param conf_data  The confidence data of inference
   * @param loc_data  The regression data of inference
   * @param numthread  The number of thread
   * @return The detected bbox
   *     @retval Ex:[x_min,y_min,x_max,y_max,class_index,score]
   *             the info of bbox
   */
  std::vector<DetectRet> process_mt_v2(
      float* conf_data,
      float* loc_data,
      int    num_thread = 4);
  void generate_multiscale_priors(
      std::vector<float>& levels,
      float               img_side,
      std::vector<float>& scales,
      std::vector<float>& aspect_ratios,
      float               offset_ratio,
      float               anchor_scale);
  virtual ~SsdTfPostprocess() {}

 private:
  void argsort_down(float* data, int* res, int start, int end);
  float area(float* box);
  // y1 x1 y2 x2
  float box_overlap(float* box1, float* box2);
  size_t class_num_;
  size_t prior_num_;
  float threshold_conf_;
  float threshold_nms_;
  size_t max_box_num_;
  std::vector<std::vector<float>> priors_;
};

/*******************************************
 * for new retail solution ssd post-process
 *******************************************/
struct Dets {
  size_t class_id;
  double score;
  int xmin;
  int ymin;
  int xmax;
  int ymax;

  Dets() : class_id(0), score(0), xmin(0), ymin(0), xmax(0), ymax(0) {}
  Dets(size_t class_id, double score, int xmin, int ymin, int xmax, int ymax) :
      class_id(class_id), score(score), xmin(xmin),
      ymin(ymin), xmax(xmax), ymax(ymax) {}
};

class NRSSDPostprocess {
 public:
  NRSSDPostprocess(size_t class_num,
                   size_t prior_num,
                   float threshold_conf,
                   float threshold_nms);

  /**
   * process - post process for new retail SSD output
   *
   * @param concated_output - the output of SSD network,
   * concat conf_data, loc_data, prior_data and variances.
   * @return - vector of Dets struct, containing the final
   * results.
   */
  std::vector<Dets> process(float *concated_output);

  /**
   * jaccard - compute the IOU of two bounding box
   *
   * @param a - Dets struct
   * @param b - Dets struct
   * @return - IOU in [0, 1] if a,b are different class,
   * return 0.
   */
  double jaccard(const Dets &a, const Dets &b);

 private:
  size_t nb_classes;
  size_t nb_priors;
  float threshold_conf;
  float threshold_nms;
  size_t cols;
  size_t length_of_dets;
};

/*******************************************
 * for new retail solution ssd post-process
 * version 2. inputs are loc_data, conf_data, prior_data
 *******************************************/
class NRSSDPostprocess_v2 {
 public:
  NRSSDPostprocess_v2(size_t class_num,
                      size_t prior_num,
                      float threshold_conf,
                      float threshold_nms);

  // TODO(kui.ye): add comments
  std::vector<Dets> process(float *conf_data,
                            float *loc_data,
                            float *anchors);

  // compute IOU for each class
  double jaccard(const Dets &a, const Dets &b);

 private:
  size_t class_num_;
  size_t prior_num_;
  float threshold_conf_;
  float threshold_nms_;
  /* size_t cols; */
  /* size_t length_of_dets; */
};

/**
 * @brief A wrapper that performs the caffe-fasterrcnn postprocess.
 *
 * An algorithm wrapper to complete the postprocessing of
 *    caffe-fasterrcnn
 */

class FasterRcnnPostprocess {
 public:
  /**
  * @brief Construct a caffe-fasterrcnn postprocess object
  *
  * @param class_num  The class number of fasterrcnn
  *     object detection network
  * @param threshold_conf  The threshold of object confidence
  * @param threshold_nms  The threshold of bbox nms
  */
  FasterRcnnPostprocess(size_t class_num,
                        float threshold_conf,
                        float threshold_nms);
  /**
   * @brief The execution function of caffe-fasterrcnn postprocessing
   *
   * @param roi_num  The number of detection roi
   * @param rois_data  The roi data of inference
   * @param bbox_delt_data  The regression data of inference
   * @param score  The confidence data of inference
   * @return The detected bbox
   *     @retval Ex:[x_min,y_min,x_max,y_max,class_index,score]
   *             the info of bbox
   */
  std::vector<DetectRet> process(int roi_num,
                                 float *rois_data,
                                 float *bbox_delt_data,
                                 float *score);
  virtual ~FasterRcnnPostprocess() {}

 private:
  size_t class_num_;
  size_t threshold_conf_;
  float threshold_nms_;
};

/*******************
for detection yolov3
********************/
struct yolov3_box {
  float x;
  float y;
  float w;
  float h;
};

struct yolov3_DetectRect {
  int left;
  int right;
  int top;
  int bot;
  float score;
  int category;
};

struct detection {
  yolov3_box bbox;
  int classes;
  float *prob;
  float *mask;
  float objectness;
  int sort_class;
};

struct layer {
  int batch;
  int total;
  int n, c, h, w;
  int out_n, out_c, out_h, out_w;
  int classes;
  int inputs, outputs;
  int *mask;
  float *biases;
  float *output;
  float *output_gpu;
};

/**
 * @brief A wrapper that performs the yolov3 postprocess.
 *
 * An algorithm wrapper to complete the postprocessing of
 *    yolov3
 */

class Yolov3Postprocess {
 public:
  /**
  * @brief Construct a yolov3 postprocess object
  *
  * @param class_num  The class number of yolov3
  *     object detection network
  * @param threshold_prob  The threshold of object confidence
  * @param threshold_nms  The threshold of bbox nms
  * @param input_shape  The input shape of yolov3
  */
  Yolov3Postprocess(
      size_t            class_num,
      float             threshold_prob,
      float             threshold_nms,
      std::vector<int>& input_shape);
  /**
   * @brief The execution function of yolov3 postprocessing
   *
   * @param inference  The object of yolov3 inference
   * @param graph_name  The graph name of yolov3
   * @param mode  The IOmode of yolov3
   * @param status  The asyncmode of yolov3
   * @param img_w  The width of original image
   * @param img_h  The height of original image
   * @return The detected bbox
   *     @retval Ex:[left,right,top,bottom,score,category]
   *             the info of bbox
   */
  std::vector<yolov3_DetectRect> process(
      BmRuntime         bmrt,
      Tensor<float>&    output_s,
      Tensor<float>&    output_m,
      Tensor<float>&    output_l,
      int               img_w,
      int               img_h);
  virtual ~Yolov3Postprocess();

 private:
  // bbox nms sort compare function
  // int nms_comparator(const void* pa, const void* pb);
  // bbox iou calculation
  float box_iou(yolov3_box a, yolov3_box b);
  // bbox nms
  void do_nms_sort(detection *det, int total, int classes, float thresh);
  // init the yolo layer
  layer make_yolo_layer(int batch, int w, int h, int n, int total, int classes);
  // free yolo layer
  void free_yolo_layer(layer l);
  // get the dst index
  int entry_index(layer l, int batch, int location, int entry);
  // copy inference output to yolo layer input
  void forward_yolo_layer(const float *input, layer l);
  // filter bbox with obj_prob
  int yolo_num_detections(layer l, float thresh);
  // calculation the bbox num
  int num_detections(std::vector<layer> layers_params, float thresh);
  // alloc for bbox
  detection *make_network_boxes(std::vector<layer> layers_params,
                                float thresh,
                                int *num);
  // link bbox to img size
  void correct_yolo_boxes(detection *det,
                          int n,
                          int w,
                          int h,
                          int netw,
                          int neth,
                          int relative);
  // link bbox to net input size
  yolov3_box get_yolo_box(float *x,
                          float *biases,
                          int n,
                          int index,
                          int i,
                          int j,
                          int lw,
                          int lh,
                          int w,
                          int h,
                          int stride);
  // get bbox
  int get_yolo_detections(layer l,
                          int w,
                          int h,
                          int netw,
                          int neth,
                          float thresh,
                          int *map,
                          int relative,
                          detection *dets);
  // fill the bbox
  void fill_network_boxes(std::vector<layer> layers_params,
                          int w,
                          int h,
                          float thresh,
                          float hier,
                          int *map,
                          int relative,
                          detection *dets);
  // return the bbox result before nms
  detection *get_network_boxes(std::vector<layer> layers_params,
                               int img_w,
                               int img_h,
                               float thresh,
                               float hier,
                               int *map,
                               int relative,
                               int *num);
  // get detection result
  detection *get_detections(std::vector<float *> blobs,
                            int img_w,
                            int img_h,
                            int *nboxes);
  // free the bbox prob
  void free_detections(detection *dets, int nboxes);
  // post process set the dst index
  int set_index(int w,
                int h,
                int num,
                int classes,
                int batch,
                int location,
                int entry);
  // reverse sigmoid
  void bmrt_return(std::vector<float *> data_i);
  // reverse sigmoid
  void bmrt_return_internal(float *data_i,
                            int w,
                            int h,
                            int num,
                            int total,
                            int classes);
  // calculation the max index label
  int max_index(float *a, int n);
  // return final bbox
  std::vector<yolov3_DetectRect> detection_yolov3_process(detection *dets,
                                                          int nboxes,
                                                          int rows,
                                                          int cols);
  size_t classes_num_;
  // yolov3 anchor (in coco && voc)
  float biases_[18] = {10, 13, 16, 30, 33, 23, 30, 61, 62,
                       45, 59, 119, 116, 90, 156, 198, 373, 326};
  float threshold_prob_;
  float threshold_nms_;
  int net_w_, net_h_;
  int out_s_, out_m_, out_l_;
};

/**
 * @brief A wrapper that performs the caffe-mtcnn postprocess.
 *
 * An algorithm wrapper to complete the postprocessing of
 *    caffe-mtcnn
 */

class MtcnnPostprocess {
 public:
  /**
  * @brief Construct a mtcnn postprocess object
  *
  * @param threshold The array of mtcnn threshold
  */
  explicit MtcnnPostprocess(double* threshold);
  /**
   * @brief The execution function of Pnet postprocessing
   *
   * @param conf  The confidence data of inference
   * @param coord  The regression data of inference
   * @param scale  The scale factor of image
   * @param net_output_width  The width of Pnet output
   * @param net_output_height  The height of Pnet output
   * @param input_width  The width of Pnet input
   * @param input_height  The height of Pnet input
   * @return The detected face
   *     @retval The info of face
   */
  std::vector<FaceInfo> Pnet_process(float* conf, float* coord,
      float scale, int net_output_width, int net_output_height,
      int input_width, int input_height);
  /**
   * @brief The execution function of Rnet postprocessing
   *
   * @param conf  The confidence data of inference
   * @param coord  The regression data of inference
   * @param frame  The original image
   * @param regressed_rects  The regressed face bbox
   * @return The detected face
   *     @retval The info of face
   */
  std::vector<FaceInfo> Rnet_process(float* conf, float* coord,
      cv::Mat& frame, std::vector<FaceInfo>& regressed_rects);
  /**
   * @brief The execution function of Onet postprocessing
   *
   * @param conf  The confidence data of inference
   * @param coord  The regression data of inference
   * @param landmark  The landmark data of inference
   * @param frame  The original image
   * @param regressed_rects  The regressed face bbox
   * @return The detected face
   *     @retval The info of face
   */
  std::vector<FaceInfo> Onet_process(float* conf, float* coord,
      float* landmark, cv::Mat& frame,
      std::vector<FaceInfo>& regressed_rects);
  /**
   * @brief The mtcnn nms function of face
   *
   * @param BBoxes  The input face bboxes
   * @param thresh  The threshold of nms
   * @param methodType  The calculation method of nms
   * @return The face after nms
   *     @retval The info of face
   */
  std::vector<FaceInfo> MultiNMS(std::vector<FaceInfo>& BBoxes,
      float thresh, char methodType);
  /**
   * @brief The mtcnn bbox padding function
   *
   * @param regressed_rects  The input face bboxes
   * @param regressed_pading  The padding face
   */
  void Padding(cv::Mat& frame,
    std::vector<FaceInfo>& regressed_rects,
    std::vector<FaceInfo>& regressed_pading);
  /**
   * @brief The mtcnn bbox regression function
   *
   * @param faceInfo  The input face info
   * @param stage  The number of mtcnn stage
   * @return The face after regression
   *     @retval The info of face
   */
  std::vector<FaceInfo> BoxRegress(std::vector<FaceInfo>& faceInfo,
     int stage);
  /**
   * @brief The mtcnn bbox2square function
   *
   * @param bboxes  The input face info
   */
  void Bbox2Square(std::vector<FaceInfo>& bboxes);
  virtual ~MtcnnPostprocess();

 private:
  // gen face bbox
  void GenerateBoundingBox(float* conf, float* coord,
    std::vector<FaceInfo>& condidate_rects,
    int width, int height,
    float scale, float thresh,
    int image_width, int image_height, int imgid);
  std::vector<double> threshold_;
};

}  // namespace sail
