/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

/** @brief    Sail Preprocess Header File
 *  @file     preprocess.hpp
 *  @author   bitmain
 *  @version  2.0.0
 *  @date     2019-02-25
 */

#pragma once
#include <vector>
#include <string>
#include <opencv2/opencv.hpp>
#include <opencv2/core/core.hpp>
#include "core/bm_cv.hpp"
#include "core/tensor.hpp"

namespace sail {

/**
 * @brief A base class for preprocess.
 *
 * An algorithm wrapper to complete basic preprocessing function
 */

class Preprocess {
 public:
  Preprocess() : bm_cv_(NULL) {}
  explicit Preprocess(Bmcv* processor) : bm_cv_(processor) {}
  virtual void process_tpu(Tensor<float>& input, Frame& frame) {}
  virtual void process_cpu(float* input, cv::Mat& frame) {}

 protected:
  Bmcv* bm_cv_;
};

/**************************************
imagenet ResNet/VGG/GoogleNet/MobileNet
for mobilenet:
  height = width = 224;
  interpolation = cv::INTER_LINEAR);
  mean_0 = 103.94;
  mean_1 = 116.78;
  mean_2 = 123.68;
  scale_factor = 0.017;
  convert_color = false;
for vgg_ssd:
  height = width = 300;
  interpolation = cv::INTER_LINEAR);
  mean_0 = 103.94;
  mean_1 = 116.78;
  mean_2 = 123.68;
  scale_factor = 0.0;
  convert_color = false;
for mobilenet_ssd:
  height = width = 300;
  interpolation = cv::INTER_LINEAR);
  mean_0 = 127.5;
  mean_1 = 127.5;
  mean_2 = 127.5;
  scale_factor = 0.007843;
  convert_color = false;
for yolov1:
  interpolation = CV_INTER_LINEAR);
  mean_0 = 127.5;
  mean_1 = 127.5;
  mean_2 = 127.5;
  scale_factor = 0.007843;
  convert_color = true;
*************************************/

/**
 * @brief A wrapper that performs a type of preprocess.
 *
 * An algorithm wrapper to complete the classification network
 * preprocessing and other algorithms with the same preprocessing
 */

class ImageNetPreprocess : public Preprocess {
 public:
  /**
   * @brief Construct an imagenet preprocess object
   *
   * using default parameters
   */
  ImageNetPreprocess();
  explicit ImageNetPreprocess(Bmcv* processor);
  /**
   * @brief Construct an imagenet preprocess object
   *
   * @param height  The height of network input
   * @param width  The width of network input
   * @param interpolation  The interpolation method of opencv resize
   * @param mean_0  The bias of channel B
   * @param mean_0  The bias of channel G
   * @param mean_0  The bias of channel R
   * @param scale_factor  The scale factor of all channel
   * @param convert_color  The flag that whether to do BGR2RGB
   */
  ImageNetPreprocess(
      int   height,
      int   width,
      int   interpolation,
      float mean_0,
      float mean_1,
      float mean_2,
      float scale_factor,
      bool  convert_color);
  ImageNetPreprocess(
      Bmcv*         processor,
      int           height,
      int           width,
      int           interpolation,
      float         mean_0,
      float         mean_1,
      float         mean_2,
      float         scale_factor,
      bool          convert_color);
  virtual void process_tpu(Tensor<float>& input, Frame& frame);
  virtual ~ImageNetPreprocess() {}
  /**
   * @brief The execution function of imagenet preprocess
   *
   * @param input The input data system mem pointer
   * @param frame The original image
   */
  virtual void process_cpu(float* input, cv::Mat& frame);

 private:
  int height_;
  int width_;
  int interpolation_;
  std::vector<float> negative_mean_;
  float scale_factor_;
  bool convert_color_;
};

/**
 * @brief A wrapper that performs the yolov2&3 preprocess.
 *
 * An algorithm wrapper to complete the preprocessing of
 *    yolov2&3
 */

class Yolov23Preprocess : public Preprocess {
 public:
  /**
   * @brief Construct an yolov23 preprocess object
   *
   * using default parameters
   */
  Yolov23Preprocess();
#ifdef BMCV
  explicit Yolov23Preprocess(Bmcv* processor);
  Yolov23Preprocess(Bmcv* processor, int height, int width);
  virtual void process_tpu(Tensor<float>& input, Frame& frame);
#endif
  /**
   * @brief Construct an yolov23 preprocess object
   *
   * @param height  The height of network input
   * @param width  The width of network input
   */
  Yolov23Preprocess(int height, int width);
  virtual ~Yolov23Preprocess() {}
  /**
   * @brief The execution function of yolov23 preprocess
   *
   * @param input The input data system mem pointer
   * @param frame The original image
   */
  virtual void process_cpu(float* input, cv::Mat& frame);

 private:
  int height_;
  int width_;
};

/**
 * @brief A wrapper that performs the caffe-fasterrcnn preprocess.
 *
 * An algorithm wrapper to complete the preprocessing of
 *    caffe-fasterrcnn
 */

class FasterRcnnPreprocess : public Preprocess {
 public:
  /**
   * @brief Construct an caffe—fasterrcnn preprocess object
   *
   * using default parameters
   */
  FasterRcnnPreprocess();
#ifdef BMCV
  explicit FasterRcnnPreprocess(Bmcv* processor);
  FasterRcnnPreprocess(Bmcv* processor, int height, int width);
  virtual void process_tpu(bm_device_mem_t* input, Frame& frame);
#endif
  /**
   * @brief Construct an caffe-fasterrcnn preprocess object
   *
   * @param height  The height of network input
   * @param width  The width of network input
   */
  FasterRcnnPreprocess(int height, int width);
  virtual ~FasterRcnnPreprocess() {}
  /**
   * @brief The execution function of caffe-fasterrcnn preprocess
   *
   * @param input The input data system mem pointer
   * @param frame The original image
   */
  virtual void process_cpu(float* input, cv::Mat& frame);
  std::vector<float> get_im_info();
  std::vector<float> get_hw_scale();

 private:
  int height_;
  int width_;
  std::vector<float> im_info_;
  std::vector<float> hw_scale_;
};

/***********************************
for mtcnn:
 height = width = 300;
 interpolation = cv::INTER_LINEAR);
 mean_0 = 127.5;
 mean_1 = 127.5;
 mean_2 = 127.5;
 scale_factor = 0.0078125;
 convert_color = false;
************************************/

typedef struct FaceRect {  // facebbox info
  float x1;
  float y1;
  float x2;
  float y2;
  float score;

  bool operator==(const FaceRect &t) const {
    return (t.x1 == this->x1 &&
        t.y1 == this->y1 &&
        t.x2 == this->x2 &&
        t.y2 == this->y2 &&
        t.score == this->score);
  }
} FaceRect;

typedef struct FacePts {  // face landmark info
  float x[5], y[5];
  bool operator==(const FacePts &t) const {
    bool ret = true;
    for (int i = 0; i < 5; ++i) {
      if (t.x[i] != this->x[i] || t.y[i] != this->y[i]) {
         ret = false;
         break;
      }
    }
    return ret;
  }
} FacePts;

typedef struct FaceInfo {  // face info
  FaceRect bbox;
  cv::Vec4f regression;
  FacePts facePts;
  double roll;
  double pitch;
  double yaw;
  double distance;
  int imgid;

  bool operator==(const FaceInfo &t) const {
    return (t.bbox == this->bbox &&
        t.regression == this->regression &&
        t.roll == this->roll &&
        t.pitch == this->pitch &&
        t.yaw == this->yaw &&
        t.distance == this->distance &&
        t.imgid == this->imgid);
  }
} FaceInfo;

/**
 * @brief A wrapper that performs the mtcnn P/R/Onet preprocess.
 *
 * An algorithm wrapper to complete the preprocessing of
 *    all stages in mtcnn
 */

class MtcnnPreprocess : public Preprocess {
 public:
  /**
   * @brief Construct an mtcnn preprocess object
   *
   * using default parameters
   */
  MtcnnPreprocess();
#ifdef BMCV
  explicit MtcnnPreprocess(Bmcv* processor);
  MtcnnPreprocess(
      Bmcv*         processor,
      float         mean_0,
      float         mean_1,
      float         mean_2,
      float         scale_factor);
  MtcnnPreprocess(
      Bmcv&         processor,
      float         mean_0,
      float         mean_1,
      float         mean_2,
      float         scale_factor);
  virtual void process_tpu(Tensor<float>& input, Frame& frame) {}
  void process_tpu(
      Tensor<float>&   input,
      Frame&           frame,
      int              height,
      int              width);
#endif
  /**
   * @brief Construct an mtcnn preprocess object
   *
   * @param mean_0  The bias of channel B
   * @param mean_0  The bias of channel G
   * @param mean_0  The bias of channel R
   * @param scale_factor  The scale factor of all channel
   */
  MtcnnPreprocess(
      float mean_0,
      float mean_1,
      float mean_2,
      float scale_factor);
  virtual ~MtcnnPreprocess() {}
  virtual void process_cpu(float* input, cv::Mat& frame) {}
  /**
   * @brief The execution function of mtcnn Pnet preprocess
   *
   * @param input  The input data system mem pointer
   * @param frame  The original image
   * @param height  The height of network input
   * @param width  The width of network input
   */
  void process_cpu(float* input, cv::Mat& frame, int height, int width);
  /**
   * @brief The execution function of mtcnn image pyramid
   *   scale factor generate
   *
   * @param frame  The original image
   * @param factor  The the basic factor
   * @param minSize  The min size of face
   */
  std::vector<double> face_scale_gen(
      cv::Mat& frame,
      double   factor,
      int      minSize);
  // for R/O net preprocess (inlcude crop pad joint)
  /**
   * @brief The execution function of mtcnn Rnet/Onet input preprocess
   *
   * @param input_data  The input data system mem pointer
   * @param input_height  The height of Pnet/Onet
   * @param input_width  The width of Pnet/Onet
   * @param regressed_pading  The padded bbox size
   * @param regressed_rects  The regression bbox size
   */
  void ronet_pre_img(
      float*                 input_data,
      cv::Mat&               frame,
      int                    input_height,
      int                    input_width,
      std::vector<FaceInfo>& regressed_pading,
      std::vector<FaceInfo>& regressed_rects);
  /**
   * @brief The execution function of mtcnn Rnet/Onet preprocess
   *
   * @param input  The network input data system mem pointer
   * @param all_data  The data of all face which detected in Pnet/Rnet
   * @param input_shape  The shape of Rnet/Onet input
   * @param num_done  The number of img that have been processed
   * @param num_work  The number of img that will be processed in this
   *     batch
   */
  void ronet_preprocess_cpu(
      float*            input,
      float*            all_data,
      std::vector<int>& input_shape,
      int               num_done,
      int               num_work);

 private:
  std::vector<float> negative_mean_;
  float scale_factor_;
  float face_factor_;
  int minSize_;
};

}  // namespace sail
