"""The implementation of face detection mtcnn
"""
from __future__ import print_function
from __future__ import division
import numpy as np
import cv2, logging
from typing import Union, List, Dict
from ...engine.base_engine import BaseEngine
from ...utils.extend_math import softmax
from ...utils.box_operation import restore_param, fix_bbox_boundary, nms
from ...kitconfig import ChipMode
from ...engine import Engine

class FaceDetectionMTCNN(BaseEngine):
    """Construct mtcnn face detector
    """
    def __init__(self, pnet, rnet, onet,
                 detected_size,
                 thresholds, nms_thresholds, xform, min_size=20, factor=0.709):
        super(FaceDetectionMTCNN, self).__init__()
        # detected_size: (h, w)
        # image will be resize before processing detection
        # return bbox with respect to original image
        logging.info("mtcnn face detector model initialization")
        # init P/R/Onet
        self.pnet = PNet(thresholds=thresholds[0], nms_thresholds=nms_thresholds[0], min_size=min_size, factor=factor, **pnet)
        self.rnet = RNet(thresholds=thresholds[1], nms_thresholds=nms_thresholds[1], **rnet)
        self.onet = ONet(thresholds=thresholds[2], nms_thresholds=nms_thresholds[2], **onet)
        self.detected_size = detected_size
        self.xform = xform

    def predict(self, images, detected_size=None):
        self.time.tic()
        if isinstance(images, list):
            use_batch = True
        else:
            logging.debug("input images will be forced to a list of np.ndarray")
            use_batch = False
            images = [images]

        result_bboxes = []
        result_points = []
        result_probs = []
        for image in images:
            image_size = image.shape
            # MTCNN read image with RGB matlab format
            # image = self.bgr2rgb(image)

            if detected_size is None:
                detected_size = self.detected_size

            image, rescale_param = self.rescale_image(image, detected_size, True)
            image = self.xform_img(image, self.xform)[0]
            # three stages
            bboxes = self.pnet.predict(image)
            bboxes = self.rnet.predict(image, bboxes)
            bboxes, points = self.onet.predict(image, bboxes)

            if len(bboxes) > 0:
                bboxes, points = restore_param(bboxes, points, rescale_param)
                bboxes = fix_bbox_boundary(bboxes, image_size)
                # probs = [bbox[-1] for bbox in bboxes]
                probs = [np.asscalar(bbox[-1]) for bbox in bboxes]
                bboxes = [np.fix(bbox[:4]) for bbox in bboxes]
                area = np.array([(bbox[2]-bbox[0])*(bbox[3]-bbox[1]) for bbox in bboxes]) * (-1)
                argidx = np.argsort(area)
                bboxes = np.array(bboxes)[argidx]
                probs = np.array(probs)[argidx]
                result_bboxes.append(bboxes.tolist())
                result_probs.append(probs.tolist())
                if points is None:
                    result_points.append([None])
                else:
                    points = np.array(points)[argidx]
                    result_points.append(points.tolist())
            else:
                result_bboxes.append([])
                result_points.append([])
                result_probs.append([])
        self.time.toc()
        if use_batch:
            return (result_bboxes, result_points, result_probs)
        else:
            return (result_bboxes[0], result_points[0], result_probs[0])

    def set_param(self, config):
        pnet_param = dict()
        rnet_param = dict()
        onet_param = dict()
        for key in config:
            if key == 'thresholds':
                pnet_param['thresholds'] = config[key][0]
                rnet_param['thresholds'] = config[key][1]
                onet_param['thresholds'] = config[key][2]
            elif key == 'nms_thresholds':
                pnet_param['nms_thresholds'] = config[key][0]
                rnet_param['nms_thresholds'] = config[key][1]
                onet_param['nms_thresholds'] = config[key][2]
            elif key == 'min_size':
                pnet_param['min_size'] = config[key]
        self.pnet.set_param(pnet_param)
        self.rnet.set_param(rnet_param)
        self.onet.set_param(onet_param)


class PNet(BaseEngine):
    def __init__(self, arch, mode,
                 thresholds, nms_thresholds, min_size, factor):

        super(PNet, self).__init__()
        self.net = Engine(arch, mode)
        self.thresholds = thresholds
        self.nms_thresholds = nms_thresholds
        self.detected_min_size = min_size  # 20
        self.factor = factor  # 0.709
        self.input_names = arch['input_names'] # ['data',...]

    def set_param(self, param):
        for key in param:
            if key == 'thresholds':
                self.thresholds = param[key]
            elif key == 'nms_thresholds':
                self.nms_thresholds = param[key]
            elif key == 'min_size':
                self.detected_min_size = param[key]

    def predict(self, img):
        self.time.tic()
        factor_count = 0
        total_boxes = np.zeros((0, 9), np.float)
        h = img.shape[0]
        w = img.shape[1]
        minl = min(h, w) # minvalue between w&&h
        img = img.astype(np.float32)
        m = 12.0 / self.detected_min_size
        minl = minl * m

        # create scale pyramid
        scales = []
        while minl >= 12:
            scales.append(m * pow(self.factor, factor_count))
            minl *= self.factor
            factor_count += 1

        # first stage
        for scale in scales:
            hs = int(np.ceil(h*scale))
            ws = int(np.ceil(w*scale))
            im_data = cv2.resize(img, (ws, hs))  # default is bilinear
            im_data = np.swapaxes(im_data, 0, 2)
            im_data = np.array([im_data], dtype=np.float32)
            input_data = {self.input_names[0]: im_data}
            # print("Pnet input name:{}".format(self.input_names[0]))
            out = self.net.predict(input_data)
            if 'prob1' not in out:
                out['prob1'] = softmax(out['conv4-1'], 1, 1)
            boxes = generateBoundingBox(out['prob1'][0, 1, :, :],\
                out['conv4-2'][0], scale, self.thresholds)
            if boxes.shape[0] != 0:
                pick = nms(boxes, 0.5, 'Union')
                if len(pick) > 0:
                    boxes = boxes[pick, :]
            if boxes.shape[0] != 0:
                total_boxes = np.concatenate((total_boxes, boxes), axis=0)
        #####
        # 1 #
        #####
        logging.debug("mtcnn pnet [1]: {:d}".format(total_boxes.shape[0]))
        numbox = total_boxes.shape[0]
        if numbox > 0:
            # nms
            pick = nms(total_boxes, self.nms_thresholds, 'Union')
            total_boxes = total_boxes[pick, :]
            # logging.debug("[2]: {:d}".format(total_boxes.shape[0]))

            # revise and convert to square
            regh = total_boxes[:, 3] - total_boxes[:, 1]
            regw = total_boxes[:, 2] - total_boxes[:, 0]
            t1 = total_boxes[:, 0] + total_boxes[:, 5]*regw
            t2 = total_boxes[:, 1] + total_boxes[:, 6]*regh
            t3 = total_boxes[:, 2] + total_boxes[:, 7]*regw
            t4 = total_boxes[:, 3] + total_boxes[:, 8]*regh
            t5 = total_boxes[:, 4]
            total_boxes = np.array([t1, t2, t3, t4, t5]).T
            total_boxes = rerec(total_boxes)  # convert box to square
            # logging.debug("[4]: {:d}".format(total_boxes.shape[0]))

            total_boxes[:, 0:4] = np.fix(total_boxes[:, 0:4])
            # logging.debug("[4.5]: {:d}".format(total_boxes.shape[0]))
        self.time.toc()
        return total_boxes


class RNet(BaseEngine):
    def __init__(self, arch, mode,
                 thresholds, nms_thresholds):

        super(RNet, self).__init__()
        self.net = Engine(arch, mode)
        self.thresholds = thresholds
        self.nms_thresholds = nms_thresholds
        self.input_names = arch['input_names']

    def set_param(self, param):
        for key in param:
            if key == 'thresholds':
                self.thresholds = param[key]
            elif key == 'nms_thresholds':
                self.nms_thresholds = param[key]

    def predict(self, image, total_boxes):
        self.time.tic()
        width = image.shape[1]
        height = image.shape[0]
        numbox = total_boxes.shape[0]
        if numbox == 0:
            return total_boxes

        [dy, edy, dx, edx, y, ey, x, ex, tmpw, tmph] =\
            pad(total_boxes, width, height)
        # second stage
        # construct input for RNet
        tempimg = np.zeros((numbox, 24, 24, 3), np.float32)  # (24, 24, 3, numbox)
        for k in range(numbox):
            tmp = np.zeros((int(tmph[k]) + 1, int(tmpw[k]) + 1, 3))
            tmp[int(dy[k]):int(edy[k])+1, int(dx[k]):int(edx[k])+1] =\
                image[int(y[k]):int(ey[k])+1, int(x[k]):int(ex[k])+1]
            tempimg[k, :, :, :] = cv2.resize(tmp, (24, 24))
        # RNet
        tempimg = np.swapaxes(tempimg, 1, 3)
        input_data = {self.input_names[0]: tempimg}
        out = self.net.predict(input_data)
        if 'prob1' not in out:
            out['prob1'] = softmax(out['conv5-1'], 1, 1)
        score = out['prob1'][:, 1]
        pass_t = np.where(score > self.thresholds)[0]
        score = np.array([score[pass_t]]).T
        total_boxes = np.concatenate((total_boxes[pass_t, 0:4], score), axis=1)
        # logging.debug("[5]: {:d}".format(total_boxes.shape[0]))
        mv = out['conv5-2'][pass_t, :].T
        if total_boxes.shape[0] > 0:
            pick = nms(total_boxes, self.nms_thresholds, 'Union')
            # print('pick', pick)
            if len(pick) > 0:
                total_boxes = total_boxes[pick, :]
                # logging.debug("[6]: {:d}".format(total_boxes.shape[0]))
                total_boxes = bbreg(total_boxes, mv[:, pick])
                # logging.debug("[7]: {:d}".format(total_boxes.shape[0]))
                total_boxes = rerec(total_boxes)
                # logging.debug("[8]: {:d}".format(total_boxes.shape[0]))
        #####
        # 2 #
        #####
        logging.debug("mtcnn rnet [2]: {:d}".format(total_boxes.shape[0]))
        self.time.toc()
        return total_boxes


class ONet(BaseEngine):
    def __init__(self, arch, mode,
                 thresholds, nms_thresholds, only_points=False):

        super(ONet, self).__init__()
        self.net = Engine(arch, mode)
        self.mode = mode
        self.thresholds = thresholds
        self.nms_thresholds = nms_thresholds
        self.only_points = only_points
        self.input_names = arch['input_names']

    def set_param(self, param):
        for key in param:
            if key == 'thresholds':
                self.thresholds = param[key]
            elif key == 'nms_thresholds':
                self.nms_thresholds = param[key]

    def predict(self, image, bboxes):
        self.time.tic()

        width = image.shape[1]
        height = image.shape[0]
        numbox = bboxes.shape[0]
        if numbox == 0:
            return bboxes, []
        if self.only_points:
            original_bboxes = bboxes.copy()
        bboxes = rerec(bboxes)
        # third stage
        total_boxes = np.fix(bboxes)

        [dy, edy, dx, edx, y, ey, x, ex, tmpw, tmph] =\
            pad(total_boxes, width, height)

        tempimg = np.zeros((numbox, 48, 48, 3), np.float32)
        for k in range(numbox):
            tmp = np.zeros((int(tmph[k]), int(tmpw[k]), 3), np.float32)
            tmp[int(dy[k]):int(edy[k])+1, int(dx[k]):int(edx[k])+1] =\
                image[int(y[k]):int(ey[k])+1, int(x[k]):int(ex[k])+1]
            tempimg[k, :, :, :] = cv2.resize(tmp, (48, 48))
        # ONet
        tempimg = np.swapaxes(tempimg, 1, 3)
        input_data = {self.input_names[0]: tempimg} 
        out = self.net.predict(input_data)
        if 'prob1' not in out:
            out['prob1'] = softmax(out['conv6-1'], 1, 1)
        assert out['prob1'].shape[0] == numbox
        score = out['prob1'][:, 1]
        points = out['conv6-3']
        pass_t = np.where(score > self.thresholds)[0]
        points = points[pass_t, :]
        score = np.array([score[pass_t]]).T
        total_boxes = np.concatenate((total_boxes[pass_t, 0:4], score), axis=1)

        if self.only_points:
            h = total_boxes[:, 3] - total_boxes[:, 1] + 1
            w = total_boxes[:, 2] - total_boxes[:, 0] + 1

            points = np.repeat([w, h], 5, axis=0).T * points
            points += np.repeat(np.stack([total_boxes[:, 0],\
                total_boxes[:, 1]]), 5, axis=0).T -1
            total_boxes = np.concatenate((original_bboxes[pass_t, 0:4], score), axis=1)
        else:
            mv = out['conv6-2'][pass_t, :].T

            # h = total_boxes[:, 3] - total_boxes[:, 1] + 1
            # w = total_boxes[:, 2] - total_boxes[:, 0] + 1

            # points[:, 0:5] = np.tile(w, (5, 1)).T * points[:, 0:5] + np.tile(total_boxes[:, 0], (5,1)).T - 1 
            # points[:, 5:10] = np.tile(h, (5, 1)).T * points[:, 5:10] + np.tile(total_boxes[:, 1], (5,1)).T -1

            idxc1 = np.where(total_boxes[:, 3] > image.shape[0])[0]  # ey>imgh
            idxc3 = np.where(total_boxes[:, 2] > image.shape[1])[0]  # ex>imgw
            idxc_union = np.union1d(idxc1, idxc3)

            h = total_boxes[:, 3] - total_boxes[:, 1] + 1
            w = total_boxes[:, 2] - total_boxes[:, 0] + 1

            w1 = ex[pass_t] - x[pass_t] + 1
            h1 = ey[pass_t] - y[pass_t] + 1

            # Old implementation
            # pointsN = points.copy()
            # pointsN[:, 0:5] = np.tile(w1, (5, 1)).T * (points[:, 0:5]) + np.tile(x[pass_t], (5, 1)).T - 1
            # pointsN[:, 5:10] = np.tile(h1, (5, 1)).T * (points[:, 5:10]) + np.tile(y[pass_t], (5, 1)).T - 1

            # points[:, 0:5] = np.tile(w, (5, 1)).T * points[:, 0:5] + np.tile(total_boxes[:, 0], (5, 1)).T - 1
            # points[:, 5:10] = np.tile(h, (5, 1)).T * points[:, 5:10] + np.tile(total_boxes[:, 1], (5, 1)).T - 1

            # if idxc1.shape[0] > 0:
            #     points[idxc1, :] = pointsN[idxc1, :]

            # if idxc3.shape[0] > 0:
            #     points[idxc3, :] = pointsN[idxc3, :]

            w[idxc_union] = w1[idxc_union]
            h[idxc_union] = h1[idxc_union]

            _total_boxes = total_boxes.copy()
            _total_boxes[idxc_union, 0] = x[pass_t[idxc_union]]
            _total_boxes[idxc_union, 1] = y[pass_t[idxc_union]]

            points = np.repeat([w, h], 5, axis=0).T * points
            points += np.repeat(np.stack([_total_boxes[:, 0], _total_boxes[:, 1]]), 5, axis=0).T -1

            if total_boxes.shape[0] > 0:
                total_boxes = bbreg(total_boxes, mv[:, :])
                # logging.debug("[10]: {:d}".format(total_boxes.shape[0]))
                pick = nms(total_boxes, self.nms_thresholds, 'Min')
                if len(pick) > 0:
                    total_boxes = total_boxes[pick, :]
                    # logging.debug("[11]: {:d}".format(total_boxes.shape[0]))
                    points = points[pick, :]
            #####
            # 3 #
            #####
            logging.debug("mtcnn onet [3]: {:d}".format(total_boxes.shape[0]))

        self.time.toc()
        return total_boxes, points


def generateBoundingBox(map, reg, scale, t):
    stride = 2
    cellsize = 12
    map = map.T
    dx1 = reg[0, :, :].T
    dy1 = reg[1, :, :].T
    dx2 = reg[2, :, :].T
    dy2 = reg[3, :, :].T
    (x, y) = np.where(map >= t)

    yy = y
    xx = x

    score = map[x, y]
    reg = np.array([dx1[x, y], dy1[x, y], dx2[x, y], dy2[x, y]])

    if reg.shape[0] == 0:
        pass
    boundingbox = np.array([yy, xx]).T

    bb1 = np.fix((stride * (boundingbox) + 1) / scale).T  # matlab index from 1, so with "boundingbox-1"
    bb2 = np.fix((stride * (boundingbox) + cellsize - 1 + 1) / scale).T  # while python don't have to
    score = np.array([score])

    boundingbox_out = np.concatenate((bb1, bb2, score, reg), axis=0)

    # print('(x,y)',x,y)
    # print('score', score)
    # print('reg', reg)

    return boundingbox_out.T

def rerec(bboxA):
    # convert bboxA to square
    w = bboxA[:, 2] - bboxA[:, 0]
    h = bboxA[:, 3] - bboxA[:, 1]
    l = np.maximum(w, h).T
    bboxA[:, :4] += np.array([w-l, h-l, l-w, l-h]).T * 0.5

    return bboxA


def bbreg(boundingbox, reg):
    reg = reg.T

    # calibrate bouding boxes
    if reg.shape[1] == 1:
        print("reshape of reg")
        pass  # reshape of reg
    w = boundingbox[:, 2] - boundingbox[:, 0] + 1
    h = boundingbox[:, 3] - boundingbox[:, 1] + 1

    boundingbox[:, 0:4] += reg * np.array([w, h, w, h]).T

    return boundingbox


def pad(boxesA, w, h):
    boxes = boxesA.copy()
    # print('#################')
    # print('boxes', boxes)
    # print('w,h', w, h)

    tmph = boxes[:, 3] - boxes[:, 1] + 1
    tmpw = boxes[:, 2] - boxes[:, 0] + 1
    numbox = boxes.shape[0]

    # print('tmph', tmph)
    # print('tmpw', tmpw)

    dx = np.ones(numbox)
    dy = np.ones(numbox)
    edx = tmpw
    edy = tmph

    x = boxes[:, 0:1][:, 0]
    y = boxes[:, 1:2][:, 0]
    ex = boxes[:, 2:3][:, 0]
    ey = boxes[:, 3:4][:, 0]

    tmp = np.where(ex > w)[0]
    if tmp.shape[0] != 0:
        edx[tmp] = -ex[tmp] + w-1 + tmpw[tmp]
        ex[tmp] = w-1

    tmp = np.where(ey > h)[0]
    if tmp.shape[0] != 0:
        edy[tmp] = -ey[tmp] + h-1 + tmph[tmp]
        ey[tmp] = h-1

    tmp = np.where(x < 1)[0]
    if tmp.shape[0] != 0:
        dx[tmp] = 2 - x[tmp]
        x[tmp] = np.ones_like(x[tmp])

    tmp = np.where(y < 1)[0]
    if tmp.shape[0] != 0:
        dy[tmp] = 2 - y[tmp]
        y[tmp] = np.ones_like(y[tmp])

    # for python index from 0, while matlab from 1
    dy = np.maximum(0, dy-1)
    dx = np.maximum(0, dx-1)
    y = np.maximum(0, y-1)
    x = np.maximum(0, x-1)
    edy = np.maximum(0, edy-1)
    edx = np.maximum(0, edx-1)
    ey = np.maximum(0, ey-1)
    ex = np.maximum(0, ex-1)

    return [dy, edy, dx, edx, y, ey, x, ex, tmpw, tmph]
