"""object detection factory
create various object detectors
"""
from __future__ import print_function
from ..utils.utils import load_predefined_config, load_config
from ..kitconfig import ChipMode, ObjDetModel
import os, sys
from . basefactory import BaseFactory

localpath = os.path.dirname(os.path.realpath(__file__))

# detected_size: (width, height)
default_yolov3_param = {'detected_size': (416, 416),
    'threshold': 0.5, 'nms_threshold': 0.45, 'num_classes': 80,
    'anchors': [[(10, 13), (16, 30), (33, 23)],
                [(30, 61), (62, 45), (59, 119)],
                [(116, 90), (156, 198), (373, 326)]]}

# global param for mobilenetssd
default_mobilenetssd_param = {'detected_size': (300, 300),
    'threshold': 0.25, 'nms_threshold': 0.45, 'num_classes': 21,
    'priorbox_num': 1917}

# global param for mobilenetyolov3
default_mobilenetyolov3_param = {'detected_size': (416, 416),
    'threshold': 0.2, 'nms_threshold': 0.45, 'num_classes': 80,
    'anchors': [[(10, 13), (16, 30), (33, 23)],
                [(30, 61), (62, 45), (59, 119)],
                [(116, 90), (156, 198), (373, 326)]]}

# global param for fasterrcnn
default_fasterrcnn_param = {'detected_size': (600, 800),
                       'thresholds': [0.6, 0.3], 'num_classes': 21,
                       'global_cfg': {'pre_nms_topN': 300, 
                       'post_nms_topN': 32,
                       'anchor_min_size': 16,
                       'nms_threshold': 0.7}}

class ObjectDetector(BaseFactory):
    def __init__(self, model_path=None):
        super(ObjectDetector, self).__init__(model_path)
    """list the detection model in sailalgokit
    """
    def get_supported_models(self):
        return list(algo_name.value for algo_name in ObjDetModel)

    def create(self, detector_model, mode, config_path=None,
            extra_args={}):
        if detector_model is ObjDetModel.YOLOV3:
            from ..algo_cv.det.object_detection_yolov3 import\
                ObjectDetectionYOLOV3 as Detector
        elif detector_model is ObjDetModel.MOBILENETSSD:
            from ..algo_cv.det.object_detection_mobilenetssd import\
                ObjectDetectionMOBILENETSSD as Detector
        elif detector_model is ObjDetModel.MOBILENETYOLOV3:
            from ..algo_cv.det.object_detection_mobilenetyolov3 import\
                ObjectDetectionMOBILENETYOLOV3 as Detector
        elif detector_model is ObjDetModel.FASTERRCNN:
            from ..algo_cv.det.object_detection_fasterrcnn import\
                ObjectDetectionFASTERRCNN as Detector
        else:
            raise ValueError('not a valid detector_model:', detector_model)
        if config_path is None:
            arch_config = load_predefined_config(self.model_path, mode, detector_model.value)
        else:
            arch_config = load_config(config_path, self.model_path, mode)
        arch_config = self.load_param(detector_model, arch_config, extra_args)
        detector = Detector(**arch_config)
        return detector

    def set_param(self):
        pass

    def load_param(self, detector_model, arch_config, extra_args):
        """unified all arguments to one dict
        """
        if detector_model is ObjDetModel.YOLOV3:
            default_param = default_yolov3_param
        elif detector_model is ObjDetModel.MOBILENETSSD:
            default_param = default_mobilenetssd_param
        elif detector_model is ObjDetModel.MOBILENETYOLOV3:
            default_param = default_mobilenetyolov3_param
        elif detector_model is ObjDetModel.FASTERRCNN:
            default_param = default_fasterrcnn_param
        else:
            raise ValueError('not a valid detector_model:', detector_model)

        for key in extra_args:
            if key not in arch_config:
                arch_config[key] = extra_args[key]
        for key in default_param:
            if key in arch_config:
                pass
            else:
                arch_config[key] = default_param[key]
        return arch_config
