"""The BM1682 model engine
"""
from __future__ import print_function
from __future__ import division
import numpy as np
from . base_engine import BaseEngine
import sail # engine core lib
from .. kitconfig import ChipMode
import functools as ft
from operator import mul

class BM1682Engine(BaseEngine):
    """Construct BM1682 engine from lib sail
    """
    def bm1682_init(self):
        """Init bm1682 inference
        """
        inference = sail.Inference(self.context_path, ChipMode.BM1682.value,
                                   self.tpus, sail.IOMode.SYSIO,
                                   self.is_dynamic, self.is_async)
        return inference

    def __init__(self, context_path, is_dynamic, is_async,
                 tpus, input_shapes, input_names, output_names):
        """Init BM1682Engine

        Args:
            All arguments init from net config json file
            context_path(str). context ir path
            is_dynamic(bool). dynamic status
            is_async(bool). async status
            tpus(str). tpu number
            input_shapes(List). net input shapes (max shape for dynamic)
        """
        super(BM1682Engine, self).__init__()
        self.is_dynamic = is_dynamic
        self.is_async = is_async
        self.context_path = context_path
        self.tpus = tpus
        self.inference = self.bm1682_init() # init bm1682 inference
        self.input_shapes = input_shapes
        self.graph_name = \
            self.inference.get_all_graph_names()[0]
        self.input_names = \
            self.inference.get_input_tensor_names(self.graph_name)
        self.output_names = \
            self.inference.get_output_tensor_names(self.graph_name)
        """
        set for async 
        _inference_status: BEGIN (first inference without d2s)
                           PROCESS (inference processing)
                           END (last d2s without inference)
        _tensor_mode: PING
                      PONG
        """
        self._tensor_mode = sail.TensorMode.PING
        self._inference_status = sail.AsyncStatus.BEGIN

    # def _set_input_shape(self, input_shape):
    #     """reset the input shape of network

    #        Args:
    #            input_shape(np.ndarray, dtype=int32)
    #     """
    #     input_shape_map = {}
    #     input_shape_map[self.input_names[0]] = input_shape
    #     if self.Async:
    #         self.inference.set_input_shape_async(self.graph_name, input_shape_map, self._tensor_mode)
    #     else:
    #         self.inference.set_input_shape_sync(self.graph_name, input_shape_map)

    def _sync_predict(self, input):
        """synchronous inference
        """
        for input_name in self.input_names:
            input[input_name] = \
                    np.array(input[input_name], dtype=np.float32)
        output = self.inference.process_sync(self.graph_name, input)
        for output_name in self.output_names:
            output_shape = output[output_name].shape
            # rm axis = x && x.value == 1 in the end
            # rm_idx = len(output_shape)
            # for idx in range(len(output_shape) - 1):
            #     idx = len(output_shape) - idx - 1
            #     if (output_shape[idx] == 1):
            #         rm_idx = idx
            # output_shape = output_shape[:rm_idx]
            # output[output_name] = output[output_name].reshape(output_shape)
            print ("{} output shape: {}".format(output_name, tuple(output_shape)))
        return output

    # def _async_predict(self, input_data):
    #     """asynchronous inference

    #     dynamic status: true | false
    #     inference status: BEGIN | PROCESS | END
    #     """
    #     data = np.array(input_data, dtype='float32')
    #     is_end = (data.shape[0] is 0) # set for async end
    #     if self.Dynamic is True: # dynamic
    #         if is_end is not True: # async not end
    #             input_shape = np.array([data.shape[0], data.shape[1],
    #                                    data.shape[2], data.shape[3]],
    #                                    dtype='int32')
    #             self._set_input_shape(input_shape)
    #             input_tensor = \
    #                 self.inference.get_input_tensor_sys(self.graph_name,
    #                                                     self.input_names[0])
    #             np.copyto(input_tensor, data)
    #             print ("{} input shape: {}".format(self.graph_name,\
    #                 input_tensor.shape))
    #             self.inference.process(self.graph_name, self._tensor_mode,\
    #                 self._inference_status)
    #             output = dict()
    #             if self._inference_status is not sail.AsyncStatus.BEGIN:
    #                 for output_name in self.output_names:
    #                     output_tensor = \
    #                         self.inference.get_output_tensor_sys(self.graph_name,\
    #                             output_name)
    #                     # sail default shape 4-dims n c h w
    #                     output_shape = \
    #                         self.inference.get_output_shape(self.graph_name,\
    #                             output_name)
    #                     if ((output_shape[2] == 1) and (output_shape[3] == 1)):
    #                         output[output_name] = \
    #                             output_tensor.copy().reshape(output_shape[0], \
    #                                 output_shape[1])
    #                     else:
    #                         output[output_name] = \
    #                             output_tensor.copy().reshape(output_shape[0], \
    #                                 output_shape[1], output_shape[2], output_shape[3])
    #                     print ("{} output shape: {}".format(output_name, output_shape))
    #             self._tensor_mode = sail.TensorMode.PING if self._tensor_mode is sail.TensorMode.PONG\
    #                 else sail.TensorMode.PONG
    #             self._inference_status = sail.AsyncStatus.PROCESS
    #             return output
    #         else: # async end
    #             self._inference_status = sail.AsyncStatus.END
    #             self.inference.process(self.graph_name, self._tensor_mode,\
    #                 self._inference_status)
    #             output = dict()
    #             for output_name in self.output_names:
    #                 output_tensor = \
    #                     self.inference.get_output_tensor_sys(self.graph_name,\
    #                         output_name)
    #                 # sail default shape 4-dims n c h w
    #                 output_shape = \
    #                     self.inference.get_output_shape(self.graph_name,\
    #                         output_name)
    #                 if ((output_shape[2] == 1) and (output_shape[3] == 1)):
    #                     output[output_name] = \
    #                         output_tensor.copy().reshape(output_shape[0], \
    #                             output_shape[1])
    #                 else:
    #                     output[output_name] = \
    #                         output_tensor.copy().reshape(output_shape[0], \
    #                             output_shape[1], output_shape[2], output_shape[3])
    #                 print ("{} output shape: {}".format(output_name, output_shape))
    #             return output
    #     else: # static
    #         if is_end is not True: # async not end
    #             input_shape = np.array([data.shape[0], data.shape[1],
    #                                    data.shape[2], data.shape[3]],
    #                                    dtype='int32') # np default type int64
    #             self._set_input_shape(input_shape)
    #             input_tensor = \
    #                 self.inference.get_input_tensor_sys(self.graph_name,
    #                                                     self.input_names[0])
    #             # copy the value to input tensor and keep the addr of
    #             # input_tensor unchanged
    #             np.copyto(input_tensor, data)
    #             print ("{} input shape: {}".format(self.graph_name,\
    #                 input_tensor.shape))
    #             self.inference.process(self.graph_name, self._tensor_mode,\
    #                 self._inference_status)
    #             output = dict()
    #             if self._inference_status is not sail.AsyncStatus.BEGIN:
    #                 for output_name in self.output_names:
    #                     output_tensor = \
    #                         self.inference.get_output_tensor_sys(self.graph_name,\
    #                             output_name)
    #                     # sail default shape 4-dims n c h w
    #                     output_shape = \
    #                         self.inference.get_output_shape(self.graph_name,\
    #                             output_name)
    #                     if ((output_shape[2] == 1) and (output_shape[3] == 1)):
    #                         output[output_name] = \
    #                             output_tensor.copy().reshape(output_shape[0], \
    #                                 output_shape[1])
    #                     else:
    #                         output[output_name] = \
    #                             output_tensor.copy().reshape(output_shape[0], \
    #                                 output_shape[1], output_shape[2], output_shape[3])
    #                     print ("{} output shape: {}".format(output_name, output_shape))
    #             self._tensor_mode = sail.TensorMode.PING if self._tensor_mode is sail.TensorMode.PONG\
    #                 else sail.TensorMode.PONG
    #             self._inference_status = sail.AsyncStatus.PROCESS
    #             return output
    #         else: # async in the end
    #             self._inference_status = sail.AsyncStatus.END
    #             self.inference.process(self.graph_name, self._tensor_mode,\
    #                 self._inference_status)
    #             output = dict()
    #             for output_name in self.output_names:
    #                 output_tensor = \
    #                     self.inference.get_output_tensor_sys(self.graph_name,\
    #                         output_name)
    #                 # sail default shape 4-dims n c h w
    #                 output_shape = \
    #                     self.inference.get_output_shape(self.graph_name,\
    #                         output_name)
    #                 if ((output_shape[2] == 1) and (output_shape[3] == 1)):
    #                     output[output_name] = \
    #                         output_tensor.copy().reshape(output_shape[0], \
    #                             output_shape[1])
    #                 else:
    #                     output[output_name] = \
    #                         output_tensor.copy().reshape(output_shape[0], \
    #                             output_shape[1], output_shape[2], output_shape[3])
    #                 print ("{} output shape: {}".format(output_name, output_shape))
    #             return output

    def predict(self, input):
        """The inference interface of BM1682Engine

        status:
            asynchronous: _async_predict
            synchronous: _sync_predict
        """
        output = {}

        if self.is_async is False:
            output = self._sync_predict(input) # for sync
        else:
            print("async coming soon...")
            # output = self._async_predict(input_data) # for async
        return output
