"""BMmodel engine

which can create engine based on chipmode
"""
from __future__ import print_function
from .base_engine import BaseEngine
from ..kitconfig import ChipMode
import logging, time
import numpy as np
from typing import Union, List, Dict

class BMEngine(BaseEngine):
    """Construct BMmodel inference engine
    """
    def __init__(self, arch, mode):
        super(BMEngine, self).__init__()
        if mode is ChipMode.BM1682:
            from . bm1682_engine import BM1682Engine as Engine
        elif mode is ChipMode.BM1684:
            from . bm1684_engine import BM1684Engine as Engine
        else:
            raise ValueError('not a valid mode: {}'.format(mode))
        self.mode = mode
        self.nets = Engine(**arch)
        self.input_info = arch['input_names']
        self.is_async = arch['is_async']
        self.max_batch_size = [item[0] for item in arch['input_shapes']][0]
        logging.debug('Initialize {} #{} cmdbuf:'.format(mode, len(arch)))

    def get_batch_list(self, num):
        """Get inference batch list from max_input_batch and
           real input batch
        """
        batch_list = []  # [max_batch, max_batch, num - n * max_batch]
        while num > 0:
            if self.max_batch_size <= num:
                batch_num = self.max_batch_size
                batch_list.append(batch_num)
                num -= batch_num
            else:
                batch_list.append(num)
                num -= num
        if self.is_async is True:
            batch_list.append(0)
        print ("inference batch list: {}".format(batch_list))
        return batch_list

    def predict(self, input_data):
        """Split the input data with batch list
           call the interface of BM1682/BM1684Engine to do inference
        """
        if len(input_data) == 1:
            num = input_data[self.input_info[0]].shape[0] # single input
        else:
            num = 1 # multi inputs
        batch_list = self.get_batch_list(num)
        print("net inference is async: {}".format(self.is_async))
        logging.debug("split input batch to subbatch with {}".format(batch_list))
        batch_num = 0 # done batch
        batch_flag = 2 # set for async
        if self.is_async is True:
            batch_flag = 0
        if len(input_data) == 1: # single input can get sub_batch
            for sub_batch in batch_list: # working batch
                subbatch_data = {}
                subbatch_data[self.input_info[0]] = \
                        np.ascontiguousarray(input_data[self.input_info[0]][batch_num:sub_batch+batch_num])
                logging.debug("forwarding subbatch input data shape: {}".format(subbatch_data[self.input_info[0]].shape))
                self.time.tic()
                partial_out = self.nets.predict(subbatch_data) # do inference
                self.time.toc()
                if batch_num == 0 or batch_flag < 2:  # first forward (sync batch_num == 0 | async batch_flag < 2)
                    out = partial_out
                else:
                    for key in out: # concatenate output
                        out[key] = np.concatenate((out[key], partial_out[key].copy()), axis=0)
                batch_num += sub_batch
                batch_flag += 1
        else: # multi_inputs batch_list = [1] for sync | batch_list = [1, 0] for async
            for sub_batch in batch_list: # check contiguous mem np.ndarray.flags['C_CONTIGUOUS']
                subbatch_data = {}
                if sub_batch == 1:
                    subbatch_data = input_data
                    logging.debug("forwarding subbatch input data num: {}".format(len(subbatch_data)))
                else:
                    for input_name in self.input_info:
                        subbatch_data[input_name] = np.ndarray((0), dtype=np.float32)
                        logging.debug("forwarding subbatch input data shape: {}".format(subbatch_data[input_name].shape))
                self.time.tic()
                partial_out = self.nets.predict(subbatch_data)
                self.time.toc()
                if batch_num == 0 or batch_flag < 2:  # first forward
                    out = partial_out
                else:
                    for key in out:
                        out[key] = np.concatenate((out[key], partial_out[key].copy()), axis=0)
                batch_num += sub_batch
                batch_flag += 1
        assert batch_num == sum(batch_list)
        return out
