"""The operation of box
"""
from __future__ import division
import numpy as np
import math
from typing import List

def get_largest_box_index(bboxes):
    """Return the largest box index
    """
    bboxes = np.array(bboxes)
    delta_x = bboxes[:, 2] - bboxes[:, 0]
    delta_y = bboxes[:, 3] - bboxes[:, 1]
    return np.argmax(delta_x*delta_y)

def extend_box(bbox, image_size, extend_ratio):
    """Zoom the input bbox
    """
    x = (bbox[2] - bbox[0]) * extend_ratio
    y = (bbox[3] - bbox[1]) * extend_ratio
    extend_bbox = []
    extend_bbox.append(int(max(0, (bbox[2] + bbox[0])/2 - x/2)))
    extend_bbox.append(int(max(0, (bbox[3] + bbox[1])/2 - y/2)))
    extend_bbox.append(int(min(image_size[1], (bbox[2] + bbox[0])/2 + x/2)))
    extend_bbox.append(int(min(image_size[0], (bbox[3] + bbox[1])/2 + y/2)))
    return extend_bbox

def fix_bbox_boundary(bboxes, image_size):
    """Fix the coordinates of bbox
    """
    bboxes[bboxes < 0] = 0
    bboxes[:, 2][bboxes[:, 2] > image_size[1]] = image_size[1]
    bboxes[:, 3][bboxes[:, 3] > image_size[0]] = image_size[0]
    return bboxes

def crop_and_shift(image, bbox, landmark):
    """Crop face && reset the landmark coordinates
    """
    landmark = np.array(landmark)
    radius = math.sqrt((bbox[2] - bbox[0])**2 + (bbox[3] - bbox[1])**2)*0.6
    new_bbox = [int(max(0, (bbox[2] + bbox[0])/2 - radius)),
                int(max(0, (bbox[3] + bbox[1])/2 - radius)),
                int(min(image.shape[1], (bbox[2] + bbox[0])/2 + radius)),
                int(min(image.shape[0], (bbox[3] + bbox[1])/2 + radius))]
    landmark[0:5] = landmark[0:5] - new_bbox[0]
    landmark[5:10] = landmark[5:10] - new_bbox[1]
    return image[new_bbox[1]:new_bbox[3], new_bbox[0]:new_bbox[2], :], landmark

def restore_param(bboxes, points, rescale_param):
    """restore bboxes with original image size when detecting in different size

    Args:
        bboxes:list. Input bbox
        points:list. Input landmark
        rescale_param: tuple. ex:(scale_y, scale_x, pad_top, pad_left)
                    scale_y or scale_x = original_size/detected_size
    
    Returns:
        bboxes:list. restore bbox
        points:list. restore point
    """
    bboxes[:, 0] = (bboxes[:, 0] - rescale_param[3])*rescale_param[0]
    bboxes[:, 1] = (bboxes[:, 1] - rescale_param[2])*rescale_param[1]
    bboxes[:, 2] = (bboxes[:, 2] - rescale_param[3])*rescale_param[0]
    bboxes[:, 3] = (bboxes[:, 3] - rescale_param[2])*rescale_param[1]
    if points is not None:
        points[:, 0:5] = (points[:, 0:5] - rescale_param[3])*rescale_param[1]
        points[:, 5:10] = (points[:, 5:10] - rescale_param[2])*rescale_param[1]
    return bboxes, points

def nms(boxes, threshold, type):
    """nms
    :boxes: [:,0:5]
    :threshold: 0.5 like
    :type: 'Min' or others ('union')
    :returns: idx list
    """
    if boxes.shape[0] == 0:
        return np.array([])
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    s = boxes[:, 4]
    area = np.multiply(x2-x1+1, y2-y1+1)
    I = np.array(s.argsort())  # read s using I

    pick = []
    while len(I) > 0:
        xx1 = np.maximum(x1[I[-1]], x1[I[0:-1]])
        yy1 = np.maximum(y1[I[-1]], y1[I[0:-1]])
        xx2 = np.minimum(x2[I[-1]], x2[I[0:-1]])
        yy2 = np.minimum(y2[I[-1]], y2[I[0:-1]])
        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        if type == 'Min':
            o = inter / np.minimum(area[I[-1]], area[I[0:-1]])
        else:
            o = inter / (area[I[-1]] + area[I[0:-1]] - inter)
        pick.append(I[-1])
        I = I[np.where(o <= threshold)[0]]
    return pick

def bbox_transform_inv(boxes, deltas):
    """proposal layer bbox operation

    to be added
    """
    if boxes.shape[0] == 0:
        return np.zeros((0, deltas.shape[1]), dtype=deltas.dtype)

    boxes = boxes.astype(deltas.dtype, copy=False)

    widths = boxes[:, 2] - boxes[:, 0] + 1.0
    heights = boxes[:, 3] - boxes[:, 1] + 1.0
    ctr_x = boxes[:, 0] + 0.5 * widths
    ctr_y = boxes[:, 1] + 0.5 * heights

    dx = deltas[:, 0::4]
    dy = deltas[:, 1::4]
    dw = deltas[:, 2::4]
    dh = deltas[:, 3::4]

    pred_ctr_x = dx * widths[:, np.newaxis] + ctr_x[:, np.newaxis]
    pred_ctr_y = dy * heights[:, np.newaxis] + ctr_y[:, np.newaxis]

    potential_of = np.where(dw > 100)[0]
    try:
        pred_w = np.exp(dw) * widths[:, np.newaxis]
        pred_h = np.exp(dh) * heights[:, np.newaxis]
    except(FloatingPointError):
        print ('WARNING: OVERFLOW OCCURED IN np.exp(dw) and/or np.exp(dh)')
        for i in range(dw.shape[0]):
            for j in range(dw.shape[1]):
                if dw[i][j]>50:
                    dw[i][j]=5
                if dh[i][j]>50:
                    dh[i][j]=5
        pred_w = np.exp(dw) * widths[:, np.newaxis]
        pred_h = np.exp(dh) * heights[:, np.newaxis]

    pred_boxes = np.zeros(deltas.shape, dtype=deltas.dtype)
    # x1
    pred_boxes[:, 0::4] = pred_ctr_x - 0.5 * pred_w
    # y1
    pred_boxes[:, 1::4] = pred_ctr_y - 0.5 * pred_h
    # x2
    pred_boxes[:, 2::4] = pred_ctr_x + 0.5 * pred_w
    # y2
    pred_boxes[:, 3::4] = pred_ctr_y + 0.5 * pred_h
    return pred_boxes