import os
import json
import logging


def exception_wrapper(message):
  def decorator(func):
    def wrapper(*args, **kwargs):
      try:
        ret = func(*args, **kwargs)
      except Exception as e:
        logging.error(message)
        print(repr(e))
        import traceback
        traceback.print_exc()
        exit(-1)
      else:
        return ret
    return wrapper
  return decorator

@exception_wrapper(message="Met an Error when split model.")
def split(platform, input_tensors, save_dir, graph_path, \
          params_path=None, outputs=None, dynamic=False, layout='NCHW'):
  """ Split the raw model into several submodels.

  Args:
    platform: Platform that trained the model.
              Options: tensorflow, mxnet, pytorch, caffe
    input_tensors: A dict contains the information of input tensors.
                  Format: {input_name: numpy.ndarray}
    save_dir: Path of directory to save submodels and splitting information.
    graph_path: Path to the graph description file of the model.
    params_path: Path to the parameters file of the model. Default None.
    outputs: A list contains the output tensor names. Default None.
    dynamic: True means input tensor shapes may change. Default False.
    layout: Layout of tensor. Default 'NCHW'.

  Returns:
    None.
  """
  save_dir = os.path.abspath(save_dir)
  graph_path = os.path.abspath(graph_path)
  if params_path is not None:
    params_path = os.path.abspath(params_path)
  supported_platforms = ["tensorflow", "mxnet", "pytorch", "caffe"]
  if platform not in supported_platforms:
    raise NotImplementedError("{} not supported by now.".format(platform))
  if platform == "mxnet":
    if params_path is None:
      raise RuntimeError("params_path should not be none for mxnet.")
    if not graph_path.endswith('.json') or not params_path.endswith('.params'):
      raise RuntimeError("models for mxnet should be like" + 
            "xxx.json and xxx.params.")
    from auto_deploy.splitter.mxnet_splitter import MxnetSplitter as SPL
    model_des = {"json_path": graph_path, "params_path": params_path,
                 "dynamic": dynamic, "input_tensors": input_tensors,
                 "layout": layout}
  elif platform == "tensorflow":
    if params_path is not None:
      raise RuntimeError("params_path should be None for tensorflow.")
    if outputs is None:
      raise RuntimeError("outputs should not be none for tensorflow.")
    if layout != 'NHWC':
      raise NotImplementedError("TF only support NHWC by now.")
    from auto_deploy.splitter.tensorflow_splitter \
      import TensorflowSplitter as SPL
    model_des = {"model_path": graph_path, "dynamic": dynamic,
                 "input_tensors": input_tensors, "output_names": outputs,
                 "layout": layout}
  splitter = SPL(model_des)
  splitter.convert_and_split(save_dir)

@exception_wrapper(message="Met an Error when compile model.")
def compile(folder, optimize=None, compare=False, target='BM1682'):
  """ Compile all the subgraphs which can deploy on sophon.

  Args:
    folder: path that contains splitted models,
            'graph_infos.json' must be under this folder

    Format of 'graph_infos.json':
    {
      "graph_num": graph_numbmer,
      "platform": "mxnet"
      "layout": "NCHW"
      "dynamic": False
      "graphs": [
        {
          "device": "cpu",
          "inputs": list of input tensor names,
          "outputs": list of output tensor names,
          "model_info": {
            "json": json_file_path,
            "params": params_file_path
          }
        }
      ]
      "tensors": [
        {
          "name": name,
          "shape": list for shape,
          "attr": "input" or "output" or "intermediate"
        }
      ]
    }
    optimize: optimizing mode, parameter of bmnet compiler.
    compare: if compare with cpu results when compiling.
    target: 'BM1682' or 'BM1684'(future).

  Returns: None.
  """
  if not os.path.isfile(os.path.join(folder, 'graph_infos.json')):
    raise RuntimeError("graph_infos.json not under {}.".format(folder))
  with open(os.path.join(folder, 'graph_infos.json'), 'r') as cfg:
    graph_infos = json.load(cfg)
  
  if graph_infos['platform'] == 'mxnet':
    from auto_deploy.compiler.mxnet_compiler import MxnetCompiler as CP
  elif graph_infos['platform'] == 'tensorflow':
    from auto_deploy.compiler.tensorflow_compiler import TensorflowCompiler as CP
  elif graph_infos['platform'] == 'pytorch':
    from auto_deploy.compiler.pytorch_compiler import PytorchCompiler as CP
  elif graph_infos['platform'] == 'caffe':
    from auto_deploy.compiler.caffe_compiler import CaffeCompiler as CP
  else:
    raise NotImplementedError("unsupport platform: {}".format\
            (graph_infos['platform']))
  compiler = CP(folder)
  compiler.compile_to_bm_models()

@exception_wrapper(message="Met an Error when infer model.")
def infer(folder, input_tensors):
  """ Run all the subgraphs as a pipeline.

  Args:
    folder: Directory path that contains splitted models and 'graph_infos.json'.

    Format of 'graph_infos.json':
    {
      "graph_num": graph_numbmer,
      "platform": "mxnet"
      "layout": "NCHW"
      "dynamic": False
      "graphs": [
        {
          "device": "cpu",
          "inputs": list of input tensor names,
          "outputs": list of output tensor names,
          "model_info": {
            "json": json_file_path,
            "params": params_file_path
          }
        }
      ]
      "tensors": [
        {
          "name": name,
          "shape": list for shape,
          "attr": "input" or "output" or "intermediate"
        }
      ]
    }
    input_tensors: Input tensors. Format: {name: value}.

  Returns:
    Output tensors. Format: {name: value}.
  """
  if not os.path.isfile(os.path.join(folder, 'graph_infos.json')):
    raise RuntimeError("graph_infos.json not under {}.".format(folder))
  with open(os.path.join(folder, 'graph_infos.json'), 'r') as cfg:
    graph_infos = json.load(cfg)
  platform = graph_infos["platform"]
  assert(platform in ["tensorflow", "mxnet", "caffe", "pytorch"])
  if platform == 'tensorflow':
    from auto_deploy.runner.tensorflow_runner import TensorflowRunner as RN
  elif platform == 'mxnet':
    from auto_deploy.runner.mxnet_runner import MxnetRunner as RN
  elif platform == 'caffe':
    from auto_deploy.runner.caffe_runner import CaffeRunner as RN
  elif platform == 'pytorch':
    from auto_deploy.runner.pytorch_runner import PytorchRunner as RN
  else:
    raise NotImplementedError("unsupport platform: {}".format\
            (graph_infos['platform']))
  runner = RN(folder)
  return runner.infer(input_tensors)

def infer_time(folder, input_tensors, save_file, mode):
  """ infer time of splitted models.
  """
  import numpy as np
  with open(os.path.join(folder, 'graph_infos.json'), 'r') as cfg:
    graph_infos = json.load(cfg)
  platform = graph_infos["platform"]
  assert(platform in ["tensorflow", "mxnet", "caffe", "pytorch"])
  if platform == 'tensorflow':
    from auto_deploy.runner.tensorflow_runner import TensorflowRunner as RN
  elif platform == 'mxnet':
    from auto_deploy.runner.mxnet_runner import MxnetRunner as RN
  elif platform == 'caffe':
    from auto_deploy.runner.caffe_runner import CaffeRunner as RN
  elif platform == 'pytorch':
    from auto_deploy.runner.pytorch_runner import PytorchRunner as RN
  runner = RN(folder, mode)
  for i in range(10):
    ts = runner.infer_time(input_tensors)
  time_array = np.zeros(len(ts))
  for i in range(1000):
    ts = runner.infer_time(input_tensors)
    time_array += np.array(ts)
  print("time: " + str(time_array))
  with open(save_file, 'w') as f:
    f.write(str(time_array))
