import os
import sys
import argparse
import json
import mxnet as mx
from mxnet import gluon
import numpy as np
sys.path.append("..")
from auto_deploy.api import split, compile, infer

def get_cls_image():
  """ Get an image for classification models.

  Args:
    None.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  url = 'https://github.com/dmlc/web-data/blob/master/mxnet/' + \
        'doc/tutorials/onnx/images/dog.jpg?raw=true'
  filename = mx.test_utils.download(url, fname='dog.jpg')
  image = mx.image.imread('dog.jpg')
  resized = mx.image.resize_short(image, 224) #minimum 224x224 images
  cropped, crop_info = mx.image.center_crop(resized, (224, 224))
  normalized = mx.image.color_normalize(cropped.astype(np.float32)/255,
                                    mean=mx.nd.array([0.485, 0.456, 0.406]),
                                    std=mx.nd.array([0.229, 0.224, 0.225]))
  # the network expect batches of the form (N,3,224,224)
  # Transposing from (224, 224, 3) to (3, 224, 224)
  transposed = normalized.transpose((2,0,1))
  # change the shape from (3, 224, 224) to (1, 3, 224, 224)
  batchified = transposed.expand_dims(axis=0)
  return batchified.asnumpy()

def get_ssd_image(size):
  """ Get an image for ssd models.

  Args:
    size: Image height an width.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  from gluoncv import model_zoo, data, utils
  im_fname = utils.download('https://github.com/dmlc/web-data/blob/master/' +
                            'gluoncv/detection/street_small.jpg?raw=true',
                            path='street_small.jpg')
  x, img = data.transforms.presets.ssd.load_test(im_fname, short=size)
  return x.asnumpy()

def get_rcnn_image():
  """ Get an image for rcnn models.

  Args:
    None.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  import gluoncv
  from gluoncv import model_zoo, data, utils
  im_fname = utils.download('https://github.com/dmlc/web-data/blob/master/' +
                            'gluoncv/detection/biking.jpg?raw=true',
                            path='biking.jpg')
  x, orig_img = data.transforms.presets.rcnn.load_test(im_fname)
  return x.asnumpy()

def get_yolo_image():
  """ Get an image for rcnn models.

  Args:
    None.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  import gluoncv
  from gluoncv import model_zoo, data, utils
  im_fname = utils.download('https://raw.githubusercontent.com/zhreshold/' +
                            'mxnet-ssd/master/data/demo/dog.jpg',
                            path='dog.jpg')
  x, img = data.transforms.presets.yolo.load_test(im_fname, short=512)
  return x.asnumpy()

def get_fcn_image():
  """ Get an image for fcn models.

  Args:
    None.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  from mxnet import image
  from mxnet.gluon.data.vision import transforms
  import gluoncv
  ctx = mx.cpu()
  url = 'https://raw.githubusercontent.com/dmlc/web-data/master/gluoncv/' + \
        'segmentation/voc_examples/1.jpg'
  filename = 'example.jpg'
  gluoncv.utils.download(url, filename)
  img = image.imread(filename)
  transform_fn = transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize([.485, .456, .406], [.229, .224, .225])
  ])
  img = transform_fn(img)
  img = img.expand_dims(0).as_in_context(ctx)
  return img.asnumpy()

def get_psp_image():
  """ Get an image for psp models.

  Args:
    None.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  from mxnet import image
  from mxnet.gluon.data.vision import transforms
  import gluoncv
  ctx = mx.cpu()
  url = 'https://github.com/zhanghang1989/image-data/blob/master/encoding/' + \
        'segmentation/ade20k/ADE_val_00001142.jpg?raw=true'
  filename = 'ade20k_example.jpg'
  gluoncv.utils.download(url, filename, True)
  img = image.imread(filename)
  img = image.imresize(img, 480, 480)
  transform_fn = transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize([.485, .456, .406], [.229, .224, .225])
  ])
  img = transform_fn(img)
  img = img.expand_dims(0).as_in_context(ctx)
  return img.asnumpy()

def get_deeplab_image():
  """ Get an image for deeplab models.

  Args:
    None.

  Returns:
    A preprocessed image with type of numpy.ndarray.
  """
  from mxnet import image
  from mxnet.gluon.data.vision import transforms
  import gluoncv
  ctx = mx.cpu()
  url = 'https://github.com/zhanghang1989/image-data/blob/master/encoding/' + \
        'segmentation/ade20k/ADE_val_00001755.jpg?raw=true'
  filename = 'ade20k_example.jpg'
  gluoncv.utils.download(url, filename, True)
  img = image.imread(filename)
  img = image.imresize(img, 480, 480)
  transform_fn = transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize([.485, .456, .406], [.229, .224, .225])
  ])
  img = transform_fn(img)
  img = img.expand_dims(0).as_in_context(ctx)
  return img.asnumpy()

def get_pose_image():
  """ Get input data for pose models.

  Args:
    None.

  Returns:
    Data with type of numpy.ndarray.
  """
  from gluoncv import model_zoo, data, utils
  from gluoncv.data.transforms.pose import detector_to_simple_pose
  detector = model_zoo.get_model('yolo3_mobilenet1.0_coco', pretrained=True)
  im_fname = utils.download('https://github.com/dmlc/web-data/blob/master/' +
                            'gluoncv/pose/soccer.png?raw=true',
                            path='soccer.png')
  x, img = data.transforms.presets.ssd.load_test(im_fname, short=512)
  class_IDs, scores, bounding_boxs = detector(x)
  pose_input, upscale_bbox = detector_to_simple_pose( \
         img, class_IDs, scores, bounding_boxs)
  return pose_input.asnumpy()

def compare_data(x, y, eps):
  """ Compare two numpy.ndarray.

  Args:
    x: numpy.ndarray.
    y: numpy.ndarray.
    eps: Threshold.

  Return:
    True for equal and False for not.
  """
  a = x.flatten()
  b = y.flatten()
  if a.shape != b.shape:
    print("shape not match")
    print(x.shape, " vs ", y.shape)
    return False
  for i in range(len(a)):
    if abs(a[i]) > 1.0:
      if abs((a[i] - b[i]) / a[i]) > eps:
        print("{} vs {}".format(a[i], b[i]))
        return False
    elif abs(a[i] - b[i]) > eps:
      print("{} vs {}".format(a[i], b[i]))
      return False
  return True

def compare_outputs(raw_out, new_out, output_names):
  """ Compare output tensors between the raw graph and subgraphs.

  Args:
    raw_out: Outputs of the raw graph. Format: list of mxnet.ndarray
    new_out: Outputs of subgraphs. Format: {output_name: numpy.ndarray}
    output_names: A list contains names of final output intensors.

  Return:
    True for equal and False for not.
  """
  eps = 1e-4
  ret = True
  if len(output_names) == 1:
    raw_data = raw_out.asnumpy()
    new_data = new_out[output_names[0]]
    ret = compare_data(raw_data, new_data, eps)
    if not ret:
      print("Compare failed for tenssor: ",output_names[0])
  else:
    for i in range(len(output_names)):
      raw_data = raw_out[i].asnumpy()
      new_data = new_out[output_names[i]]
      ret = compare_data(raw_data, new_data, eps)
      if not ret:
        print("Compare failed for tenssor: ",output_names[i])
  return ret

def infer_and_compare(save_dir, json_path, \
                      params_path, input_tensors):
  """ Inference by running the submodels as a pipline.

  Args:
    save_dir: Directory saves model informations.
    json_path: Path of json file of raw mxnet model.
    params_path: Path of params file of raw mxnet model.
    input_tensors: A dict contains the information of input tensors.
                  Format: {input_name: numpy.ndarray}

  Return:
    None.
  """
  mxnet_input = mx.nd.array(input_tensors["data"])
  net = gluon.SymbolBlock.imports(json_path, \
                                  list(input_tensors.keys()), \
                                  params_path, \
                                  ctx=mx.cpu())
  predictions = net(mxnet_input)
  outputs = infer(save_dir, input_tensors)
  sorted_output_names = sorted(list(outputs.keys()))
  ret = compare_outputs(predictions, outputs, sorted_output_names)
  if ret:
    print("Comparision Passed!")

def main():
  """Program entry point.
  """
  if os.path.exists(FLAGS.save_dir) and os.path.isfile(FLAGS.save_dir):
    print("{} is a file, please input a directory path.")
    exit(-1)
  if not os.path.exists(FLAGS.save_dir):
    os.mkdir(FLAGS.save_dir)

  save_dir = os.path.abspath(FLAGS.save_dir)
  if FLAGS.model_type == "cls":
    image = get_cls_image()
  elif FLAGS.model_type == "ssd":
    image = get_ssd_image(FLAGS.size)
  elif FLAGS.model_type == "rcnn":
    image = get_rcnn_image()
  elif FLAGS.model_type == "yolo":
    image = get_yolo_image()
  elif FLAGS.model_type == "fcn":
    image = get_fcn_image()
  elif FLAGS.model_type == "pose":
    image = get_pose_image()
  elif FLAGS.model_type == "psp":
    image = get_psp_image()
  elif FLAGS.model_type == "deeplab":
    image = get_deeplab_image()
  else:
    print("Wrong model_type: ", model_type)
    exit(-1)
  input_tensors = {'data': image}
#  input_shapes = {'data': image.shape}
  if FLAGS.mode & 1 == 1:
    split("mxnet", input_tensors, save_dir, FLAGS.json_path, params_path=FLAGS.params_path)
    print("Splitting done!\n")

  if FLAGS.mode & 2 == 2:
    compile(save_dir)
    print("Compilation done!\n")

  input_tensors = {'data': image}
  if FLAGS.mode & 4 == 4:
    infer_and_compare(save_dir, FLAGS.json_path, \
                      FLAGS.params_path, input_tensors)
    print("Inference done!")

if __name__ == "__main__":
  PARSER = argparse.ArgumentParser()
  PARSER.add_argument(
      '--save_dir',
      type=str,
      default=os.path.join(os.getcwd(), "mxnet_split_info"),
      help='Absolute path of a directory to save splitting information.'
  )
  PARSER.add_argument(
      '--json_path',
      type=str,
      required=True,
      help='Path to symbol file.'
  )
  PARSER.add_argument(
      '--params_path',
      type=str,
      required=True,
      help='Path to parameter file.'
  )
  PARSER.add_argument(
      '--model_type',
      type=str,
      required=True,
      help='Options: cls, ssd, rcnn, yolo, fcn, pose, psp, deeplab'
  )
  PARSER.add_argument(
      '--size',
      type=int,
      required=True,
      help='Size of image height and width.'
  )
  PARSER.add_argument(
      '--mode',
      type=int,
      default=0,
      help='Options: 1:S,2:C,3:SC,4:R,5:SR,6:CR,7:SCR(S:split,C:compile,R:run)'
  )
  FLAGS, UNPARSED = PARSER.parse_known_args()
  main()
