#include <ctime>
#include "util.hpp"

using namespace std;

void Preprocess(float* input_ptr, cv::Mat image, int width, int height) {
    image.convertTo(image, CV_32FC3);
    vector<cv::Mat> input_channels;
    int size_per_channel = height * width;
    for (int i = 0; i < 3; i++) {
        cv::Mat channel(height, width, CV_32FC1, input_ptr);
        input_channels.push_back(channel);
        input_ptr += size_per_channel;
    }
    cv::resize(image, image, cv::Size(width, height), 0, 0, cv::INTER_LINEAR);
    float mean[3] = {-103.94, -116.78, -123.68};
    image += cv::Scalar(mean[0], mean[1], mean[2]);
    cv::split(image, input_channels);
}

vector<int> Postprocess(float* output, int classes, int top_k) {
    // save idx
    vector<int> idx(classes);
    iota(idx.begin(), idx.end(), 0);
    sort(idx.begin(), idx.end(),
         [&output](int idx1, int idx2) {
             return output[idx1] > output[idx2];
             });
    idx.resize(top_k);
    return idx;
}

int main(int argc, char** argv) {
    string context_dir;
    string image_path;
    if (argc == 1) {
        cout << "Usage ./main context_path input_file_path"<< endl;
        exit(-1);
    }
    if (argc > 2) {
        context_dir = argv[1];
        image_path = argv[2];
    }
    int dev_count = 0, ret;
    ret = bmrt_dev_getcount(NULL, &dev_count);
    if (0 != ret) {
        cout << "can not get device count on the host" << endl;
        return -1;
    }
    // init info from context_dir
    string CHIPNAME = "BM1682_PCIE";
    struct bmrt_info g_bmrtinfo;
    void* p_bmrt = create_bmruntime_per_device(NULL, CHIPNAME.c_str(), 0);
    g_bmrtinfo.p_bmrt = p_bmrt;
    bmrt_load_context(p_bmrt, context_dir.c_str());
    // fill bmrt net info
    bmrt_init_net_info(p_bmrt, g_bmrtinfo);
    bmrt_print_net_info(g_bmrtinfo);
    bmrt_alloc_memory(g_bmrtinfo);
    // run network
    for (int netidx = 0; netidx < g_bmrtinfo.net_num; netidx++) {
        // network: inceptionv1 net_num == 1
        bmrt_net_info* pnetinfo = &g_bmrtinfo.netinfo[0];
        bm_device_mem_t* input_mem = NULL;
        bm_device_mem_t* output_mem = NULL;
        cv::Mat frame;
        frame = cv::imread(image_path);
        double preprocess_cost = 0.0,
               inference_cost = 0.0,
               postprocess_cost = 0.0;
        for (int i = 0; i < 1; i++) {
            input_mem = pnetinfo->input_mem_ping;
            output_mem = pnetinfo->output_mem_ping;
            clock_t start_time = clock();
            int net_h = pnetinfo->input_shapes[2];
            int net_w = pnetinfo->input_shapes[3];
            Preprocess(pnetinfo->input_tensor_sys[0], frame, net_w, net_h);
            clock_t after_preprocess = clock();
            // copy input data to device
            bmrt_memcpy_s2d(g_bmrtinfo.p_bmrt, input_mem[0],
                    reinterpret_cast<void *>(pnetinfo->input_tensor_sys[0]));
            // launch network do inference
            bmrt_launch_shape(g_bmrtinfo.p_bmrt, netidx, NULL,
                              (const void*)input_mem,
                              pnetinfo->itensor_num,
                              pnetinfo->input_dims,
                              pnetinfo->input_shapes,
                              (const void*)output_mem,
                              pnetinfo->otensor_num);
            // copy output data back to system
            bmrt_memcpy_d2s(g_bmrtinfo.p_bmrt,
                    reinterpret_cast<void *>(pnetinfo->output_tensor_sys[0]),
                    output_mem[0]);
            clock_t after_inference = clock();
            vector<int> result = Postprocess(pnetinfo->output_tensor_sys[0],
                                             1000, 5);
            cout << "top" << 5 << endl;
            for (int i = 0; i < 5; i++) {
                cout << result[i] << " ";
            }
            cout << endl;
            clock_t after_postprocess = clock();
            preprocess_cost += static_cast<double>
                (after_preprocess-start_time)/CLOCKS_PER_SEC;
            // inference_cost +=
            //     (double)(after_inference-after_preprocess)/CLOCKS_PER_SEC;
            postprocess_cost += static_cast<double>
                (after_postprocess-after_inference)/CLOCKS_PER_SEC;
        }
        // cout << "Total frame number: " << frame_num << endl;
        // cout << " preprocess cost avg: " << 1000*preprocess_cost/frame_num
        // << " ms" << endl;
        // cout << " inference cost avg: " << 1000*inference_cost/frame_num
        // << " ms" << endl;
        // cout << " postprocess cost avg: " << 1000*postprocess_cost/frame_num
        //  << " ms" << endl;
        pnetinfo++;
    }
    // free memory and info
    bmrt_free_memory(g_bmrtinfo);
    bmrt_delete_net_info(g_bmrtinfo);
    destroy_bmruntime(p_bmrt);
    return 0;
}
