/*
 * Copyright (c) 2012 Stefano Sabatini
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/**
 * @file
 * Demuxing and decoding example.
 *
 * Show how to use the libavformat and libavcodec API to demux and
 * decode audio and video data.
 * @example demuxing_decoding.c
 */

extern "C" {
#include <libavutil/imgutils.h>
#include <libavutil/samplefmt.h>
#include <libavutil/timestamp.h>
#include <libavutil/frame.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>
#include <libavcodec/avcodec.h>
}
#include <opencv2/opencv.hpp>
#include <bmdnn_api.h>
#include <bmcv_api.h>
#include <bmruntime_interface.h>
// #include <bmblob.h>
#include <bmlib_runtime.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <errno.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <pthread.h>
//#include <thread.h>

#include <iostream>
#include <fstream>

using namespace std;

int pcie_board_id = 0;

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"

#define MAX_INSTANCE_NUM 8
#define MAX_THREAD MAX_INSTANCE_NUM
typedef struct thread_context_t {
    AVFormatContext *fmt_ctx;
    AVCodecContext *video_dec_ctx;
    enum AVPixelFormat pix_fmt;
    AVFrame *out_frame = NULL;
    AVFrame *frame;
    AVPacket pkt;
    int video_stream_idx;
    int width;
    int height;
    AVStream *video_stream;
    char *src_filename;
    bm_handle_t handle = NULL;
    int totalframes;
} thread_context;

/* Enable or disable frame reference counting. You are not supposed to support
 * both paths in your application but pick the one most appropriate to your
 * needs. Look for the use of refcount in this example to see what are the
 * differences of API usage between them. */
static int refcount = 0;
static AVFrame *alloc_picture(enum AVPixelFormat pix_fmt,
                              int width,
                              int height) {
    AVFrame *picture;
    int ret;

    picture = av_frame_alloc();
    if (!picture)
        return NULL;

    picture->format = pix_fmt;
    picture->width = width;
    picture->height = height;

    /* allocate the buffers for the frame data */
    ret = av_frame_get_buffer(picture, 4);
    if (ret < 0) {
        fprintf(stderr, "Could not allocate frame data.\n");
        exit(1);
    }

    return picture;
}

static int decode_packet(int *got_frame,
                         int cached,
                         thread_context *t_context) {
    int ret = 0;
    int decoded = t_context->pkt.size;
    AVFrame *frame = t_context->frame;
    *got_frame = 0;
    if (t_context->pkt.stream_index == t_context->video_stream_idx) {
        /* decode video frame */
        int response = avcodec_send_packet(t_context->video_dec_ctx,
                                           &t_context->pkt);
        if (response < 0) {
            return response;
        }
        response = avcodec_receive_frame(t_context->video_dec_ctx, frame);
        if (response == AVERROR(EAGAIN) || response == AVERROR_EOF) {
            sleep(1);
            return response;
        } else if (response < 0) {
            return response;
        }
        if (response >= 0) {
            if (frame->width != t_context->width ||
                frame->height != t_context->height ||
                frame->format != t_context->pix_fmt) {
                /* To handle this change, one could call 
                av_image_alloc again and decode the following
                frames into another rawvideo file. */
                fprintf(stderr, "Error: Width, height and pixel format "
                                " have to be constant in a rawvideo file,"
                                "but the width, height or pixel format of"
                                " the input video changed:\n"
                                "old: width = %d, height = %d, format = %s\n"
                                "new: width = %d, height = %d, format = %s\n",
                        t_context->width,
                        t_context->height,
                        av_get_pix_fmt_name(t_context->pix_fmt),
                        t_context->frame->width,
                        t_context->frame->height,
                        av_get_pix_fmt_name(
                            (AVPixelFormat)t_context->frame->format));
                return -1;
            }

            t_context->totalframes++;
            if ((t_context->totalframes % 10) == 0) {
                printf("decoded %d frames!\n",
                       t_context->totalframes);
            }

            if (frame->format == AV_PIX_FMT_NV12 ||
                frame->format == AV_PIX_FMT_YUV420P ||
                frame->format == AV_PIX_FMT_YUVJ420P) {
                int y_linesize = frame->width;
                int uv_linesize = frame->width;
                int yuv_device_mem_malloc = 0;
                bm_device_mem_t dmem_y;
                bm_device_mem_t dmem_uv;
                bm_device_mem_t dmem_bgr;
                bm_malloc_device_byte(t_context->handle,
                    &dmem_bgr,
                    (frame->width * frame->height * 3 + 3) / 4 * 4);
                if (frame->format == AV_PIX_FMT_YUV420P ||
                    frame->format == AV_PIX_FMT_YUVJ420P) {
                    // when decode mp4 on pc with software decode,
                    // the output would be YUV420p
                    // we malloc device memory and convert yuv420p to nv12
                    // and copy to device memory for further processing
                    yuv_device_mem_malloc = 1;
                    bm_malloc_device_byte(t_context->handle,
                        &dmem_y,
                        (frame->width * frame->height + 3) / 4 * 4);
                    bm_malloc_device_byte(t_context->handle,
                        &dmem_uv,
                        (frame->width * frame->height / 2 + 3) / 4 * 4);
                    unsigned char *y_temp =
                      (unsigned char *)malloc(frame->width * frame->height);
                    unsigned char *uv_temp =
                      (unsigned char *)
                      malloc(frame->width * frame->height / 2);
                    unsigned char *u_temp = uv_temp;
                    unsigned char *v_temp = uv_temp + 1;
                    unsigned char *uv_t = uv_temp;
                    for (int i = 0; i < frame->height; i++) {
                        memcpy(y_temp + frame->width * i,
                               frame->data[0] + i * frame->linesize[0],
                               frame->width);
                    }
                    for (int i = 0; i < frame->height / 2; i++) {
                        u_temp = frame->data[1] + i * frame->linesize[1];
                        v_temp = frame->data[2] + i * frame->linesize[2];
                        for (int j = 0; j < frame->width / 2; j++) {
                            *uv_t++ = *u_temp++;
                            *uv_t++ = *v_temp++;
                        }
                    }
                    bm_memcpy_s2d(t_context->handle,
                        dmem_y,
                        bm_mem_from_system(y_temp));
                    bm_memcpy_s2d(t_context->handle,
                        dmem_uv,
                        bm_mem_from_system(uv_temp));
                    free(y_temp);
                    free(uv_temp);
                    y_linesize = frame->width;
                    uv_linesize = frame->width;
                } else {
                    y_linesize = frame->linesize[4];
                    uv_linesize = frame->linesize[5];
                    dmem_y = bm_mem_from_device(
                        (unsigned long long)frame->data[4],
                        y_linesize * frame->height);
                    dmem_uv = bm_mem_from_device((
                        unsigned long long)frame->data[5],
                        uv_linesize * frame->height);
                }

                bmcv_image input;
                input.color_space = COLOR_YUV;
                input.data_format = DATA_TYPE_BYTE;
                input.image_format = NV12;
                input.data[0] = dmem_y;
                input.stride[0] = y_linesize;
                input.data[1] = dmem_uv;
                input.stride[1] = uv_linesize;
                input.image_width = frame->width;
                input.image_height = frame->height;

                bmcv_image output = input;
                output.color_space = COLOR_RGB;
                output.image_format = BGR;
                output.data[0] = dmem_bgr;
                output.stride[0] = frame->width;

                bm_img_yuv2bgr(t_context->handle, input, output);

                cv::Mat detect_frame;
                detect_frame.create(frame->height, frame->width, CV_8UC3);
                cv::Mat detect_frame1;
                // may be aligned dmemory is larger than mat reserved?
                bm_memcpy_d2s(t_context->handle,
                              bm_mem_from_system(detect_frame.data),
                              dmem_bgr);
                {
                    vector<cv::Mat> channels;
                    cv::Mat b = cv::Mat(frame->height,
                                        frame->width,
                                        CV_8U,
                                        detect_frame.data);
                    cv::Mat g = cv::Mat(frame->height,
                       frame->width,
                       CV_8U,
                       detect_frame.data + frame->width * frame->height);
                    cv::Mat r = cv::Mat(frame->height,
                       frame->width,
                       CV_8U,
                       detect_frame.data + 2 * frame->width * frame->height);

                    channels.push_back(b);
                    channels.push_back(g);
                    channels.push_back(r);
                    /// Merge the three channels
                    cv::merge(channels, detect_frame1);
                }
                if (yuv_device_mem_malloc) {
                    bm_free_device(t_context->handle, dmem_y);
                    bm_free_device(t_context->handle, dmem_uv);
                }
                bm_free_device(t_context->handle, dmem_bgr);
            }
        }
    }
    /* If we use frame reference counting, we own the data and need
     * to de-reference it when we don't use it anymore */
    if (*got_frame && refcount)
        av_frame_unref(t_context->frame);
    return decoded;
}

static int open_codec_context(int *stream_idx,
                              AVCodecContext **dec_ctx,
                              AVFormatContext *fmt_ctx,
                              enum AVMediaType type,
                              thread_context *t_context) {
    int ret, stream_index;
    AVStream *st;
    AVCodec *dec = NULL;
    AVDictionary *opts = NULL;

    ret = av_find_best_stream(fmt_ctx, type, -1, -1, NULL, 0);
    if (ret < 0) {
        fprintf(stderr, "Could not find %s stream in input file '%s'\n",
                av_get_media_type_string(type), t_context->src_filename);
        return ret;
    } else {
        stream_index = ret;
        st = fmt_ctx->streams[stream_index];

        /* find decoder for the stream */
        dec = avcodec_find_decoder(st->codecpar->codec_id);
        if (!dec) {
            fprintf(stderr, "Failed to find %s codec\n",
                    av_get_media_type_string(type));
            return AVERROR(EINVAL);
        }

        /* Allocate a codec context for the decoder */
        *dec_ctx = avcodec_alloc_context3(dec);
        if (!*dec_ctx) {
            fprintf(stderr, "Failed to allocate the %s codec context\n",
                    av_get_media_type_string(type));
            return AVERROR(ENOMEM);
        }

        /* Copy codec parameters from input stream to output codec context */
        if ((ret = avcodec_parameters_to_context(*dec_ctx, st->codecpar)) < 0) {
            fprintf(stderr,
                    "Failed to copy %s codec parameters to decoder context\n",
                    av_get_media_type_string(type));
            return ret;
        }

        /* Init the decoders, with or without reference counting */
        av_dict_set(&opts, "refcounted_frames", refcount ? "1" : "0", 0);
        av_dict_set_int(&opts, "pcie_board_id", pcie_board_id, 0x0);
        if ((ret = avcodec_open2(*dec_ctx, dec, &opts)) < 0) {
            fprintf(stderr, "Failed to open %s codec\n",
                    av_get_media_type_string(type));
            return ret;
        }
        *stream_idx = stream_index;
    }

    return 0;
}

static void (*av_log_callback)(void *, int, const char *, va_list) = NULL;

static pthread_t decoding_threads[MAX_INSTANCE_NUM];

void *decoding_thread(void *msg) {
    int ret = 0, got_frame = 0;
    unsigned int thid = (unsigned int)pthread_self();
    thread_context *t_context = reinterpret_cast<thread_context *>(msg);
    void *pp_bmrt = create_bmrt_helper(NULL, "BM1682_PCIE", pcie_board_id);
    t_context->handle = (bm_handle_t)bmrt_get_bm_handle(pp_bmrt);
    int seeking_type = AVSEEK_FLAG_FRAME;

    av_register_all();
    av_log_set_callback(av_log_callback);

    fprintf(stderr, "now to open source file:%s\n", t_context->src_filename);
    /* open input file, and allocate format context */
    if (avformat_open_input(&t_context->fmt_ctx,
                            t_context->src_filename,
                            NULL,
                            NULL) < 0) {
        fprintf(stderr, "Could not open source file %s\n",
                t_context->src_filename);
        return 0;
    }

    /* retrieve stream information */
    if (avformat_find_stream_info(t_context->fmt_ctx, NULL) < 0) {
        fprintf(stderr, "Could not find stream information\n");
        return 0;
    }

    if (open_codec_context(&t_context->video_stream_idx,
                           &t_context->video_dec_ctx,
                           t_context->fmt_ctx,
                           AVMEDIA_TYPE_VIDEO,
                           t_context) >= 0) {
        t_context->video_stream =
            t_context->fmt_ctx->streams[t_context->video_stream_idx];
        /* allocate image where the decoded image will be put */
        t_context->width = t_context->video_dec_ctx->width;
        t_context->height = t_context->video_dec_ctx->height;
        t_context->pix_fmt = t_context->video_dec_ctx->pix_fmt;
        t_context->out_frame = alloc_picture(AV_PIX_FMT_BGRA, 1280, 720);
    }

    /* dump input information to stderr */
    av_dump_format(t_context->fmt_ctx, 0, t_context->src_filename, 0);

    if (t_context->video_dec_ctx->codec_id == AV_CODEC_ID_HEVC) {
        seeking_type = AVSEEK_FLAG_BYTE;
    }

    fprintf(stderr, "decoding size %dx%d fmt %d\n",
                    t_context->video_dec_ctx->width,
                    t_context->video_dec_ctx->height,
                    t_context->video_dec_ctx->pix_fmt);

    do {
        if (!t_context->video_stream) {
            fprintf(stderr,
              "Could not find audio or video stream in the input,"
              " aborting\n");
            ret = 1;
            break;
        }

        t_context->frame = av_frame_alloc();
        if (!t_context->frame) {
            fprintf(stderr, "Could not allocate frame\n");
            ret = AVERROR(ENOMEM);
            break;
        }

        /* initialize packet, set data to NULL, let the demuxer fill it */
        av_init_packet(&t_context->pkt);
        t_context->pkt.data = NULL;
        t_context->pkt.size = 0;

        /* read frames from the file */
        int read = 0;
        while ((read = av_read_frame(t_context->fmt_ctx,
                                     &t_context->pkt)) >= 0) {
            do {
                ret = decode_packet(&got_frame, 0, t_context);
                if (ret < 0) {
                    break;
                }
                t_context->pkt.data += ret;
                t_context->pkt.size -= ret;
            } while (t_context->pkt.size > 0);
            av_packet_unref(&t_context->pkt);
            printf("thread ID:%x frame:%d \n", thid, t_context->totalframes);
        }
        /* flush cached frames */
        t_context->pkt.data = NULL;
        t_context->pkt.size = 0;
        do {
            decode_packet(&got_frame, 1, t_context);
        } while (got_frame);
        av_seek_frame(t_context->fmt_ctx, -1, 1, seeking_type);
        av_frame_free(&t_context->frame);
    } while (0);

    destroy_bmruntime(pp_bmrt);
    avcodec_free_context(&t_context->video_dec_ctx);
    avformat_close_input(&t_context->fmt_ctx);
    return 0;
}

int main(int argc, char **argv) {
    int ret = 0;
    char *src_filename = NULL;
    int trd_num = 0;
    int oc;
    char *b_opt_arg;
    while ((oc = getopt(argc, argv, "v:t:i:")) != -1) {
        switch (oc) {
        case 'v':
            src_filename = optarg;
            break;
        case 't':
            trd_num = atoi(optarg);
            printf("threads number is %d\n", trd_num);
            if (trd_num > MAX_THREAD) {
                printf("max threads num is %d\n", MAX_THREAD);
                exit(-1);
            }
            break;
        case 'i':
            pcie_board_id = atoi(optarg);
            break;
        case '?':
            printf("usage: vpu_decode "
                   "-v file.mp4 "
                   "-t 1 "
                   "-i [[pcie_board_id] -d [debug]]\n");
            printf("arguments error!\n");
            break;
        }
    }
    for (int i = 0; i < trd_num; i++) {
        thread_context *t_context = new thread_context;
        memset(t_context, 0, sizeof(thread_context));
        t_context->video_stream_idx = -1;
        t_context->src_filename = src_filename;
        t_context->fmt_ctx = NULL;
        if (0 != pthread_create(&decoding_threads[i],
                                NULL,
                                decoding_thread,
                                reinterpret_cast<void *>(t_context))) {
            fprintf(stderr, "wrong in create thread, exit!\n");
            exit(0);
        }
        sleep(1);
    }
    for (int i = 0; i < trd_num; i++) {
        pthread_join(decoding_threads[i], NULL);
    }
    return ret < 0;
}
#pragma GCC diagnostic pop
