/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <getopt.h>
#include <iomanip>
#include "process/preprocess.hpp"
#include "process/postprocess.hpp"
#include "core/inference.hpp"
#include "core/statistics.hpp"
#include "core/utils.hpp"
using std::vector;
using std::string;
using std::set;
using std::cout;
using std::endl;
using std::ios;
using std::setw;
using std::setprecision;

void print_result(
    float*                   output_tensor,
    vector<vector<size_t>>&  class_ids,
    int                      class_num) {
  for (int i = 0; i < class_ids.size(); ++i) {
    for (int j = 0; j < class_ids[0].size(); ++j) {
      cout.flags(ios::right);
      cout << setw(3) << class_ids[i][j] << "(";
      cout.setf(ios::fixed);
      cout << setprecision(4) << output_tensor[class_num * i + class_ids[i][j]]
           << ") \t";
    }
    cout << endl;
  }
}

void cls_mobilenetv1(
      const string& context_dir,
      const string& input_path,
      const string& tpu) {
  sail::Inference<float> inference(context_dir, "BM1682_PCIE", tpu,
                                   sail::SYSIO, false, false);
  string graph_name = *inference.get_all_graph_names().begin();
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  set<string> output_names = inference.get_output_tensor_names(graph_name);
  // only one input tensor in the graph
  string input_tensor_name = *(input_names.begin());
  string output_tensor_name = *(output_names.begin());
  vector<int> input_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
  vector<int> output_shape = inference.get_output_shape(graph_name,
                                                        output_tensor_name);
  float* input = inference.get_input_tensor_sys(graph_name, input_tensor_name);

  sail::ImageNetPreprocess preprocessor(
      input_shape[2], input_shape[3],
      cv::INTER_LINEAR,
      104.00698853, 116.66876984, 112.67891693,
      0.017,
      false);
  sail::Statistics timer;
  int top_k = 5;
  int class_num = output_shape[1];
  sail::ImageNetPostprocess postprocessor(output_shape[0], class_num, top_k);
  cout << "top_1\t\ttop_2\t\ttop_3\t\ttop_4\t\ttop_5" << endl;
  timer.record_start("preprocess");
  cv::Mat frame = cv::imread(input_path);
  preprocessor.process_cpu(input, frame);
  timer.record_end("preprocess");
  timer.record_start("inference");
  inference.process(graph_name);
  timer.record_end("inference");
  timer.record_start("postprocess");
  float* output_tensor = inference.get_output_tensor_sys(
      graph_name, output_tensor_name);
  vector<vector<size_t> > result = postprocessor.process(output_tensor);
  timer.record_end("postprocess");
  print_result(output_tensor, result, class_num);
  // print the time
  timer.print_time();
}

void print_usage() {
  cout << "usage: cls_mobilenetv1 <args> " << endl
    <<  "  --context_dir:" << endl
    <<  "    Required; The dir of context after compilation." << endl
    <<  "  --input_path:" << endl
    <<  "    Required; Path to input image." << endl
    <<  "  --tpu:" << endl
    <<  "    Optional; Specify the TPU IDs to use, separated by ','." << endl
    <<  "    Default: '-tpu all' to run on all available TPUs." << endl;
}

int main(int argc, char** argv) {
  const char* optstring = "c:i:th";
  int c = 0;
  int index = 0;
  string context_dir;
  string input_path;
  string tpu("all");
  struct option opts[] = {{"context_dir", required_argument, 0, 0},
                          {"input_path", required_argument, 0, 0},
                          {"tpu", optional_argument, 0, 0},
                          {"help", no_argument, 0, 0}};
  while ((c = getopt_long_only(argc, argv, optstring, opts, &index)) != -1) {
    switch (index) {
      case 0:
        context_dir = optarg;
        break;
      case 1:
        input_path = optarg;
        break;
      case 2:
        tpu = optarg;
        break;
      case 3:
        print_usage();
        return 0;
      default:
        break;
    }
  }
  if (context_dir.empty() || input_path.empty()) {
    print_usage();
  } else {
    cls_mobilenetv1(context_dir, input_path, tpu);
  }
  return 0;
}
