/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <getopt.h>
#include "process/preprocess.hpp"
#include "process/postprocess.hpp"
#include "core/bm_cv.hpp"
#include "core/inference.hpp"
#include "core/statistics.hpp"
#include "core/utils.hpp"
using std::string;
using std::vector;
using std::set;
using std::cout;
using std::endl;

void det_mobilessd_vpu(
    const string& context_dir,
    const string& input_path,
    const string& tpu) {
  sail::Inference<float> inference(context_dir, "BM1682_PCIE", tpu,
                                   sail::SYSO, false, false);
  string graph_name = *inference.get_all_graph_names().begin();
  // only one input tensor in the graph
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  string input_tensor_name = *(input_names.begin());

  // for mobilenet_ssd
  sail::Bmcv bmcv(inference.get_bmrt());
  sail::ImageNetPreprocess preprocessor(&bmcv, 300, 300, cv::INTER_NEAREST,
                                        127.5, 127.5, 127.5, 0.007843, false);
  sail::SsdPostprocess postprocessor(21, 1917, 0.5, 0.5);
  sail::Statistics timer;
  sail::Frame frame;
  sail::VideoDecoder video_decoder(input_path);
  if (!video_decoder.is_opened()) {
     cout << "Error: can not open the video." << endl;
     return;
  }
  int frame_id = 0;
  while (!video_decoder.read(frame)) {
    frame_id++;
    sail::Tensor<float> input =
        inference.get_input_tensor_dev(graph_name, input_tensor_name);
    timer.record_start("preprocess");
    preprocessor.process_tpu(input, frame);
    timer.record_end("preprocess");
    timer.record_start("inference");
    inference.process(graph_name);

    float* conf_data = inference.get_output_tensor_sys(
        graph_name, "mbox_conf_flatten");
    float* loc_data = inference.get_output_tensor_sys(graph_name, "mbox_loc");
    float* prior_data = inference.get_output_tensor_sys(
        graph_name, "mbox_priorbox");
    timer.record_end("inference");
    timer.record_start("postprocess");

    vector<sail::DetectRet> bboxes = postprocessor.process(conf_data,
                                                           loc_data,
                                                           prior_data);
    timer.record_end("postprocess");
    for (int i = 0; i < bboxes.size(); i++) {
        cout <<"frame_id:"<<frame_id<< " bboxes: " << bboxes[i].x_min_ratio << " "
             << bboxes[i].y_min_ratio << " " << bboxes[i].x_max_ratio
             << " " << bboxes[i].y_max_ratio << " ";
        cout << "class_id: " << bboxes[i].class_id << " "
             << "score: " << bboxes[i].score << endl;
    }
  }
  cout<<"frame_id:"<<frame_id<<endl;
  timer.print_time();
}

void print_usage() {
  cout << "usage: det_mobilessd_vpu <args> " << endl
    <<  "  --context_dir:" << endl
    <<  "    Required; The dir of context after compilation." << endl
    <<  "  --input_path:" << endl
    <<  "    Required; The path of the input video." << endl
    <<  "  --tpu:" << endl
    <<  "    Optional; Specified the TPU IDs to use, separated by ','." << endl
    <<  "    Default: '-tpu all' to run on all available TPUs." << endl;
}

int main(int argc, char** argv) {
  const char* optstring = "c:i:th";
  int c = 0;
  int index = 0;
  string context_dir;
  string input_path;
  string tpu("0");
  struct option opts[] = {{"context_dir", required_argument, 0, 0},
                          {"input_path", required_argument, 0, 0},
                          {"tpu", optional_argument, 0, 0},
                          {"help", no_argument, 0, 0}};
  while ((c = getopt_long_only(argc, argv, optstring, opts, &index)) != -1) {
    switch (index) {
      case 0:
        context_dir = optarg;
        break;
      case 1:
        input_path = optarg;
        break;
      case 2:
        tpu = optarg;
        break;
      case 3:
        print_usage();
        return 0;
      default:
        break;
    }
  }
  if (context_dir.empty() || input_path.empty()) {
    print_usage();
    return 0;
  } else {
    det_mobilessd_vpu(context_dir, input_path, tpu);
  }
  return 0;
}
