/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <getopt.h>
#include "process/preprocess.hpp"
#include "process/postprocess.hpp"
#include "core/inference.hpp"
#include "core/statistics.hpp"
#include "core/utils.hpp"
using std::string;
using std::vector;
using std::map;
using std::pair;
using std::set;
using std::cout;
using std::endl;

vector<sail::FaceInfo> run_Pnet(  // PNet deploy
    sail::Inference<float>& inference, sail::MtcnnPreprocess& preprocessor,
    sail::MtcnnPostprocess& postprocessor, vector<double>& scales,
    cv::Mat& frame, const string& graph_name) {
  vector<sail::FaceInfo> total_boxes;
  cv::Mat img;
  frame.copyTo(img);
  int width =  img.cols;
  int height = img.rows;
  set<string> input_names =
              inference.get_input_tensor_names(graph_name);
  string input_tensor_name = *(input_names.begin());
  vector<int> init_shape = inference.get_input_shape(
      graph_name, input_tensor_name);
  // check init PNet input shape
  // cout << "init input shape: "
  //     << init_shape[0] << " " << init_shape[1] << " "
  //     << init_shape[2] << " " << init_shape[3] << endl;
  for (int i = 0 ; i < scales.size(); i++) {
    double scale = scales[i];
    int ws = ceil(width * scale);
    int hs = ceil(height * scale);
    // cout << "height: " << hs << " width: "
    //     << ws << endl;
    vector<int> input_new_shape(4);
    input_new_shape[0] = 1;
    input_new_shape[1] = 3;
    input_new_shape[2] = hs;
    input_new_shape[3] = ws;
    map<string, vector<int>> new_input_info;
    new_input_info.insert(pair<string, vector<int>>
        (input_tensor_name, input_new_shape));
    inference.set_input_shape(graph_name, new_input_info);
    vector<int> now_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
    // check PNet now input shape
    // cout << "new input shape: "
    //     << now_shape[0] << " " << now_shape[1] << " "
    //     << now_shape[2] << " " << now_shape[3] << endl;
    // do PNet preprocess
    float* input = inference.get_input_tensor_sys(graph_name,
                                                  input_tensor_name);
    preprocessor.process_cpu(input, img, hs, ws);
    // do PNet inference
    inference.process(graph_name);
    // get the output_tensor ptr
    float* conf = inference.get_output_tensor_sys(graph_name, "prob1");
    float* coord = inference.get_output_tensor_sys(graph_name, "conv4-2");
    // do postprocess
    vector<int> output_shape = inference.get_output_shape(graph_name, "prob1");
    // check PNet now output shape
    // cout << "new output shape: "
    //     << output_shape[0] << " " << output_shape[1] << " "
    //     << output_shape[2] << " " << output_shape[3] << endl;
    vector<sail::FaceInfo> bboxes_nms =
         postprocessor.Pnet_process(conf, coord, scale, output_shape[3],
                                    output_shape[2], ws, hs);
    total_boxes.insert(total_boxes.end(), bboxes_nms.begin(),
                       bboxes_nms.end());
  }
  int numBox = total_boxes.size();
  cout << "The PNet total numBox is: " << numBox << endl;
  return total_boxes;
}

vector<sail::FaceInfo> run_Rnet(  // RNet deploy
    sail::Inference<float>& inference,
    sail::MtcnnPreprocess&  preprocessor,
    sail::MtcnnPostprocess& postprocessor,
    vector<sail::FaceInfo>& PNet_output,
    cv::Mat&                frame,
    const string&           graph_name) {
  cv::Mat img;
  frame.copyTo(img);
  vector<sail::FaceInfo> regressed_rects;
  vector<sail::FaceInfo> regressed_pading;
  int numBox = PNet_output.size();
  cout << "RNet input bbox num: " << numBox << endl;
  regressed_rects = postprocessor.BoxRegress(PNet_output, 1);
  postprocessor.Bbox2Square(regressed_rects);
  postprocessor.Padding(img, regressed_rects, regressed_pading);
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  string input_tensor_name = *(input_names.begin());
  vector<int> input_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
  // check RNet input/output shape
  // cout << "init input shape: "
  //     << input_shape[0] << " " << input_shape[1] << " "
  //     << input_shape[1] << " " << input_shape[3] << endl;
  vector<int> conf_shape = inference.get_output_shape(graph_name, "prob1");
  // cout << "init conf shape: "
  //     << conf_shape[0] << " " << conf_shape[1] << " "
  //     << conf_shape[2] << " " << conf_shape[3] << endl;
  vector<int> coord_shape = inference.get_output_shape(graph_name, "conv5-2");
  // cout << "init coord shape: "
  //     << coord_shape[0] << " " << coord_shape[1] << " "
  //     << coord_shape[2] << " " << coord_shape[3] << endl;
  // alloc total sys mem (num > do_inference_num)
  int input_size = numBox * input_shape[1] * input_shape[2] * input_shape[3];
  float* input_data = new float[input_size];
  int conf_size = numBox * conf_shape[1];
  float* conf_data = new float[conf_size];
  int coord_size = numBox * coord_shape[1];
  float* coord_data = new float[coord_size];
  // input img joint
  preprocessor.ronet_pre_img(input_data, img, input_shape[2],
                             input_shape[3], regressed_pading, regressed_rects);
  int num_done = 0;
  int num_fix = input_shape[0];  // num_fix == 32
  while (num_done < numBox) {
    if ((num_done + num_fix) > numBox) {
      num_fix = numBox - num_done;
    }
    int num_work = num_fix;
    cout << "*********************************" << endl;
    cout << "inference batch: " << num_work
         << " done batch: " << num_done << endl;
    cout << "*********************************"<< endl;
    vector<int> input_new_shape(4);
    input_new_shape[0] = num_work;
    input_new_shape[1] = input_shape[1];
    input_new_shape[2] = input_shape[2];
    input_new_shape[3] = input_shape[3];
    map<string, vector<int>> new_input_info;
    new_input_info.insert(pair<string, vector<int>>(input_tensor_name,
                                                    input_new_shape));
    inference.set_input_shape(graph_name, new_input_info);
    vector<int> now_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
    // check RNet now input shape
    // cout << "new input shape: "
    //     << now_shape[0] << " " << now_shape[1] << " "
    //     << now_shape[1] << " " << now_shape[3] << endl;
    // do RNet preprocess
    float* input = inference.get_input_tensor_sys(graph_name,
                                                  input_tensor_name);
    preprocessor.ronet_preprocess_cpu(input, input_data,
                                      now_shape, num_done, num_work);
    // do Rnet inference
    inference.process(graph_name);
    // get the output_tensor ptr
    float* conf = inference.get_output_tensor_sys(graph_name, "prob1");
    float* coord = inference.get_output_tensor_sys(graph_name, "conv5-2");
    // output data joint
    int conf_stride = conf_shape[1];
    memcpy(conf_data + num_done * conf_stride,
          conf, num_work * conf_stride * sizeof(float));
    int coord_stride = coord_shape[1];
    memcpy(coord_data + num_done * coord_stride,
          coord, num_work * coord_stride * sizeof(float));
    // update the num done
    num_done += num_work;
  }
  // do postprocess
  vector<sail::FaceInfo> Pnet_bboxes = postprocessor.Rnet_process(
      conf_data, coord_data, img, regressed_rects);
  int num = Pnet_bboxes.size();
  cout << "The RNet total numBox is: " << num << endl;
  delete [] input_data;
  delete [] conf_data;
  delete [] coord_data;
  return Pnet_bboxes;
}

vector<sail::FaceInfo> run_Onet(  // ONet deploy
    sail::Inference<float>& inference,
    sail::MtcnnPreprocess&  preprocessor,
    sail::MtcnnPostprocess& postprocessor,
    vector<sail::FaceInfo>& RNet_output,
    cv::Mat&                frame,
    const string&           graph_name) {
  cv::Mat img;
  frame.copyTo(img);
  vector<sail::FaceInfo> regressed_rects(RNet_output);
  vector<sail::FaceInfo> regressed_pading;
  int numBox = regressed_rects.size();
  cout << "ONet input bbox num: " << numBox << endl;
  postprocessor.Padding(img, regressed_rects, regressed_pading);
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  string input_tensor_name = *(input_names.begin());
  vector<int> input_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
  // check ONet input/output shape
  // cout << "init input shape: "
  //      << input_shape[0] << " " << input_shape[1] << " "
  //      << input_shape[2] << " " << input_shape[3] << endl;
  vector<int> conf_shape = inference.get_output_shape(graph_name, "prob1");
  // cout << "init conf shape: "
  //      << conf_shape[0] << " " << conf_shape[1] << " "
  //      << conf_shape[2] << " " << conf_shape[3] << endl;
  vector<int> coord_shape = inference.get_output_shape(graph_name, "conv6-2");
  // cout << "init coord shape: "
  //      << coord_shape[0] << " " << coord_shape[1] << " "
  //      << coord_shape[2] << " " << coord_shape[3] << endl;
  vector<int> landmark_shape = inference.get_output_shape(graph_name,
                                                          "conv6-3");
  // cout << "init landmark shape: "
  //      << landmark_shape[0] << " " << landmark_shape[1] << " "
  //      << landmark_shape[2] << " " << landmark_shape[3] << endl;
  // alloc total sys mem (num > do_inference_num)
  int input_size = numBox * input_shape[1] * input_shape[2] * input_shape[3];
  float* input_data = new float[input_size];
  int conf_size = numBox * conf_shape[1];
  float* conf_data = new float[conf_size];
  int coord_size = numBox * coord_shape[1];
  float* coord_data = new float[coord_size];
  int landmark_size = numBox * landmark_shape[1];
  float* landmark_data = new float[landmark_size];
  // input img joint
  preprocessor.ronet_pre_img(input_data, img, input_shape[2], input_shape[3],
                             regressed_pading, regressed_rects);
  int num_done = 0;
  int num_fix = input_shape[0];  // num_fix == 32
  while (num_done < numBox) {
    if ((num_done + num_fix) > numBox) {
      num_fix = numBox - num_done;
    }
    int num_work = num_fix;
    cout << "*********************************" << endl;
    cout << "inference batch: " << num_work
         << " done batch: " << num_done << endl;
    cout << "*********************************"<< endl;
    vector<int> input_new_shape(4);
    input_new_shape[0] = num_work;
    input_new_shape[1] = input_shape[1];
    input_new_shape[2] = input_shape[2];
    input_new_shape[3] = input_shape[3];
    map<string, vector<int>> new_input_info;
    new_input_info.insert(pair<string, vector<int>>(input_tensor_name,
                                                    input_new_shape));
    inference.set_input_shape(graph_name, new_input_info);
    vector<int> now_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
    // check ONet now input shape
    // cout << "new input shape: "
    //      << now_shape[0] << " " << now_shape[1] << " "
    //      << now_shape[2] << " " << now_shape[3] << endl;
    // do RNet preprocess
    float* input = inference.get_input_tensor_sys(graph_name,
                                                  input_tensor_name);
    preprocessor.ronet_preprocess_cpu(input, input_data,
                                      now_shape, num_done, num_work);
    // do Rnet inference
    inference.process(graph_name);
    // get the output_tensor ptr
    float* conf = inference.get_output_tensor_sys(graph_name, "prob1");
    float* coord = inference.get_output_tensor_sys(graph_name, "conv6-2");
    float* landmark = inference.get_output_tensor_sys(graph_name, "conv6-3");
    // output data joint
    int conf_stride = conf_shape[1];
    memcpy(conf_data + num_done * conf_stride,
           conf,
           num_work * conf_stride * sizeof(float));
    int coord_stride = coord_shape[1];
    memcpy(coord_data + num_done * coord_stride,
           coord,
           num_work * coord_stride * sizeof(float));
    int landmark_stride = landmark_shape[1];
    memcpy(landmark_data + num_done * landmark_stride,
           landmark,
           num_work * landmark_stride * sizeof(float));
    // update the num done
    num_done += num_work;
  }
  // do postprocess
  vector<sail::FaceInfo> Onet_bboxes = postprocessor.Onet_process(
      conf_data, coord_data, landmark_data, img, regressed_rects);
  int num = Onet_bboxes.size();
  cout << "The ONet total numBox is: " << num << endl;
  delete [] input_data;
  delete [] conf_data;
  delete [] coord_data;
  delete [] landmark_data;
  return Onet_bboxes;
}

void make_face_bbox(  // gen face bbox
    vector<sail::FaceInfo>& faceinfos, cv::Mat& img) {
  if (faceinfos.size() == 0) {
      cout << "No Face In This Img" << endl;
  }
  int w = img.cols, h = img.rows;
  cout << "******MTCNN DETECTION RESULT*******" << endl;
  for (int i = 0; i < faceinfos.size(); i++) {
    cv::Rect rc;
    rc.x = (faceinfos[i].bbox.x1) > 0 ? faceinfos[i].bbox.x1 : 0;
    rc.y = (faceinfos[i].bbox.y1) > 0 ? faceinfos[i].bbox.y1 : 0;
    rc.width = faceinfos[i].bbox.x2 - faceinfos[i].bbox.x1;
    rc.height = faceinfos[i].bbox.y2 - faceinfos[i].bbox.y1;
    cout << "face bbox: "<< rc.x << " " << rc.y << " " << rc.width <<
     " " << rc.height << " " << endl;
    cout << "face score:" << faceinfos[i].bbox.score << endl;
    // cv::rectangle(img, rc, cv::Scalar(0, 0, 255), 2, 1, 0);
    cout << "face points: ";
    for (int j = 0; j < 5; j++) {
        cout << " x" << j << ": " << faceinfos[i].facePts.x[j]
            << " y" << j << ": " << faceinfos[i].facePts.y[j];
        // cv::circle(img, cv::Point(faceinfos[i].facePts.x[j],
        //             faceinfos[i].facePts.y[j]), 1, cv::Scalar(0, 255, 0));
    }
    cout << endl;
  }
  // cv::imshow("test", img.t());
  // cv::waitKey(0);
}

void det_mtcnn(
    const string& context_dir,
    const string& input_path,
    const string& tpu) {
  cv::Mat frame = cv::imread(input_path);
  cv::Mat input_image = frame.t();
  // init inference
  vector<sail::FaceInfo> faceinfos;
  sail::Inference<float> inference(context_dir, "BM1682_PCIE", tpu,
                                   sail::SYSIO, true, false);
  sail::Statistics part_timer;
  // mtcnn preprocess
  sail::MtcnnPreprocess preprocessor(127.5, 127.5, 127.5, 0.0078125);
  // gen face scale factor: face_factor=0.89 minSize=40
  part_timer.record_start("mtcnn");
  vector<double> scales = preprocessor.face_scale_gen(input_image, 0.89, 40);
  // mtcnn postprocess
  double threshold[3] = {0.5, 0.3, 0.7};
  sail::MtcnnPostprocess postprocessor(threshold);
  // check gen scale factor
  for (auto& iter : scales) {
    cout << iter  << " ";
  }
  cout << endl;
  // run Pnet
  string graph_name = "PNet";
  part_timer.record_start("pnet");
  vector<sail::FaceInfo> PNet_output = run_Pnet(
          inference, preprocessor, postprocessor,
          scales, input_image, graph_name);
  part_timer.record_end("pnet");
  cout << "**************Pnet finished**********" <<endl;
  cout << PNet_output.size()  <<endl;

  if (PNet_output.size() != 0) {
    // pnet output nms
    PNet_output = postprocessor.MultiNMS(PNet_output, 0.7, 'u');
    // run Rnet
    graph_name = "RNet";
    part_timer.record_start("rnet");
    vector<sail::FaceInfo> RNet_output = run_Rnet(
          inference, preprocessor, postprocessor, PNet_output,
          input_image, graph_name);
    part_timer.record_end("rnet");
    cout << "**************Rnet finished**********" <<endl;
    if (RNet_output.size() != 0) {
      // run Onet
      graph_name = "ONet";
      part_timer.record_start("onet");
      faceinfos = run_Onet(
          inference, preprocessor, postprocessor, RNet_output,
          input_image, graph_name);
      part_timer.record_end("onet");
      cout << "************Onet finished*********" <<endl;
    }
  }
  part_timer.record_end("mtcnn");
  // make face bbox
  make_face_bbox(faceinfos, input_image);
  part_timer.print_time();
}

void print_usage() {
  cout << "usage: det_mtcnn <args> " << endl
    <<  "  --context_dir:" << endl
    <<  "    Required; The dir of context after compilation." << endl
    <<  "  --input_path:" << endl
    <<  "    Required; The path to a image file." << endl
    <<  "  --tpu:" << endl
    <<  "    Optional; Specify the TPU IDs to use, separated by ','." << endl
    <<  "    Default: '-tpu all' to run on all available TPUs." << endl;
}

int main(int argc, char** argv) {
  const char* optstring = "c:i:th";
  int c = 0;
  int index = 0;
  string context_dir;
  string input_path;
  string tpu("all");
  struct option opts[] = {{"context_dir", required_argument, 0, 0},
                          {"input_path", required_argument, 0, 0},
                          {"tpu", optional_argument, 0, 0},
                          {"help", no_argument, 0, 0}};
  while ((c = getopt_long_only(argc, argv, optstring, opts, &index)) != -1) {
    switch (index) {
      case 0:
        context_dir = optarg;
        break;
      case 1:
        input_path = optarg;
        break;
      case 2:
        tpu = optarg;
        break;
      case 3:
        print_usage();
        return 0;
      default:
        break;
    }
  }
  if (context_dir.empty() || input_path.empty()) {
    print_usage();
  } else {
    det_mtcnn(context_dir, input_path, tpu);
  }
  return 0;
}
