/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <getopt.h>
#include "process/preprocess.hpp"
#include "process/postprocess.hpp"
#include "core/inference.hpp"
#include "core/tensor.hpp"
#include "core/statistics.hpp"
#include "core/utils.hpp"
using std::string;
using std::vector;
using std::set;
using std::cout;
using std::endl;

void det_yolov3(
    const string& context_dir,
    const string& input_path,
    const string& tpu) {
  sail::Inference<float> inference(context_dir, "BM1682_PCIE", tpu,
                                   sail::SYSI, false, false);
  string graph_name = *inference.get_all_graph_names().begin();
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  string input_tensor_name = *(input_names.begin());
  sail::Yolov23Preprocess preprocessor(416, 416);
  std::vector<int> input_shape;
  input_shape.push_back(1);
  input_shape.push_back(3);
  input_shape.push_back(416);
  input_shape.push_back(416);
  sail::Yolov3Postprocess postprocess(80, 0.5, 0.45, input_shape);
  sail::Statistics timer;
  timer.record_start("yolov3_time");
  cv::Mat frame = cv::imread(input_path);
  cv::Mat frame_show;
  frame.copyTo(frame_show);
  float* input = inference.get_input_tensor_sys(graph_name, input_tensor_name);
  timer.record_start("preprocess");
  preprocessor.process_cpu(input, frame);
  timer.record_end("preprocess");
  timer.record_start("inference");
  inference.process(graph_name);
  timer.record_end("inference");
  timer.record_start("postprocess");
  sail::Tensor<float> output_s = inference.get_output_tensor_dev(graph_name,
                                                           "layer82-conv");
  sail::Tensor<float> output_m = inference.get_output_tensor_dev(graph_name,
                                                           "layer94-conv");
  sail::Tensor<float> output_l = inference.get_output_tensor_dev(graph_name,
                                                           "layer106-conv");
  vector<sail::yolov3_DetectRect> dets = postprocess.process(
      inference.get_bmrt(), output_s, output_m,
      output_l, frame.cols, frame.rows);
  timer.record_end("postprocess");
  timer.record_end("yolov3_time");
  timer.print_time();
  int height = frame_show.rows;
  int width = frame_show.cols;
  cout << "detected num: " << dets.size() << endl;
  for (size_t i = 0; i < dets.size(); i++) {
    int x_min = dets[i].left;
    int x_max = dets[i].right;
    int y_min = dets[i].top;
    int y_max = dets[i].bot;
    cout << "cls: " << dets[i].category << " score: " << dets[i].score
        << " coord: " << x_min << " " << x_max << " " << y_min << " "
        << y_max << endl;
    cv::Rect rc;
    rc.x = x_min;
    rc.y = y_min;;
    rc.width = x_max - x_min;
    rc.height = y_max - y_min;
    cv::rectangle(frame_show, rc, cv::Scalar(255, 0, 255), 2, 1, 0);
  }
  // cv::imshow("results", frame_show);
  // cv::waitKey(0);
}

void print_usage() {
  cout << "usage: det_yolov3 <args> " << endl
    <<  "  --context_dir:" << endl
    <<  "    Required; The dir of context after compilation." << endl
    <<  "  --input_path:" << endl
    <<  "    Required; The path to a image file." << endl
    <<  "  --tpu:" << endl
    <<  "    Optional; Specify the TPU IDs to use, separated by ','." << endl
    <<  "    Default: '-tpu all' to run on all available TPUs." << endl;
}

int main(int argc, char** argv) {
  const char* optstring = "c:i:th";
  int c = 0;
  int index = 0;
  string context_dir;
  string input_path;
  string tpu("all");
  struct option opts[] = {{"context_dir", required_argument, 0, 0},
                          {"input_path", required_argument, 0, 0},
                          {"tpu", optional_argument, 0, 0},
                          {"help", no_argument, 0, 0}};
  while ((c = getopt_long_only(argc, argv, optstring, opts, &index)) != -1) {
    switch (index) {
      case 0:
        context_dir = optarg;
        break;
      case 1:
        input_path = optarg;
        break;
      case 2:
        tpu = optarg;
        break;
      case 3:
        print_usage();
        return 0;
      default:
        break;
    }
  }
  if (context_dir.empty() || input_path.empty()) {
    print_usage();
  } else {
    det_yolov3(context_dir, input_path, tpu);
  }
  return 0;
}
