/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <getopt.h>
#include <iomanip>
#include "process/preprocess.hpp"
#include "process/postprocess.hpp"
#include "core/inference.hpp"
#include "core/statistics.hpp"
#include "core/utils.hpp"
using std::vector;
using std::string;
using std::set;
using std::cout;
using std::endl;
using std::ios;
using std::setw;
using std::setprecision;
using std::copy;

void img_cls_sync(
      const string& context_dir,
      const string& tpu,
      int input_n, int input_c, int input_h, int input_w) {
  sail::Inference<float> inference(context_dir, "BM1682_PCIE", tpu,
                                   sail::SYSIO, false, false);

  set<string> graph_names = inference.get_all_graph_names();
  string graph_name = *(graph_names.begin());
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  set<string> output_names = inference.get_output_tensor_names(graph_name);
  // only one input tensor in the graph
  string input_tensor_name = *(input_names.begin());
  string output_tensor_name = *(output_names.begin());
  vector<int> input_shape = inference.get_input_shape(graph_name,
                                                       input_tensor_name);
  // vector<int> output_shape = inference.get_output_shape(graph_name,
  //                                                       output_tensor_name);
  float* input = inference.get_input_tensor_sys(graph_name, input_tensor_name);

  sail::Statistics timer;
  int init_data_size = sail::get_size_of_shape(input_shape);
  int input_data_size = input_n * input_c * input_h * input_w;
  cout << "model init data size: " << init_data_size << " input data size: "
      << input_data_size << endl;
  if (init_data_size != input_data_size) {
      cout << "input shape error !" << endl;
      exit(-1);
  }

  // fill random data
  float* random_data = new float[input_data_size];
  for (int i = 0; i < input_data_size; i++) {
    random_data[i] = rand() % 256;
  }

  for (int i = 0; i < 120; i++) {
    timer.record_start("inference");
    memcpy(input, random_data, input_data_size * sizeof(float));
    // copy(random_data, random_data+input_data_size, input);
    inference.process(graph_name);
    timer.record_end("inference");
  }
  // print the time
  cout << graph_name << " batch " << input_n <<" ";
  timer.print_time();
}

void img_cls_async(
      const string& context_dir,
      const string& tpu,
      int input_n, int input_c, int input_h, int input_w) {
  sail::Inference<float> inference(context_dir, "BM1682_PCIE", tpu,
                                   sail::SYSIO, false, true);

  set<string> graph_names = inference.get_all_graph_names();
  string graph_name = *(graph_names.begin());
  set<string> input_names = inference.get_input_tensor_names(graph_name);
  set<string> output_names = inference.get_output_tensor_names(graph_name);
  // only one input tensor in the graph
  string input_tensor_name = *(input_names.begin());
  string output_tensor_name = *(output_names.begin());
  vector<int> input_shape = inference.get_input_shape(graph_name,
                                                      input_tensor_name);
  // vector<int> output_shape = inference.get_output_shape(graph_name,
  //                                                       output_tensor_name);
  float* input = inference.get_input_tensor_sys(graph_name, input_tensor_name);

  sail::Statistics timer;
  int init_data_size = sail::get_size_of_shape(input_shape);
  int input_data_size = input_n * input_c * input_h * input_w;
  cout << "model init data size: " << init_data_size << " input data size: "
      << input_data_size << endl;
  if (init_data_size != input_data_size) {
      cout << "input shape error !" << endl;
      exit(-1);
  }

  // fill random data
  float* random_data = new float[input_data_size];
  for (int i = 0; i < input_data_size; i++) {
    random_data[i] = rand() % 256;
  }

  sail::TensorMode mode = sail::PING;
  sail::AsyncStatus status = sail::BEGIN;
  for (int i = 0; i < 120; i++) {
    timer.record_start("inference");
    // memcpy(input, random_data, input_data_size * sizeof(float));
    copy(random_data, random_data+input_data_size, input);
    inference.process(graph_name, mode, status);
    timer.record_end("inference");
    mode = (mode == sail::PING) ? sail::PONG : sail::PING;
    status = sail::PROCESS;
  }
  timer.record_start("inference");
  inference.process(graph_name, mode, sail::END);
  timer.record_end("inference");
  // print the time
  cout << "async!!!!!!" << endl;
  cout << graph_name << " batch " << input_n <<" ";
  timer.print_time();
}

void print_usage() {
  cout << "usage: val <args> " << endl
    <<  "  --context_dir:" << endl
    <<  "    Required; The dir of context after compilation." << endl
    <<  "  --batch:" << endl
    <<  "    Required; model batch size" << endl
    <<  "  --channel:" << endl
    <<  "    Required; model channel size" << endl
    <<  "  --height:" << endl
    <<  "    Required; model height size" << endl
    <<  "  --width:" << endl
    <<  "    Required; model width size" << endl
    <<  "  --sync_mode:" << endl
    <<  "    Required; Specify the processing mode " << endl;
}

int main(int argc, char** argv) {
  const char* optstring = "r:n:c:h:w:s";
  int c = 0;
  int batch = 1, channel = 1, height = 1, width = 1;
  int index = 0;
  string context_dir;
  string tpu("0");
  string sync_mode("sync");
  struct option opts[] = {{"context_dir", required_argument, 0, 0},
                          {"batch", required_argument, 0, 0},
                          {"channel", required_argument, 0, 0},
                          {"height", required_argument, 0, 0},
                          {"width", required_argument, 0, 0},
                          {"sync_mode", required_argument, 0, 0}};
  while ((c = getopt_long_only(argc, argv, optstring, opts, &index)) != -1) {
    switch (index) {
      case 0:
        context_dir = optarg;
        break;
      case 1:
        batch = atoi(optarg);
        break;
      case 2:
        channel = atoi(optarg);
        break;
      case 3:
        height = atoi(optarg);
        break;
      case 4:
        width = atoi(optarg);
        break;
      case 5:
        sync_mode = optarg;
        break;
      case 6:
        print_usage();
        return 0;
      default:
        break;
    }
  }
  if (context_dir.empty()) {
    print_usage();
  } else {
    if (sync_mode.compare("sync") == 0) {
      img_cls_sync(context_dir, tpu, batch, channel, height, width);
    } else if (sync_mode.compare("async") == 0) {
      img_cls_async(context_dir, tpu, batch, channel, height, width);
    } else {
      cout << "Wrong sync_mode" << endl;
    }
  }
  return 0;
}
