# Copyright 2016-2022 Bitmain Technologies Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Entry of python samples.
"""

from __future__ import print_function
import time
import argparse
import sys
import cv2
from algokit.algofactory.objectdetection_factory import ObjectDetector
from algokit.algofactory.facedetection_factory import FaceDetector
from algokit.algofactory.generalclassification_factory import GeneralClassifier
from algokit.algofactory.autodeploy_factory import ObjectDetector \
                                            as Auto_ObjectDetector
from algokit.algofactory.autodeploy_factory import SemanticSegment \
                                            as Auto_SemanticSegment
from algokit.kitconfig import ChipMode, ObjDetModel, FaceDetModel
from algokit.kitconfig import ClassificationModel, SegmentationModel

def run_ojbdetection():
  """
  test object detection
  """
  model_name = None
  for iterator in ObjDetModel:
    if ARGS.algo_name == iterator.value:
      model_name = iterator
  if model_name is None:
    raise ValueError('not a valid algorithm name', ARGS.algo_name)
  if ARGS.algo_name in ['fasterrcnn_resnet50_tf']:
    detector = Auto_ObjectDetector().create(model_name, ChipMode.BM1682)
  else:
    detector = ObjectDetector().create(model_name, ChipMode.BM1682)
  input_image = cv2.imread(ARGS.input_path)
  t_0 = time.time()
  bboxes, classes, probs = detector.predict(input_image)
  print('detect time: {}'.format(time.time()-t_0))
  print("\x1b[32m\x1b" + "#"*31 + "Detection Results" +"#"*32+ "\x1b[0m")
  i = 0
  for bbox, cls, prob in zip(bboxes, classes, probs):
    print("{}: bbox: {} cls: {} prob: {}".format(i, bbox, cls, prob))
    i += 1
  print("\x1b[32m\x1b" + "#"*80+ "\x1b[0m")
  # result = draw_obj_bbox(input_image, bboxes, classes, probs, "voc")
  # cv2.imshow("test", result)
  # cv2.waitKey()

def run_facedetection():
  """
  test face detection
  """
  model_name = None
  for iterator in FaceDetModel:
    if ARGS.algo_name == iterator.value:
      model_name = iterator
  if model_name is None:
    raise ValueError('not a valid algorithm name', ARGS.algo_name)
  detector = FaceDetector().create(model_name, ChipMode.BM1682)
  input_image = cv2.imread(ARGS.input_path)
  t_0 = time.time()
  bboxes, points, probs = detector.predict(input_image)
  print('detect time: {}'.format(time.time()-t_0))
  print("\x1b[32m\x1b" + "#"*31 + "Detection Results" +"#"*32+ "\x1b[0m")
  i = 0
  for bbox, point, prob in zip(bboxes, points, probs):
    print("{}: bbox: {} prob: {}".format(i, bbox, prob))
    print("point: {}".format(point))
    i += 1
  print("\x1b[32m\x1b" + "#"*80+ "\x1b[0m")
  # result = draw_face_bbox(input_image, bboxes, probs, points)
  # cv2.imshow("test", result)
  # cv2.waitKey()

def run_classification():
  """
  test classification && recognition
  """
  model_name = None
  for iterator in ClassificationModel:
    if ARGS.algo_name == iterator.value:
      model_name = iterator
  if model_name is None:
    raise ValueError('not a valid algorithm name', ARGS.algo_name)
  classifier = GeneralClassifier().create(model_name, ChipMode.BM1682)
  input_image = cv2.imread(ARGS.input_path)
  t_0 = time.time()
  top1, top5 = classifier.predict(input_image)
  print('detect time: {}'.format(time.time() - t_0))
  print("top1_idx: {} top1_score: {}".format(top1['top1_idx'], \
         top1['top1_score']))
  print("top5_idx: {} top5_score: {}".format(top5['top5_idx'], \
        top5['top5_score']))

def run_segmentation():
  """
  test segmentation
  """
  model_name = None
  for iterator in SegmentationModel:
    if ARGS.algo_name == iterator.value:
      model_name = iterator
  if model_name is None:
    raise ValueError('not a valid algorithm name', ARGS.algo_name)
  segment = Auto_SemanticSegment().create(model_name, ChipMode.BM1682)
  input_image = cv2.imread(ARGS.input_path)
  t_0 = time.time()
  segment_results = segment.predict(input_image)
  print('segment time: {}'.format(time.time()-t_0))
  print("segment results: {} {}".format(segment_results.shape, \
                                        segment_results))

def run_algokit():
  """
  test supported algorithms in algokit
  """
  if ARGS.check_list:
    if ARGS.algo_type == "cls":
      run_classification()
    elif ARGS.algo_type == "obj_det":
      run_ojbdetection()
    elif ARGS.algo_type == "face_det":
      run_facedetection()
    elif ARGS.algo_type == "seg":
      run_segmentation()
    else:
      raise ValueError('not a valid algorithmtype: ', ARGS.algo_type)
  else:
    print("The Algorithm List Of Algokit:")
    classification_list = GeneralClassifier().get_supported_models()
    print("Classification_List:\n {}".format(classification_list))
    objectdetection_list = ObjectDetector().get_supported_models()
    print("Objectdetection_List:\n {}".format(objectdetection_list))
    facedetection_list = FaceDetector().get_supported_models()
    print("Facedetection_List:\n {}".format(facedetection_list))
    semanticsegmentation_list = Auto_SemanticSegment().get_supported_models()
    print("Semanticsegmentation_List:\n {}".format(semanticsegmentation_list))

if __name__ == '__main__':
  # algokit argparse
  if len(sys.argv) < 2:
    print("usage1 check algorithm list in algokit: ")
    print("python run_algokit.py --check_list")
    print("usage2 test algorithm in algokit: ")
    print("python run_algokit.py ")
    print("--algo_type (cls | obj_det | face_det | seg) \\ ")
    print("--algo_name (resnet50|yolov3|mtcnn|deeplabv3_mobilenetv2_tf) \\ ")
    print("--input file_path")
    sys.exit()
  PARSER = argparse.ArgumentParser(description='for algokit test')
  PARSER.add_argument('--algo_type', help='algorithm type', default='cls')
  PARSER.add_argument('--algo_name', help='algorithm name', default='resnet50')
  PARSER.add_argument('--input_path', help='test file path', \
                      default='../test/test_data/cls_1.jpg')
  PARSER.add_argument('--check_list', help='check algorithm list', \
                      action="store_false")
  ARGS = PARSER.parse_args()
  run_algokit()
