/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include "core/common.hpp"
#include "core/bm_cv.hpp"
#include "core/utils.hpp"

namespace sail {

//#ifdef USING_FFMPEG
VideoDecoder::VideoDecoder(const string& file_path)
    : file_path_(file_path), fmt_ctx_(NULL), dec_ctx_(NULL),
      pix_fmt_(AV_PIX_FMT_NONE), packet_(NULL), stream_(NULL),
      pcie_board_id_(0), refcount_(0), stream_index_(-1),
      height_(0), width_(0), status_(false), total_frames_(0),
      seeking_type_(AVSEEK_FLAG_FRAME), read_end_(false),
      inside_decode_(false), start_(true) {
  av_register_all();
  av_log_set_callback(av_log_callback);
  int seeking_type = AVSEEK_FLAG_FRAME;
  int ret = avformat_open_input(&fmt_ctx_, file_path_.c_str(), NULL, NULL);
  if (ret < 0) {
    cerr << "Failed to open input file: " << file_path_ << endl;
    return;
  }
  ret = avformat_find_stream_info(fmt_ctx_, NULL);
  if (ret < 0) {
    cerr << "Failed to find stream information." << endl;
    return;
  }
  ret = open_codec_context();
  if (!ret) {
    cerr << "Failed to open codec context." << endl;
    return;
  }
  if (dec_ctx_->codec_id == AV_CODEC_ID_HEVC) {
    seeking_type = AVSEEK_FLAG_BYTE;
  }
  // initialize packet, set data to NULL, let the demuxer fill it
  packet_ = av_packet_alloc();
  av_init_packet(packet_);
  packet_->data = NULL;
  packet_->size = 0;
  if (!packet_) {
    cerr << "Failed to alloc packet." << endl;
  }
  status_ = true;
}

bool VideoDecoder::open_codec_context() {
  AVMediaType type = AVMEDIA_TYPE_VIDEO;
  int ret = av_find_best_stream(fmt_ctx_, type, -1, -1, NULL, 0);
  if (ret < 0) {
    cerr << "Failed to  find a video stream in the input file." << endl;
    return false;
  }
  stream_index_ = ret;
  stream_ = fmt_ctx_->streams[stream_index_];
  // find decoder for the stream
  AVCodec* dec = avcodec_find_decoder(stream_->codecpar->codec_id);
  if (!dec) {
    cerr << "Failed to find codec." << endl;
    return false;
  }

  // Allocate a codec context for the decoder
  dec_ctx_ = avcodec_alloc_context3(dec);
  if (!dec_ctx_) {
    cerr << "Failed to allocate codec context." << endl;
    return false;
  }

  // Copy codec parameters from input stream to output codec context
  ret = avcodec_parameters_to_context(dec_ctx_, stream_->codecpar);
  if (ret < 0) {
    cerr << "Failed to copy codec parameters to decoder context." << endl;
    return false;
  }
  height_ = dec_ctx_->height;
  width_ = dec_ctx_->width;
  pix_fmt_ = dec_ctx_->pix_fmt;

  // Init the decoders, with or without reference counting
  AVDictionary* opts = NULL;
  av_dict_set(&opts, "refcounted_frames", refcount_ ? "1" : "0", 0);
  av_dict_set_int(&opts, "pcie_board_id", pcie_board_id_, 0x0);
  if ((ret = avcodec_open2(dec_ctx_, dec, &opts)) < 0) {
    cerr << "Failed to open codec" << endl;
  	return false;
  }
  return true;
}

bool VideoDecoder::is_opened() {
  return status_;
}

vector<int> VideoDecoder::get_frame_shape() {
  vector<int> shape(4);
  shape[0] = 1;
  shape[1] = 3;
  shape[2] = height_;
  shape[3] = width_;
  return shape;
}

VideoDecoder::~VideoDecoder() {
  av_seek_frame(fmt_ctx_, -1, 1, seeking_type_);
  av_packet_free(&packet_);
  avcodec_free_context(&dec_ctx_);
  avformat_close_input(&fmt_ctx_);
}

int VideoDecoder::decode_packet(AVFrame* frame) {
  int decoded = packet_->size;
  if (packet_->stream_index == stream_index_) {
    int ret = avcodec_send_packet(dec_ctx_, packet_);
    if (ret < 0) {
      cout << "avcodec_send_packet < 0" << endl;
      return ret;
    }
    ret = avcodec_receive_frame(dec_ctx_, frame);
    if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF) {
      cout << "ret == AVERROR(EAGAIN) || ret == AVERROR_EOF" << endl;
      return ret;
    } else if (ret < 0) {
      return ret;
    }
    if (ret >= 0) {
      if (frame->width != width_ || frame->height != height_
          || frame->format != pix_fmt_) {
        cerr << "frame shape not match" << endl;
        return -1;
      }
      ++total_frames_;
    }
  }
  return decoded;
}

int VideoDecoder::read(Frame& frame) {
  AVFrame* p_frame = frame.get();
  int status = 1;
  if (!read_end_) {
    while (status > 0) {
      if (!start_ && !inside_decode_) {
        av_packet_unref(packet_);
        if (packet_->pts > 15000000) {
          av_seek_frame(fmt_ctx_, -1, 1, seeking_type_);
        }
        if (total_frames_ > 40000) {
          av_seek_frame(fmt_ctx_, -1, 1, seeking_type_);
          total_frames_ = 0;
        }
      }
      if (!inside_decode_) {
        status = av_read_frame(fmt_ctx_, packet_);
        if (status < 0) {
          read_end_ = true;
        } else if (start_) {
          start_ = false;
        }
      }
      if (!read_end_) {
        int ret = decode_packet(p_frame);
        if (ret < 0) {
           inside_decode_ = false;
           status = 1;
        } else {
           packet_->data += ret;
           packet_->size -= ret;
           status = 0;
          if (packet_->size <= 0) {
            inside_decode_ = false;
          } else {
            inside_decode_ = true;
          }
        }
      }
    }
  } else {
    packet_->data = NULL;
    packet_->size = 0;
    decode_packet(p_frame);
  }
  return status;
}

//#ifdef BMDNN_CV
Bmcv::~Bmcv() {
}

bmcv_image Bmcv::get_bmcv_image(Tensor<float>& tensor) {
  vector<int> shape = tensor.shape();
  bmcv_image img;
  img.color_space = COLOR_RGB;
  img.data_format = DATA_TYPE_BYTE;
  img.image_format = BGR;
  img.image_width = shape[3];
  img.image_height = shape[2];
  img.data[0] = *tensor.dev_data();
  img.stride[0] = shape[3];
  return img;
}

void Bmcv::get_bgr_from_frame(Frame& frame, Tensor<float>& bgr) {
  AVFrame* p_frame = frame.get();
  if (p_frame->format != AV_PIX_FMT_NV12) {
    cerr << "Error: Only AV_PIX_FMT_NV12 is supported." << endl;
  }
  bm_device_mem_t dmem_y = bm_mem_from_device(
      (unsigned long long)p_frame->data[4],
      p_frame->linesize[4] * p_frame->height);
  bm_device_mem_t dmem_uv = bm_mem_from_device(
      (unsigned long long)p_frame->data[5],
      p_frame->linesize[5] * p_frame->height);
  vector<int> y_shape(2);
  y_shape[0] = p_frame->linesize[4];
  y_shape[1] = p_frame->height;
  Tensor<float> y(p_bmrt_, y_shape, &dmem_y);
  vector<int> uv_shape(2);
  uv_shape[0] = p_frame->linesize[5];
  uv_shape[1] = p_frame->height;
  Tensor<float> uv(p_bmrt_, uv_shape, &dmem_uv);
  int ret = yuv2bgr(y, uv, bgr);
}

int Bmcv::resize(
    Tensor<float>&        input,
    Tensor<float>&        output) {
  bmcv_image in_img = get_bmcv_image(input);
  bmcv_image out_img = get_bmcv_image(output);
  out_img.data_format = DATA_TYPE_FLOAT;
  int batch_size = input.shape()[0];
  int ret = bmcv_img_scale(handle_, in_img, batch_size,
                           0, 0, 0, 0, 0,
                           1, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0,
                           out_img);
  return ret;
}

int Bmcv::crop_resize_norm(
    Tensor<float>&        input,
    Tensor<float>&        output,
    int                   top,
    int                   left,
    int                   crop_height,
    int                   crop_width,
    float                 scale_b,
    float                 bias_b,
    float                 scale_g,
    float                 bias_g,
    float                 scale_r,
    float                 bias_r) {
  bmcv_image in_img = get_bmcv_image(input);
  bmcv_image out_img = get_bmcv_image(output);
  out_img.data_format = DATA_TYPE_FLOAT;
  int batch_size = input.shape()[0];
  int ret = bmcv_img_scale(handle_, in_img, batch_size,
                           1, top, left, crop_height, crop_width,
                           1, 0, 0, 0,
                           1, scale_b, bias_b, scale_g, bias_g, scale_r, bias_r,
                           out_img);
  return ret;
}

int Bmcv::resize_norm(
    Tensor<float>&        input,
    Tensor<float>&        output,
    float                 scale_b,
    float                 bias_b,
    float                 scale_g,
    float                 bias_g,
    float                 scale_r,
    float                 bias_r) {
  bmcv_image in_img = get_bmcv_image(input);
  bmcv_image out_img = get_bmcv_image(output);
  out_img.data_format = DATA_TYPE_FLOAT;
  int batch_size = input.shape()[0];
  int ret = bmcv_img_scale(handle_, in_img, batch_size,
                           0, 0, 0, 0, 0,
                           1, 0, 0, 0,
                           1, scale_b, bias_b, scale_g, bias_g, scale_r, bias_r,
                           out_img);
  return ret;
}

int Bmcv::crop(
    Tensor<float>&        input,
    Tensor<float>&        output,
    int                   top,
    int                   left) {
  bmcv_image in_img = get_bmcv_image(input);
  bmcv_image out_img = get_bmcv_image(output);
  int channels = input.shape()[1];
  int ret = bmcv_img_crop(handle_, in_img, channels, top, left, out_img);
  return ret;
}

int Bmcv::yuv2bgr(
    Tensor<float>&        input_y,
    Tensor<float>&        input_uv,
    Tensor<float>&        output) {
  bmcv_image in_img = get_bmcv_image(output);
  in_img.color_space = COLOR_YUV;
  in_img.image_format = NV12;
  in_img.data[0] = *input_y.dev_data();
  in_img.stride[0] = input_y.shape()[0];
  in_img.data[1] = *input_uv.dev_data();
  in_img.stride[1] = input_uv.shape()[0];
  bmcv_image out_img = get_bmcv_image(output);
  int ret = bm_img_yuv2bgr(handle_, in_img, out_img);
  return ret;
}

int Bmcv::transpose(
    Tensor<float>&        input,
    Tensor<float>&        output) {
  bmcv_image in_img = get_bmcv_image(input);
  bmcv_image out_img = get_bmcv_image(output);
  int ret = bmcv_img_transpose(handle_, in_img, out_img);
  return ret;
}

int Bmcv::bgrsplit(
    Tensor<float>&        input,
    Tensor<float>&        output) {
  bmcv_image in_img = get_bmcv_image(input);
  bmcv_image out_img = get_bmcv_image(output);
  int ret = bmcv_img_bgrsplit(handle_, in_img, out_img);
  return ret;
}
//#endif

}  // namespace sail
