/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <iostream>
#include "core/common.hpp"
#include "core/utils.hpp"
#include "core/graph.hpp"

namespace sail {

template <typename Dtype>
Graph<Dtype>::Graph(
    void*                     p_bmrt,
    const string&             name,
    IOMode                    mode,
    bool                      dynamic,
    bool                      async,
    map<string, vector<int>>& input_shapes,
    map<string, vector<int>>& output_shapes,
    map<string, Dtype*>&      input_tensors,
    map<string, Dtype*>&      output_tensors)
    : p_bmrt_(p_bmrt), name_(name), io_mode_(mode), dynamic_(dynamic),
      async_(async), input_mem_ping_(NULL), input_mem_pong_(NULL),
      output_mem_ping_(NULL), output_mem_pong_(NULL) {
  if (!p_bmrt) {
    spdlog::error("Error while constructing Graph: bmruntime is null");
    return;
  }
  get_input_tensor_idx();
  get_output_tensor_idx();
  input_mem_ping_ = new bm_device_mem_t[input_tensors.size()];
  output_mem_ping_ = new bm_device_mem_t[output_tensors.size()];
  if (async_) {
    input_mem_pong_ = new bm_device_mem_t[input_tensors.size()];
    output_mem_pong_ = new bm_device_mem_t[output_tensors.size()];
  }
  input_tensors_.assign(input_map_.size(), 0);
  for (auto iter = input_shapes.begin(); iter != input_shapes.end(); ++iter) {
    int idx = input_map_[iter->first];
    bm_device_mem_t* pong = async_ ? input_mem_pong_ + idx : NULL;
    Tensor<Dtype>* tensor = new Tensor<Dtype>(p_bmrt_, iter->second,
        false, false, async_, input_tensors[iter->first],
        input_mem_ping_ + idx, pong);
    input_tensors_[idx] = tensor;
    tensor_map_[iter->first] = tensors_.size();
    tensors_.push_back(std::shared_ptr<Tensor<Dtype>>(tensor));
  }
  output_tensors_.assign(output_map_.size(), 0);
  for (auto iter = output_shapes.begin(); iter != output_shapes.end(); ++iter) {
    int idx = output_map_[iter->first];
    bm_device_mem_t* pong = async_ ? output_mem_pong_ + idx : NULL;
    Tensor<Dtype>* tensor = new Tensor<Dtype>(p_bmrt_, iter->second,
        false, false, async_, output_tensors[iter->first],
        output_mem_ping_ + idx, pong);
    output_tensors_[idx] = tensor;
    tensor_map_[iter->first] = tensors_.size();
    tensors_.push_back(std::shared_ptr<Tensor<Dtype>>(tensor));
  }
}

template <typename Dtype>
void Graph<Dtype>::set_dynamic_output_shapes(
    map<string, vector<int>>* p_output_shapes) {
  dynamic_output_shapes_ = p_output_shapes;
}

template <typename Dtype>
void Graph<Dtype>::free_tensors() {
  tensors_.clear();
  input_tensors_.clear();
  output_tensors_.clear();
  tensor_map_.clear();
  if (input_mem_ping_) {
    delete [] input_mem_ping_;
    input_mem_ping_ = NULL;
  }
  if (input_mem_pong_) {
    delete [] input_mem_pong_;
    input_mem_pong_ = NULL;
  }
  if (output_mem_ping_) {
    delete [] output_mem_ping_;
    output_mem_ping_ = NULL;
  }
  if (output_mem_pong_) {
    delete [] output_mem_pong_;
    output_mem_pong_ = NULL;
  }
}

template <typename Dtype>
Graph<Dtype>::~Graph() {
  free_tensors();
}

template <typename Dtype>
Dtype* Graph<Dtype>::get_tensor_sys(const string& name) const {
  auto it = tensor_map_.find(name);
  if (it != tensor_map_.end()) {
    return tensors_[it->second]->sys_data();
  } else {
    return NULL;
  }
}

template <typename Dtype>
bm_device_mem_t* Graph<Dtype>::get_tensor(const string& name) {
  auto it = tensor_map_.find(name);
  if (it != tensor_map_.end()) {
    return tensors_[it->second]->dev_data();
  } else {
    return NULL;
  }
}

template <typename Dtype>
bm_device_mem_t* Graph<Dtype>::get_tensor(
    const std::string& name,
    TensorMode         mode) {
  auto it = tensor_map_.find(name);
  if (it != tensor_map_.end()) {
    return tensors_[it->second]->dev_data(mode);
  } else {
    return NULL;
  }
}

template <typename Dtype>
void Graph<Dtype>::reshape(
    map<string, vector<int>>& input_shapes,
    map<string, Dtype*>&      input_tensors,
    bool                      force_reshape) {
  for (auto iter = input_shapes.begin(); iter != input_shapes.end(); ++iter) {
    size_t idx = tensor_map_[iter->first];
    if (force_reshape || get_size_of_shape(iter->second) !=
        get_size_of_shape(tensors_[idx]->shape())) {
      tensors_[idx]->reshape(iter->second);
      tensors_[idx]->reset_sys_data(input_tensors[iter->first]);
    }
  }
}

template <typename Dtype>
void Graph<Dtype>::reshape(
    map<string, vector<int>>& input_shapes,
    map<string, Dtype*>&      input_tensors,
    bool                      force_reshape,
    TensorMode                mode) {
  for (auto iter = input_shapes.begin(); iter != input_shapes.end(); ++iter) {
    size_t idx = tensor_map_[iter->first];
    if (force_reshape || get_size_of_shape(iter->second) !=
        get_size_of_shape(tensors_[idx]->shape())) {
      tensors_[idx]->reshape(iter->second, mode);
      tensors_[idx]->reset_sys_data(input_tensors[iter->first]);
    }
  }
}

template <typename Dtype>
void Graph<Dtype>::get_input_tensor_idx() {
  int index = 0;
  int input_num = 0;
  const char** input_names = NULL;
  bmrt_get_input_tensor(p_bmrt_, 0, name_.c_str(), &input_num, &input_names);
  for (int i = 0; i < input_num; ++i) {
    input_map_[input_names[i]] = i;
  }
  free(input_names);
}

template <typename Dtype>
void Graph<Dtype>::get_output_tensor_idx() {
  int index = 0;
  int output_num = 0;
  const char** output_names = NULL;
  bmrt_get_output_tensor(p_bmrt_, 0, name_.c_str(), &output_num, &output_names);
  for (int i = 0; i < output_num; ++i) {
    output_map_[output_names[i]] = i;
  }
  free(output_names);
}

template <typename Dtype>
void Graph<Dtype>::input_s2d() {
  for (size_t i = 0; i < input_tensors_.size(); ++i) {
    input_tensors_[i]->sync_s2d();
  }
}

template <typename Dtype>
void Graph<Dtype>::input_s2d(TensorMode mode) {
  for (size_t i = 0; i < input_tensors_.size(); ++i) {
    input_tensors_[i]->sync_s2d(mode);
  }
}

template <typename Dtype>
void Graph<Dtype>::output_d2s() {
  for (size_t i = 0; i < output_tensors_.size(); ++i) {
    output_tensors_[i]->sync_d2s();
  }
}

template <typename Dtype>
void Graph<Dtype>::output_d2s(const vector<int>& size) {
  for (size_t i = 0; i < output_tensors_.size(); ++i) {
    output_tensors_[i]->sync_d2s(size[i]);
  }
}

template <typename Dtype>
void Graph<Dtype>::output_d2s(TensorMode mode) {
  for (size_t i = 0; i < output_tensors_.size(); ++i) {
    output_tensors_[i]->sync_d2s(mode);
  }
}

template <typename Dtype>
void Graph<Dtype>::output_d2s(TensorMode mode, const vector<int>& size) {
  for (size_t i = 0; i < output_tensors_.size(); ++i) {
    output_tensors_[i]->sync_d2s(mode, size[i]);
  }
}

template <typename Dtype>
bm_device_mem_t* Graph<Dtype>::get_input_mems() {
  return input_mem_ping_;
}

template <typename Dtype>
bm_device_mem_t* Graph<Dtype>::get_input_mems(TensorMode mode) {
  return (mode == PING) ? input_mem_ping_ : input_mem_pong_;
}

template <typename Dtype>
bm_device_mem_t* Graph<Dtype>::get_output_mems() {
  return output_mem_ping_;
}

template <typename Dtype>
bm_device_mem_t* Graph<Dtype>::get_output_mems(TensorMode mode) {
  return (mode == PING) ? output_mem_ping_ : output_mem_pong_;
}

// asynchronous version of inference
template <typename Dtype>
void Graph<Dtype>::inference(TensorMode mode, AsyncStatus status) {
  TensorMode opposite_mode = (mode == PING) ? PONG : PING;
  // Copy input data from system memory to device memory
  if (io_mode_ == SYSI || io_mode_ == SYSIO) {
    if (status != END) {
      input_s2d(mode);
    }
  }

  // dynamic async batch
  static int dyn_batch_ping = 0, dyn_batch_pong = 0;
  if (mode == PING)
    dyn_batch_ping = input_tensors_[0]->shape()[0];
  else
    dyn_batch_pong = input_tensors_[0]->shape()[0];

  // Copy output data from device memory to system memory
  if (io_mode_ == SYSO || io_mode_ == SYSIO) {
    if (status != BEGIN) {
      if (dynamic_) {
        vector<int> size(output_tensors_.size());
        for (auto it = dynamic_output_shapes_->begin(); it != dynamic_output_shapes_->end(); ++it) {
          bmrt_get_accurate_output_shape(p_bmrt_, it->first.c_str(), 0, name_.c_str(),
                                it->second.data());
          it->second[0] = (opposite_mode == PING) ? dyn_batch_ping :
          it->second[0] = (opposite_mode == PING) ? dyn_batch_ping :
              dyn_batch_pong; // output batch == input batch
          int index = output_map_[it->first];
          size[index] = get_size_of_shape(it->second) * sizeof(Dtype);
        }
        output_d2s(opposite_mode, size);
      } else {
        output_d2s(opposite_mode);
      }
    }
  }
  // Inference
  if (status != END) {
    if (dynamic_) {
      // vector<int> shape = input_tensors_[0]->shape();
      vector<int> input_dims;
      vector<int> input_shape;
      for (int i = 0; i < input_tensors_.size(); i++) {
        vector<int> shape = input_tensors_[i]->shape();
        input_dims.push_back(shape.size());
        input_shape.insert(input_shape.end(), shape.begin(), shape.end());
      }
      bool res = bmrt_launch_shape(p_bmrt_,
                                  0,
                                  name_.c_str(),
                                  get_input_mems(),
                                  input_tensors_.size(),
                                  &input_dims[0],
                                  &input_shape[0],
                                  get_output_mems(),
                                  output_tensors_.size());
    } else {
      bool res = bmrt_launch(p_bmrt_,
                             0,
                             name_.c_str(),
                             get_input_mems(mode),
                             input_tensors_.size(),
                             get_output_mems(mode),
                             output_tensors_.size());
    }
  }
}

// synchronous version of inference
template <typename Dtype>
void Graph<Dtype>::inference() {
  // Copy input data from system memory to device memory
  if (io_mode_ == SYSI || io_mode_ == SYSIO) {
    input_s2d();
  }
  // get all input tensor shape before inference
  vector<int> shape_for_output = input_tensors_[0]->shape();
  vector<int> input_dims;
  vector<int> input_shape;
  for (int i = 0; i < input_tensors_.size(); i++) {
    vector<int> shape = input_tensors_[i]->shape();
    input_dims.push_back(shape.size());
    input_shape.insert(input_shape.end(), shape.begin(), shape.end());
  }
  // Inference
  if (dynamic_) {
    bool res = bmrt_launch_shape(p_bmrt_,
                                0,
                                name_.c_str(),
                                get_input_mems(),
                                input_tensors_.size(),
                                &input_dims[0],
                                &input_shape[0],
                                get_output_mems(),
                                output_tensors_.size());
  } else {
    bool res = bmrt_launch(p_bmrt_,
                           0,
                           name_.c_str(),
                           get_input_mems(),
                           input_tensors_.size(),
                           get_output_mems(),
                           output_tensors_.size());
  }

  // Copy output data from device memory to system memory
  if (io_mode_ == SYSO || io_mode_ == SYSIO) {
    if (dynamic_) {
      vector<int> size(output_tensors_.size());
      for (auto it = dynamic_output_shapes_->begin(); it != dynamic_output_shapes_->end(); ++it) {
        bmrt_get_accurate_output_shape(p_bmrt_, it->first.c_str(), 0, name_.c_str(),
                              it->second.data());
        if (input_tensors_.size() == 1) {
          it->second[0] = shape_for_output[0]; // output batch == input batch
        }
        int index = output_map_[it->first];
        size[index] = get_size_of_shape(it->second) * sizeof(Dtype);
      }
      output_d2s(size);
    } else {
      output_d2s();
    }
  } else {
    bmrt_thread_sync(p_bmrt_);
  }
}

template class Graph<float>;

}  // namespace sail
