/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <string>
#include <vector>
#include <algorithm>
#include <memory>
#include "core/common.hpp"
#include "core/utils.hpp"
#include "core/inference.hpp"

namespace sail {

template <typename Dtype>
Inference<Dtype>::Inference(
    const string& context_dir,
    const string& chip_name,
    const string& tpus,
    IOMode        mode,
    bool          dynamic,
    bool          async)
    : io_mode_(mode), dynamic_(dynamic), async_(async), worker_num_(0) {
  int ret = init_bmrts(context_dir, chip_name, tpus);
  if (ret != 0) {
    spdlog::error("Init bmruntime failed.");
    return;
  }
  set<string> graph_names = get_all_graph_names();
  for (auto gname : graph_names) {
    input_shapes_[gname]             = get_max_input_shapes(gname, true);
    output_shapes_[gname]            = get_max_output_shapes(gname, true);
    max_output_shapes_single_[gname] = get_max_output_shapes(gname, false);
    max_input_shapes_single_[gname]  = get_max_input_shapes(gname, false);
  }

  ret = init_graphs();
  if (ret != 0) {
    spdlog::error("Init graphs failed.");
    return;
  }
}

template <typename Dtype>
Inference<Dtype>::Inference(const Inference& other) {
  io_mode_       = other.io_mode_;
  async_         = other.async_;
  tpus_          = other.tpus_;
  input_shapes_  = other.input_shapes_;
  output_shapes_ = other.output_shapes_;
  max_input_shapes_single_  = other.max_input_shapes_single_;
  max_output_shapes_single_ = other.max_output_shapes_single_;
  dynamic_output_shapes_    = other.dynamic_output_shapes_;
  // dynamic_ must be assigned after free_pointers_sys()
  free_pointers();
  dynamic_ = other.dynamic_;
  graphs_  = other.graphs_;
}

template <typename Dtype>
Inference<Dtype>& Inference<Dtype>::operator=(const Inference& other){
  if (this != &other) {
    io_mode_       = other.io_mode_;
    async_         = other.async_;
    tpus_          = other.tpus_;
    input_shapes_  = other.input_shapes_;
    output_shapes_ = other.output_shapes_;
    max_input_shapes_single_  = other.max_input_shapes_single_;
    max_output_shapes_single_ = other.max_output_shapes_single_;
    dynamic_output_shapes_    = other.dynamic_output_shapes_;
    // dynamic_ must be assigned after free_pointers_sys()
    free_pointers();
    dynamic_ = other.dynamic_;
    graphs_  = other.graphs_;
    p_bmrts_ = other.p_bmrts_;
  }
  return *this;
}

template <typename Dtype>
void Inference<Dtype>::free_pointers() {
  // Destroy bmruntime pointers
  for (auto pbmrt : p_bmrts_) {
    destroy_bmruntime(pbmrt);
    pbmrt = NULL;
  }
  // free graphs, and device memories will be freed by deconstructor of Graph
  for (auto g : graphs_) {
    for (auto i : g.second) {
      // call destructor of Graph
      i.reset();
    }
  }
  // free system memories of input tensors
  for (auto t : input_tensors_) {
    for (auto i : t.second) {
      free(i.second);
      i.second = NULL;
    }
  }
  // free system memories of output tensors
  for (auto t : output_tensors_) {
    for (auto i : t.second) {
      free(i.second);
      i.second = NULL;
    }
  }
  // free system memories of dynamic output tensors
  if (dynamic_) {
    for (auto t : dynamic_output_tensors_) {
      for (auto i : t.second) {
        free(i.second);
        i.second = NULL;
      }
    }
  }
}

template <typename Dtype>
Inference<Dtype>::~Inference() {
  free_pointers();
}

template <typename Dtype>
BmRuntime Inference<Dtype>::get_bmrt() {
  BmRuntime bmrt(p_bmrts_[0]);
  return bmrt;
}

template <typename Dtype>
set<string> Inference<Dtype>::get_all_graph_names() {
  set<string> graph_names;
  int graph_num = bmrt_get_network_number(p_bmrts_[0]);
  const char** names;
  bmrt_get_network_names(p_bmrts_[0], &names);
  for (int i = 0; i < graph_num; ++i) {
    graph_names.insert(names[i]);
  }
  free(names);
  return graph_names;
}

template <typename Dtype>
set<string> Inference<Dtype>::get_input_tensor_names(const string& graph_name) {
  set<string> tensor_names;
  int input_num;
  const char** input_names;
  bmrt_get_input_tensor(p_bmrts_[0], 0, graph_name.c_str(),
                        &input_num, &input_names);
  for (int i = 0; i < input_num; ++i) {
    tensor_names.insert(input_names[i]);
  }
  return tensor_names;
}

template <typename Dtype>
set<string> Inference<Dtype>::get_output_tensor_names(
    const string &graph_name) {
  set<string> tensor_names;
  int output_num;
  const char** output_names;
  bmrt_get_output_tensor(p_bmrts_[0], 0, graph_name.c_str(),
                         &output_num, &output_names);
  for (int i = 0; i < output_num; ++i) {
    tensor_names.insert(output_names[i]);
  }
  return tensor_names;
}

template <typename Dtype>
bool Inference<Dtype>::is_input_batch_size_equal(
    map<string, vector<int>>& input_shapes,
    int&                      batch_size) {
  bool ret = true;
  for (auto it = input_shapes.begin(); it != input_shapes.end(); ++it) {
    if(it == input_shapes.begin()) {
      batch_size = it->second[0];
      continue;
    }
    if (it->second[0] != batch_size) {
      ret = false;
      break;
    }
  }
  return ret;
}

template <typename Dtype>
int Inference<Dtype>::reshape(
    const string&                     graph_name,
    map<string, vector<int>>&         input_shapes,
    map<string, bool>&                input_flags,
    map<string, vector<vector<int>>>& input_shapes_v) {
  // graphs_[graph_name] does not exist while constructing at this step.
  auto it_find = graphs_.find(graph_name);
  bool flag = (it_find == graphs_.end());
  int batch_size = 0;
  bool batch_equal = is_input_batch_size_equal(input_shapes, batch_size);
  int bmrt_size = p_bmrts_.size();
  worker_num_ = batch_equal ? min(batch_size, bmrt_size) : 1;
  for (int i = 0; i < worker_num_; ++i) {
    map<string, vector<int>> input_shapes_i = input_shapes;
    for (auto it = input_shapes_i.begin(); it != input_shapes_i.end(); ++it) {
      it->second[0] = it->second[0] / worker_num_
                      + (i < it->second[0] % worker_num_);
      input_shapes_v[it->first].push_back(it->second);
    }
    if (i == 0 && is_input_shape_valid(graph_name, input_shapes_i) != 0) {
      return 1;
    }
    if (flag && dynamic_) {
      dynamic_output_shapes_[graph_name].push_back(
          max_output_shapes_single_[graph_name]);
    }
  }
  if (!flag) {
    for (auto it = input_shapes.begin(); it != input_shapes.end(); ++it) {
      input_flags[it->first] = shapes_are_equal(input_shapes[it->first],
                               input_shapes_[graph_name][it->first]);
      if (!input_flags[it->first]) {
        input_shapes_[graph_name][it->first] = input_shapes[it->first];
      }
    }
  }
  return 0;
}

template <typename Dtype>
void Inference<Dtype>::alloc_sys_data(
    const string&             graph_name,
    map<string, bool>&        input_flags) {
  // graphs_[graph_name] does not exist while constructing at this step.
  auto it_find = graphs_.find(graph_name);
  bool flag = (it_find == graphs_.end());
  for (auto it = input_shapes_[graph_name].begin(); it != input_shapes_[graph_name].end(); ++it) {
    if (io_mode_ == SYSI || io_mode_ == SYSIO) {
      int data_size = get_size_of_shape(it->second) * sizeof(Dtype);
      if (!flag && !input_flags[it->first]) {
        free(input_tensors_[graph_name][it->first]);
      }
      if (flag || !input_flags[it->first]) {
        input_tensors_[graph_name][it->first] =
            reinterpret_cast<Dtype*>(malloc(data_size));
      }
    } else if (flag) {
      input_tensors_[graph_name][it->first] = NULL;
    }
  }
  if (flag) {
    for (auto it = output_shapes_[graph_name].begin(); it != output_shapes_[graph_name].end(); ++it) {
      if (io_mode_ == SYSO || io_mode_ == SYSIO) {
        int data_size = get_size_of_shape(it->second) * sizeof(Dtype);
        output_tensors_[graph_name][it->first] =
            reinterpret_cast<Dtype*>(malloc(data_size));
        if (dynamic_) {
          dynamic_output_tensors_[graph_name][it->first] =
            reinterpret_cast<Dtype*>(malloc(data_size));
        }
      } else {
        output_tensors_[graph_name][it->first] = NULL;
      }
    }
  }
}

template <typename Dtype>
map<string, Dtype*> Inference<Dtype>::split_sys_data(
    map<string, Dtype*>&      tensors,
    map<string, vector<int>>& shapes,
    map<string, size_t>&      offsets) {
  if (io_mode_ == SYSI || io_mode_ == SYSIO) {
    map<string, Dtype*> ret;
    for (auto it = shapes.begin(); it != shapes.end(); ++it) {
      ret[it->first] = tensors[it->first] + offsets[it->first];
      offsets[it->first] += get_size_of_shape(it->second);
    }
    return ret;
  } else {
    return tensors;
  }
}

template <typename Dtype>
void Inference<Dtype>::alloc_dev_data(
    const string&                     graph_name,
    map<string, vector<vector<int>>>& input_shapes_v) {
  // graphs_[graph_name] does not exist while constructing before this step.
  auto it_find = graphs_.find(graph_name);
  bool flag = (it_find == graphs_.end());
  map<string, size_t> input_offsets;
  for (auto it = input_shapes_[graph_name].begin(); it != input_shapes_[graph_name].end(); ++it) {
    input_offsets[it->first] = 0;
  }
  map<string, size_t> output_offsets;
  if (flag) {
    for (auto it = output_shapes_[graph_name].begin(); it != output_shapes_[graph_name].end(); ++it) {
      output_offsets[it->first] = 0;
    }
  }
  vector<std::shared_ptr<Graph<Dtype>>> graphs;
  for (int i = 0; i < worker_num_; ++i) {
    map<string, vector<int>> input_shapes;
    for (auto it = input_shapes_v.begin(); it != input_shapes_v.end(); ++it) {
      input_shapes[it->first] = it->second[i];
    }
    map<string, Dtype*> input_tensors = split_sys_data(
        input_tensors_[graph_name], input_shapes, input_offsets);
    if (flag) {
      map<string, Dtype*> output_tensors = split_sys_data(
          output_tensors_[graph_name], max_output_shapes_single_[graph_name],
          output_offsets);
      std::shared_ptr<Graph<Dtype>> graph(new Graph<Dtype>(p_bmrts_[i], graph_name,
          io_mode_, dynamic_, async_, input_shapes,
          max_output_shapes_single_[graph_name],
          input_tensors, output_tensors));
      if (dynamic_) {
          graph->set_dynamic_output_shapes(
              &dynamic_output_shapes_[graph_name][i]);
      }
      graphs.push_back(graph);
    } else {
      graphs_[graph_name][i]->reshape(input_shapes, input_tensors, false);
    }
  }
  if (flag) {
    graphs_[graph_name] = graphs;
  }
}

template <typename Dtype>
void Inference<Dtype>::alloc_dev_data(
    const string&                     graph_name,
    map<string, vector<vector<int>>>& input_shapes_v,
    TensorMode                        mode) {
  map<string, size_t> input_offsets;
  for (auto it = input_shapes_[graph_name].begin(); it != input_shapes_[graph_name].end(); ++it) {
    input_offsets[it->first] = 0;
  }
  for (size_t i = 0; i < p_bmrts_.size(); ++i) {
    map<string, vector<int>> input_shapes;
    for (auto it = input_shapes_v.begin(); it != input_shapes_v.end(); ++it) {
      input_shapes[it->first] = it->second[i];
    }
    map<string, Dtype*> input_tensors = split_sys_data(
        input_tensors_[graph_name], input_shapes, input_offsets);
    graphs_[graph_name][i]->reshape(input_shapes, input_tensors, false, mode);
  }
}

template <typename Dtype>
void Inference<Dtype>::set_input_shape(
    const string&             graph_name,
    map<string, vector<int>>& input_shapes) {
  map<string, vector<vector<int>>> input_shapes_v;
  map<string, bool> input_flags;

  // For static models, input shapes and output shapes can not be changed.
  // For dynamic models, output shapes are comfirmed after launching. So system
  // memory and device memory for output tensors are allocated for the maximum
  // shapes. Just reset the shapes of input tensors for dynamic models.
  reshape(graph_name, input_shapes, input_flags, input_shapes_v);
  alloc_sys_data(graph_name, input_flags);
  alloc_dev_data(graph_name, input_shapes_v);
}

template <typename Dtype>
void Inference<Dtype>::set_input_shape(
    const string&             graph_name,
    map<string, vector<int>>& input_shapes,
    TensorMode                mode) {
  map<string, vector<vector<int>>> input_shapes_v;
  map<string, bool> input_flags;

  // For static models, input shapes and output shapes can not be changed.
  // For dynamic models, output shapes are comfirmed after launching. So system
  // memory and device memory for output tensors are allocated for the maximum
  // shapes. Just reset the shapes of input tensors for dynamic models.
  reshape(graph_name, input_shapes, input_flags, input_shapes_v);
  alloc_sys_data(graph_name, input_flags);
  alloc_dev_data(graph_name, input_shapes_v, mode);
}

template <typename Dtype>
map<string, vector<int>> Inference<Dtype>::get_max_input_shapes(
    const string& graph_name,
    bool          is_total) {
  map<string, vector<int>> input_shapes;
  set<string> input_names = get_input_tensor_names(graph_name);
  for (auto iter = input_names.begin(); iter != input_names.end(); ++iter) {
    vector<int> shape(8);
    int dim = bmrt_get_input_blob_max_shape(p_bmrts_[0], iter->c_str(),
        0, graph_name.c_str(), shape.data());
    shape.resize(dim);
    if (is_total) {
      shape[0] *= tpus_.size();
    }
    input_shapes[*iter] = shape;
  }
  return input_shapes;
}

template <typename Dtype>
vector<int> Inference<Dtype>::get_input_shape(
    const string& graph_name,
    const string& tensor_name) {
  return input_shapes_[graph_name][tensor_name];
}

template <typename Dtype>
vector<int> Inference<Dtype>::get_output_shape(
    const string& graph_name,
    const string& tensor_name) {
  return output_shapes_[graph_name][tensor_name];
}

template <typename Dtype>
Dtype* Inference<Dtype>::get_input_tensor_sys(
    const string& graph_name,
    const string& tensor_name) {
  if (io_mode_ == SYSI || io_mode_ == SYSIO) {
    return input_tensors_[graph_name][tensor_name];
  } else {
    spdlog::info("Can not get sys_data of input tensor when \
                  IOMode is not SYSI or SYSIO.");
    return NULL;
  }
}

template <typename Dtype>
Dtype* Inference<Dtype>::get_output_tensor_sys(
    const string& graph_name,
    const string& tensor_name) {
  if (io_mode_ == SYSO || io_mode_ == SYSIO) {
    if (dynamic_) {
      return dynamic_output_tensors_[graph_name][tensor_name];
    } else {
      return output_tensors_[graph_name][tensor_name];
    }
  } else {
    spdlog::info("Can not get sys_data of output tensor when \
                  IOMode is not SYSI or SYSIO.");
    return NULL;
  }
}

template <typename Dtype>
bm_device_mem_t* Inference<Dtype>::get_input_tensor_dev_mem(
    const string& graph_name,
    const string& tensor_name) {
  if (io_mode_ == SYSI || io_mode_ == SYSIO) {
    spdlog::info("Can not get dev_data of input tensor when \
                  IOMode is not SYSI or SYSIO.");
    return NULL;
  } else {
    return graphs_[graph_name][0]->get_tensor(tensor_name);
  }
}

template <typename Dtype>
Tensor<Dtype> Inference<Dtype>::get_input_tensor_dev(
    const string& graph_name,
    const string& tensor_name) {
  vector<int> shape = get_input_shape(graph_name, tensor_name);
  bm_device_mem_t* data = get_input_tensor_dev_mem(graph_name, tensor_name);
  Tensor<Dtype> tensor(p_bmrts_[0], shape, data);
  return tensor;
}

template <typename Dtype>
bm_device_mem_t* Inference<Dtype>::get_input_tensor_dev(
    const string& graph_name,
    const string& tensor_name,
    TensorMode    mode) {
  if (io_mode_ == SYSI || io_mode_ == SYSIO) {
    spdlog::info("Can not get dev_data of input tensor when \
                  IOMode is not SYSI or SYSIO.");
    return NULL;
  } else {
    return graphs_[graph_name][0]->get_tensor(tensor_name, mode);
  }
}

template <typename Dtype>
bm_device_mem_t* Inference<Dtype>::get_output_tensor_dev_mem(
    const string& graph_name,
    const string& tensor_name) {
  if (io_mode_ == SYSO || io_mode_ == SYSIO) {
    spdlog::info("Can not get sys_data of output tensor when \
                  IOMode is not SYSI or SYSIO.");
    return NULL;
  } else {
    return graphs_[graph_name][0]->get_tensor(tensor_name);
  }
}

template <typename Dtype>
Tensor<Dtype> Inference<Dtype>::get_output_tensor_dev(
    const string& graph_name,
    const string& tensor_name) {
  vector<int> shape = get_output_shape(graph_name, tensor_name);
  bm_device_mem_t* data = get_output_tensor_dev_mem(graph_name, tensor_name);
  Tensor<Dtype> tensor(p_bmrts_[0], shape, data);
  return tensor;
}

template <typename Dtype>
bm_device_mem_t* Inference<Dtype>::get_output_tensor_dev(
    const string& graph_name,
    const string& tensor_name,
    TensorMode    mode) {
  if (io_mode_ == SYSO || io_mode_ == SYSIO) {
    spdlog::info("Can not get sys_data of output tensor when \
                  IOMode is not SYSI or SYSIO.");
    return NULL;
  } else {
    return graphs_[graph_name][0]->get_tensor(tensor_name, mode);
  }
}

template <typename Dtype>
void* Inference<Dtype>::multi_process_sync(void* parameter) {
  MultiSync<Dtype>* param =
      reinterpret_cast<MultiSync<Dtype>*>(parameter);
  std::shared_ptr<Graph<Dtype>> graph = param->graph;
  graph->inference();
  return NULL;
}

template <typename Dtype>
vector<int> Inference<Dtype>::get_tpus(const string& tpus) {
  vector<int> tpu_ids;
  if (tpus.compare("all") == 0) {
    int count = 0;
    int ret = bm_dev_getcount(&count);
    ASSERT(BM_SUCCESS == ret);
    for (int i = 0; i < count; ++i) {
      tpu_ids.push_back(i);
    }
  } else if (!tpus.empty()) {
    size_t found = tpus.find_first_of(",");
    size_t start = 0;
    string token = tpus.substr(start, found - start);
    while (found != string::npos) {
      if (!token.empty()) {
        tpu_ids.push_back(atoi(token.c_str()));
      }
      start = found + 1;
      found = tpus.find_first_of(",", start);
      token = tpus.substr(start, found - start);
    }
    if (found == string::npos && !token.empty()) {
      tpu_ids.push_back(atoi(token.c_str()));
    }
    if (tpu_ids.size() > 1) {
      sort(tpu_ids.begin(), tpu_ids.end());
      auto it = unique(tpu_ids.begin(), tpu_ids.end());
      if (it != tpu_ids.end()) {
        tpu_ids.resize(distance(tpu_ids.begin(), it));
      }
    }
  }
  ASSERT(tpu_ids.size() > 0);
  return tpu_ids;
}

template <typename Dtype>
int Inference<Dtype>::init_bmrts(
    const string& context_dir,
    const string& chip_name,
    const string& tpus) {
  tpus_ = get_tpus(tpus);
  if (tpus_.size() == 0) {
    spdlog::error("No TPU is assigned.");
    return 1;
  }
  p_bmrts_.clear();
  for (size_t i = 0; i < tpus_.size(); ++i) {
    void* p_bmrt = create_bmruntime_per_device(
        NULL, chip_name.c_str(), tpus_[i]);
    if (!bmrt_load_context(p_bmrt, context_dir.c_str())) {
      spdlog::error("Load {} failed", context_dir);
      return 2;
    }
    p_bmrts_.push_back(p_bmrt);
  }
  return 0;
}

template <typename Dtype>
int Inference<Dtype>::is_input_shape_valid(
    const string&                   graph_name,
    const map<string, vector<int>>& input_shapes) {
  set<string> input_tensor_names = get_input_tensor_names(graph_name);
  for (auto iter = input_shapes.begin(); iter != input_shapes.end(); ++iter) {
    string tensor_name = iter->first;
    auto it = input_tensor_names.find(tensor_name);
    if (it == input_tensor_names.end()) {
      spdlog::error("Wrong input tensor name of {}: {}.", graph_name, tensor_name);
      return 1;
    }
    const vector<int>& tensor_shape = iter->second;
    vector<int>& max_shape = max_input_shapes_single_[graph_name][tensor_name];
    int dim = max_shape.size();
    if (!bmrt_can_batch_size_change(p_bmrts_[0], 0, graph_name.c_str()) &&
        tensor_shape[0] != max_shape[0]) {
      spdlog::error("Batch size not allowed to change: {}: {}.", graph_name, tensor_name);
      return 2;
    } else if (!bmrt_can_height_and_width_change(p_bmrts_[0],
        0, graph_name.c_str()) && dim ==4 && tensor_shape[2] != max_shape[2]) {
      spdlog::error("Height size not allowed to change: {}: {}.", graph_name, tensor_name);
      return 3;
    } else if (!bmrt_can_height_and_width_change(p_bmrts_[0],
        0, graph_name.c_str()) && dim ==4 && tensor_shape[3] != max_shape[3]) {
      spdlog::error("Width size not allowed to change: {}: {}.", graph_name, tensor_name);
      return 4;
    } else if (tensor_shape[0] > max_shape[0]) {
      spdlog::error("Batch size exceeded muximum: {}: {}.", graph_name, tensor_name);
      return 5;
    } else if (dim >= 3 && tensor_shape[2] > max_shape[2]) {
      spdlog::error("Height exceeded muximum: {}: {}.", graph_name, tensor_name);
      return 6;
    } else if (dim >= 4 && tensor_shape[3] > max_shape[3]) {
      spdlog::error("Width exceeded muximum: {}: {}.", graph_name, tensor_name);
      return 7;
    } else {
      return 0;
    }
  }
}

template <typename Dtype>
map<string, vector<int>> Inference<Dtype>::get_max_output_shapes(
    const string& graph_name,
    bool          is_total) {
  map<string, vector<int>> output_shapes;
  set<string> output_names = get_output_tensor_names(graph_name);
  for (auto iter = output_names.begin(); iter != output_names.end(); ++iter) {
    vector<int> shape(8);
    int dim = bmrt_get_output_blob_max_shape(p_bmrts_[0], iter->c_str(),
        0, graph_name.c_str(), shape.data());
    shape.resize(dim);
    if (is_total) {
      shape[0] *= tpus_.size();
    }
    output_shapes[*iter] = shape;
  }
  return output_shapes;
}

template <typename Dtype>
int Inference<Dtype>::init_graphs() {
  graphs_.clear();
  for (auto it = input_shapes_.begin(); it != input_shapes_.end(); ++it) {
    set_input_shape(it->first, input_shapes_[it->first]);
  }
  return 0;
}

template <typename Dtype>
void Inference<Dtype>::reset_input_tensors(
    const string&             graph_name,
    map<string, vector<int>>& input_shapes,
    map<string, Dtype*>&      input_tensors) {
  map<string, vector<vector<int>>> input_shapes_v;
  map<string, bool> input_flags;
  int ret = reshape(graph_name, input_shapes, input_flags, input_shapes_v);
  if (ret != 0) {
    return;
  }
  map<string, size_t> offsets;
  for (auto it = input_shapes_[graph_name].begin(); it != input_shapes_[graph_name].end(); ++it) {
    offsets[it->first] = 0;
  }
  for (int i = 0; i < worker_num_; ++i) {
    map<string, vector<int>> shapes;
    for (auto it = input_shapes_v.begin(); it != input_shapes_v.end(); ++it) {
      shapes[it->first] = it->second[i];
    }
    map<string, Dtype*> tensors = split_sys_data(input_tensors,
                                                 shapes, offsets);
    graphs_[graph_name][i]->reshape(shapes, tensors, true);
  }
}

template <typename Dtype>
void Inference<Dtype>::reset_input_tensors(
    const string&             graph_name,
    map<string, vector<int>>& input_shapes,
    map<string, Dtype*>&      input_tensors,
    TensorMode                mode) {
  map<string, vector<vector<int>>> input_shapes_v;
  map<string, bool> input_flags;
  int ret = reshape(graph_name, input_shapes, input_flags, input_shapes_v);
  if (ret != 0) {
    return;
  }
  map<string, size_t> offsets;
  for (auto it = input_shapes_[graph_name].begin(); it != input_shapes_[graph_name].end(); ++it) {
    offsets[it->first] = 0;
  }
  for (size_t i = 0; i < p_bmrts_.size(); ++i) {
    map<string, vector<int>> shapes;
    for (auto it = input_shapes_v.begin(); it != input_shapes_v.end(); ++it) {
      shapes[it->first] = it->second[i];
    }
    map<string, Dtype*> tensors = split_sys_data(
        input_tensors, shapes, offsets);
    graphs_[graph_name][i]->reshape(shapes, tensors, true, mode);
  }
}

template <typename Dtype>
void Inference<Dtype>::prepare_dynamic_outputs(const string& graph_name) {
  for (auto it = output_shapes_[graph_name].begin(); it != output_shapes_[graph_name].end(); ++it) {
    it->second = dynamic_output_shapes_[graph_name][0][it->first];
    for (int i = 1; i < worker_num_; ++i) {
      it->second[0] += dynamic_output_shapes_[graph_name][i][it->first][0];
    }
  }

  if (io_mode_ == SYSO || io_mode_ == SYSIO) {
    for (auto it = output_tensors_[graph_name].begin(); it != output_tensors_[graph_name].end(); ++it) {
      int offset = 0;
      int raw_offset = get_size_of_shape(
          max_output_shapes_single_[graph_name][it->first]);
      for (int i = 0; i < worker_num_; ++i) {
        int size = get_size_of_shape(
            dynamic_output_shapes_[graph_name][i][it->first]);
        memcpy(dynamic_output_tensors_[graph_name][it->first] + offset,
               output_tensors_[graph_name][it->first] + i * raw_offset,
               size * sizeof(Dtype));
        offset += size;
      }
    }
  }
}

// asynchronization version of process
template <typename Dtype>
void Inference<Dtype>::process(
    const string& graph_name,
    TensorMode    mode,
    AsyncStatus   status) {
  if (worker_num_ == 0) {
    return;
  } else if (worker_num_ == 1) {
    graphs_[graph_name][0]->inference(mode, status);
  } else {
    if (async_) {
      spdlog::error("Cannot run a model on multiple TPUs with asynchronous mode \
                    in this way. Please refer to src/demo/cpp/multi_async.cpp.");
      exit(-1);
    }
  }
  if (dynamic_) {
    prepare_dynamic_outputs(graph_name);
  }
}

// asynchronization version of process
template <typename Dtype>
void Inference<Dtype>::process(
    const string&             graph_name,
    map<string, vector<int>>& input_shapes,
    map<string, Dtype*>&      input_tensors,
    TensorMode                mode,
    AsyncStatus               status) {
  if (status != END) {
    reset_input_tensors(graph_name, input_shapes, input_tensors, mode);
  }
  if (worker_num_ == 0) {
    return;
  } else if (worker_num_ == 1) {
    graphs_[graph_name][0]->inference(mode, status);
  } else {
    if (async_) {
      spdlog::error("Cannot run a model on multiple TPUs with asynchronous mode \
                    in this way. Please refer to src/demo/cpp/multi_async.cpp.");
      exit(-1);
    }
  }
  if (dynamic_) {
    prepare_dynamic_outputs(graph_name);
  }
}

// synchronization version of process
template <typename Dtype>
void Inference<Dtype>::process(const string& graph_name) {
  if (worker_num_ == 0 ) {
    return;
  } else if (worker_num_ == 1 ) {
    graphs_[graph_name][0]->inference();
  } else {
    MultiSync<Dtype>* params = new MultiSync<Dtype>[worker_num_];
    for (int i = 0; i < worker_num_; ++i) {
      params[i].graph = graphs_[graph_name][i];
      if (pthread_create(&params[i].thread_id, NULL,
                         multi_process_sync,
                         &params[i])) {
        delete [] params;
        spdlog::error("Create thread failed.");
        return;
      }
    }
    for (int i = 0; i < worker_num_; ++i) {
      int ret = pthread_join(params[i].thread_id, NULL);
      if (ret != 0) {
        delete [] params;
        spdlog::error("Thread join failed.");
        return;
      }
    }
    delete [] params;
  }
  if (dynamic_) {
    prepare_dynamic_outputs(graph_name);
  }
}

// synchronization version of process
template <typename Dtype>
void Inference<Dtype>::process(
    const string&             graph_name,
    map<string, vector<int>>& input_shapes,
    map<string, Dtype*>&      input_tensors) {
  reset_input_tensors(graph_name, input_shapes, input_tensors);
  if (worker_num_ == 0) {
    return;
  } else if (worker_num_ == 1) {
    graphs_[graph_name][0]->inference();
  } else {
    MultiSync<Dtype>* params = new MultiSync<Dtype>[worker_num_];
    for (int i = 0; i < worker_num_; ++i) {
      params[i].graph = graphs_[graph_name][i];
      if (pthread_create(&params[i].thread_id, NULL,
                         multi_process_sync,
                         &params[i])) {
        delete [] params;
        spdlog::error("Create thread failed.");
        return;
      }
    }
    for (int i = 0; i < worker_num_; ++i) {
      int ret = pthread_join(params[i].thread_id, NULL);
      if (ret != 0) {
        delete [] params;
        spdlog::error("Thread join failed.");
        return;
      }
    }
    delete [] params;
  }
  if (dynamic_) {
    prepare_dynamic_outputs(graph_name);
  }
}

#ifdef PYTHON
template <typename Dtype>
bp::list Inference<Dtype>::get_all_graph_names_py() {
  set<string> graph_names = get_all_graph_names();
  bp::list ret = set_to_list(graph_names);
  return ret;
}

template <typename Dtype>
bp::list Inference<Dtype>::get_input_tensor_names_py(
    const string& graph_name) {
  set<string> tensor_names = get_input_tensor_names(graph_name);
  bp::list ret = set_to_list(tensor_names);
  return ret;
}

template <typename Dtype>
bp::list Inference<Dtype>::get_output_tensor_names_py(
    const string& graph_name) {
  set<string> tensor_names = get_output_tensor_names(graph_name);
  bp::list ret = set_to_list(tensor_names);
  return ret;
}

template <typename Dtype>
void Inference<Dtype>::set_input_shape_sync_py(
    const string& graph_name,
    bp::dict&     input_shapes) {
  map<string, vector<int>> shapes = extract_shapes(input_shapes);
  set_input_shape(graph_name, shapes);
}

template <typename Dtype>
void Inference<Dtype>::set_input_shape_async_py(
    const string& graph_name,
    bp::dict&     input_shapes,
    TensorMode    mode) {
  map<string, vector<int>> shapes = extract_shapes(input_shapes);
  set_input_shape(graph_name, shapes, mode);
}

template <typename Dtype>
bp::dict Inference<Dtype>::get_max_input_shapes_py(
    const string& graph_name) {
  map<string, vector<int>> shapes = get_max_input_shapes(graph_name, true);
  bp::dict ret;
  for (auto it = shapes.begin(); it != shapes.end(); ++it) {
    ret[it->first] = convert_shape(it->second);
  }
  return ret;
}

template <typename Dtype>
bpn::ndarray Inference<Dtype>::get_input_shape_py(
    const string& graph_name,
    const string& tensor_name) {
  vector<int> shape = get_input_shape(graph_name, tensor_name);
  bpn::ndarray ret = convert_shape(shape);
  return ret;
}

template <typename Dtype>
bpn::ndarray Inference<Dtype>::get_output_shape_py(
    const string& graph_name,
    const string& tensor_name) {
  vector<int> shape = get_output_shape(graph_name, tensor_name);
  bpn::ndarray ret = convert_shape(shape);
  return ret;
}

template <typename Dtype>
void Inference<Dtype>::set_input_tensor_sys_py(
    const string& graph_name,
    const string& tensor_name,
    bpn::ndarray& data) {
  if (io_mode_ != SYSI && io_mode_ != SYSIO) {
    spdlog::error("Can not set sys_data of input tensor \
                  when IOMode is not SYSI or SYSIO.");
    return;
  }
  Dtype* data_ptr = reinterpret_cast<Dtype*>(data.get_data());
  Dtype* tensor = get_input_tensor_sys(graph_name, tensor_name);
  vector<int> shape = input_shapes_[graph_name][tensor_name];
  int size = get_size_of_shape(shape) * sizeof(Dtype);
  memcpy(tensor, data_ptr, size);
}

template <typename Dtype>
bpn::ndarray Inference<Dtype>::get_input_tensor_sys_py(
    const string& graph_name,
    const string& tensor_name) {
  Dtype* tensor = get_input_tensor_sys(graph_name, tensor_name);
  vector<int> shape = input_shapes_[graph_name][tensor_name];
  bpn::ndarray ret = convert_to_ndarray(tensor, shape);
  return ret;
}

template <typename Dtype>
bpn::ndarray Inference<Dtype>::get_output_tensor_sys_py(
    const string& graph_name,
    const string& tensor_name) {
  Dtype* tensor = get_output_tensor_sys(graph_name, tensor_name);
  vector<int> shape = output_shapes_[graph_name][tensor_name];
  bpn::ndarray ret = convert_to_ndarray(tensor, shape);
  return ret;
}

template <typename Dtype>
Tensor<Dtype> Inference<Dtype>::get_input_tensor_dev_py(
    const string& graph_name,
    const string& tensor_name) {
  vector<int> shape = get_input_shape(graph_name, tensor_name);
  bm_device_mem_t* dev_mem =
      get_input_tensor_dev_mem(graph_name, tensor_name);
  Tensor<Dtype> ret(p_bmrts_[0], shape, false, false,
                    false, NULL, dev_mem, NULL);
  return ret;
}

template <typename Dtype>
Tensor<Dtype> Inference<Dtype>::get_input_tensor_dev_py(
    const string& graph_name,
    const string& tensor_name,
    TensorMode    mode) {
  vector<int> shape = get_input_shape(graph_name, tensor_name);
  bm_device_mem_t* dev_mem =
      get_input_tensor_dev(graph_name, tensor_name, mode);
  Tensor<Dtype> ret(p_bmrts_[0], shape, false, false,
                    false, NULL, dev_mem, NULL);
  return ret;
}

template <typename Dtype>
Tensor<Dtype> Inference<Dtype>::get_output_tensor_dev_py(
    const string& graph_name,
    const string& tensor_name) {
  vector<int> shape = get_output_shape(graph_name, tensor_name);
  bm_device_mem_t* dev_mem =
      get_output_tensor_dev_mem(graph_name, tensor_name);
  Tensor<Dtype> ret(p_bmrts_[0], shape, false, false,
                    false, NULL, dev_mem, NULL);
  return ret;
}

template <typename Dtype>
Tensor<Dtype> Inference<Dtype>::get_output_tensor_dev_py(
    const string& graph_name,
    const string& tensor_name,
    TensorMode    mode) {
  vector<int> shape = get_output_shape(graph_name, tensor_name);
  bm_device_mem_t* dev_mem =
      get_output_tensor_dev(graph_name, tensor_name, mode);
  Tensor<Dtype> ret(p_bmrts_[0], shape, false, false,
                    false, NULL, dev_mem, NULL);
  return ret;
}

template <typename Dtype>
bp::dict Inference<Dtype>::process_sync(
    const string& graph_name,
    bp::dict&     input_tensors) {
  map<string, vector<int>> shapes;
  map<string, Dtype*> tensors;
  bp::list names = input_tensors.keys();
  int size = bp::len(names);
  for (int i = 0; i < size; ++i) {
    string name = static_cast<string>(bp::extract<string>(names[i]));
    bpn::ndarray data = bp::extract<bpn::ndarray>(input_tensors.get(name));
    int dimension = data.get_nd();
    vector<int> shape(dimension);
    for (int j = 0; j < dimension; ++j) {
      shape[j] = data.shape(j);
    }
    shapes[name] = shape;
    Dtype* data_ptr = reinterpret_cast<Dtype*>(data.get_data());
    tensors[name] = data_ptr;
  }
  process(graph_name, shapes, tensors);
  bp::dict ret;
  shapes = output_shapes_[graph_name];
  for (auto it = shapes.begin(); it != shapes.end(); ++it) {
    Dtype* tensor = get_output_tensor_sys(graph_name, it->first);
    ret[it->first] = convert_to_ndarray(tensor, it->second);
  }
  return ret;
}

template <typename Dtype>
bp::dict Inference<Dtype>::process_async(
    const string& graph_name,
    bp::dict&     input_tensors,
    TensorMode    mode,
    AsyncStatus   status) {
  map<string, vector<int>> shapes;
  map<string, Dtype*> tensors;
  if (status != END) {
    bp::list names = input_tensors.keys();
    int size = bp::len(names);
    for (int i = 0; i < size; ++i) {
      string name = static_cast<string>(bp::extract<string>(names[i]));
      bpn::ndarray data = bp::extract<bpn::ndarray>(input_tensors.get(name));
      int dimension = data.get_nd();
      vector<int> shape(dimension);
      for (int j = 0; j < dimension; ++j) {
        shape[j] = data.shape(j);
      }
      shapes[name] = shape;
      Dtype* data_ptr = reinterpret_cast<Dtype*>(data.get_data());
      tensors[name] = data_ptr;
    }
  }
  process(graph_name, shapes, tensors, mode, status);
  bp::dict ret;
  if (status != BEGIN) {
    map<string, vector<int>> shapes = output_shapes_[graph_name];
    for (auto it = shapes.begin(); it != shapes.end(); ++it) {
      Dtype* tensor = get_output_tensor_sys(graph_name, it->first);
      ret[it->first] = convert_to_ndarray(tensor, it->second);
    }
  }
  return ret;
}

#endif

template class Inference<float>;

}  // namespace sail
