/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <iostream>
#include <cstring>
#include "core/common.hpp"
#include "core/utils.hpp"
#include "core/tensor.hpp"

namespace sail {

template <typename Dtype>
Tensor<Dtype>::Tensor(const vector<int>& shape)
    : p_bmrt_(NULL), shape_(shape), own_sys_data_(true),
      own_dev_data_(false), async_(false), sys_data_(NULL),
      dev_data_ping_(NULL), dev_data_pong_(NULL) {
  if (shape_is_valid(shape)) {
    int data_size = get_size_of_shape(shape_) * sizeof(Dtype);
    sys_data_ = reinterpret_cast<Dtype*>(malloc(data_size));
  }
}

template <typename Dtype>
Tensor<Dtype>::Tensor(
    void*              p_bmrt,
    const vector<int>& shape)
    : p_bmrt_(p_bmrt), shape_(shape), own_sys_data_(false),
      own_dev_data_(true), async_(false), sys_data_(NULL),
      dev_data_ping_(NULL), dev_data_pong_(NULL) {
  if (shape_is_valid(shape)) {
    dev_data_ping_ = new bm_device_mem_t;
    int data_size = get_size_of_shape(shape_) * sizeof(Dtype);
    malloc_dev_data(PING, data_size);
  }
}

template <typename Dtype>
Tensor<Dtype>::Tensor(
    void*              p_bmrt,
    const vector<int>& shape,
    int                bytes)
    : p_bmrt_(p_bmrt), shape_(shape), own_sys_data_(false),
      own_dev_data_(true), async_(false), sys_data_(NULL),
      dev_data_ping_(NULL), dev_data_pong_(NULL) {
  dev_data_ping_ = new bm_device_mem_t;
  malloc_dev_data(PING, bytes);
}

template <typename Dtype>
Tensor<Dtype>::Tensor(const Tensor& other) {
  p_bmrt_ = other.p_bmrt_;
  shape_ = other.shape_;
  sys_data_ = other.sys_data_;
  own_sys_data_ = other.own_sys_data_;
  own_dev_data_ = other.own_dev_data_;
  async_ = other.async_;
  dev_data_ping_ = other.dev_data_ping_;
  dev_data_pong_ = other.dev_data_pong_;
}

template <typename Dtype>
Tensor<Dtype>& Tensor<Dtype>::operator=(const Tensor& other){
  if (this != &other) {
    free_sys_data();
    free_dev_data_and_struct();
    p_bmrt_ = other.p_bmrt_;
    shape_ = other.shape_;
    sys_data_ = other.sys_data_;
    own_sys_data_ = other.own_sys_data_;
    own_dev_data_ = other.own_dev_data_;
    async_ = other.async_;
    dev_data_ping_ = other.dev_data_ping_;
    dev_data_pong_ = other.dev_data_pong_;
  }
  return *this;
}

#ifdef PYTHON
template <typename Dtype>
Tensor<Dtype>::Tensor(bpn::ndarray& shape)
    : p_bmrt_(NULL), own_sys_data_(true),
      own_dev_data_(false), async_(false), sys_data_(NULL),
      dev_data_ping_(NULL), dev_data_pong_(NULL) {
  shape_ = extract_shape(shape);
  if (shape_is_valid(shape_)) {
    int data_size = get_size_of_shape(shape_) * sizeof(Dtype);
    sys_data_ = reinterpret_cast<Dtype*>(malloc(data_size));
  }
}

template <typename Dtype>
Tensor<Dtype>::Tensor(
    BmRuntime     bmrt,
    bpn::ndarray& shape)
    : p_bmrt_(bmrt.get_ptr()), own_sys_data_(false),
      own_dev_data_(true), async_(false), sys_data_(NULL),
      dev_data_ping_(NULL), dev_data_pong_(NULL) {
  shape_ = extract_shape(shape);
  if (shape_is_valid(shape_)) {
    dev_data_ping_ = new bm_device_mem_t;
    int data_size = get_size_of_shape(shape_) * sizeof(Dtype);
    malloc_dev_data(PING, data_size);
  }
}

template <typename Dtype>
bpn::ndarray Tensor<Dtype>::shape_py() {
  bpn::ndarray ret = convert_shape(shape_);
  return ret;
}

template <typename Dtype>
bpn::ndarray Tensor<Dtype>::sys_data_py() {
  bpn::ndarray ret = convert_to_ndarray(sys_data_, shape_);
  return ret;
}
#endif

template <typename Dtype>
Tensor<Dtype>::Tensor(
    void*                p_bmrt,
    const vector<int>&   shape,
    bool                 own_sys_data,
    bool                 own_dev_data,
    bool                 async,
    Dtype*               data,
    bm_device_mem_t*     dev_data_ping,
    bm_device_mem_t*     dev_data_pong)
    : p_bmrt_(p_bmrt), shape_(shape), own_sys_data_(own_sys_data),
      own_dev_data_(own_dev_data), async_(async), sys_data_(data),
      dev_data_ping_(dev_data_ping), dev_data_pong_(dev_data_pong) {
  if (shape_is_valid(shape)) {
    int data_size = get_size_of_shape(shape_) * sizeof(Dtype);
    if (own_sys_data_) {
      sys_data_ = reinterpret_cast<Dtype*>(malloc(data_size));
    }
    if (own_dev_data_) {
      dev_data_ping_ = new bm_device_mem_t;
      if (async_) {
        dev_data_pong_ = new bm_device_mem_t;
      }
    }
    malloc_dev_data(PING, data_size);
    if (async_) {
      malloc_dev_data(PONG, data_size);
    }
  }
}

template <typename Dtype>
Tensor<Dtype>::Tensor(
    void*                p_bmrt,
    const vector<int>&   shape,
    bm_device_mem_t*     data)
    : p_bmrt_(p_bmrt), shape_(shape), own_sys_data_(false),
      own_dev_data_(false), async_(false), sys_data_(NULL),
      dev_data_ping_(data), dev_data_pong_(NULL) {
  if (!shape_is_valid(shape)) {
    spdlog::error("Invaild tensor shape!");
  }
}

template <typename Dtype>
void Tensor<Dtype>::malloc_dev_data(TensorMode mode, int data_size) {
  if (mode == PING && dev_data_ping_ && data_size > 0) {
    bmrt_malloc_device_byte(p_bmrt_, dev_data_ping_, data_size);
  } else if (mode == PONG && dev_data_pong_ && data_size > 0) {
    bmrt_malloc_device_byte(p_bmrt_, dev_data_pong_, data_size);
  }
}

template <typename Dtype>
void Tensor<Dtype>::free_dev_data(TensorMode mode) {
  if (mode == PING && dev_data_ping_) {
    bmrt_free_device(p_bmrt_, *dev_data_ping_);
  } else if (mode == PONG && dev_data_pong_) {
    bmrt_free_device(p_bmrt_, *dev_data_pong_);
  }
}

template <typename Dtype>
void Tensor<Dtype>::free_dev_data() {
  free_dev_data(PING);
  free_dev_data(PONG);
}

template <typename Dtype>
void Tensor<Dtype>::free_sys_data() {
  if (own_sys_data_ && sys_data_) {
    delete [] sys_data_;
    sys_data_ = NULL;
  }
}

template <typename Dtype>
void Tensor<Dtype>::free_dev_data_and_struct() {
  free_dev_data();
  if (own_dev_data_ && dev_data_ping_ != NULL) {
    delete dev_data_ping_;
    dev_data_ping_ = NULL;
  }
  if (own_dev_data_ && dev_data_pong_ != NULL) {
    delete dev_data_pong_;
    dev_data_pong_ = NULL;
  }
}

template <typename Dtype>
Tensor<Dtype>::~Tensor() {
  free_sys_data();
  free_dev_data_and_struct();
}

template <typename Dtype>
const vector<int>& Tensor<Dtype>::shape() const {
  return shape_;
}

template <typename Dtype>
void Tensor<Dtype>::reshape(const vector<int>& shape) {
  if (!shape_is_valid(shape)) {
    spdlog::error("Invaild tensor shape!");
    return;
  }
  if (get_size_of_shape(shape_) != get_size_of_shape(shape)) {
    int data_size = get_size_of_shape(shape) * sizeof(Dtype);
    if (own_sys_data_) {
      free_sys_data();
      sys_data_ = reinterpret_cast<Dtype*>(malloc(data_size));
    }

    free_dev_data();
    malloc_dev_data(PING, data_size);
    if (async_) {
      malloc_dev_data(PONG, data_size);
    }
  }
  shape_ = shape;
}

template <typename Dtype>
void Tensor<Dtype>::reshape(const vector<int>& shape, TensorMode mode) {
  if (!shape_is_valid(shape)) {
    spdlog::error("Invaild tensor shape!");
    return;
  }
  if (get_size_of_shape(shape_) != get_size_of_shape(shape)) {
    int data_size = get_size_of_shape(shape) * sizeof(Dtype);
    if (own_sys_data_) {
      free_sys_data();
      sys_data_ = reinterpret_cast<Dtype*>(malloc(data_size));
    }

    free_dev_data(mode);
    malloc_dev_data(mode, data_size);
  }
  shape_ = shape;
}

template <typename Dtype>
void Tensor<Dtype>::reset_sys_data(Dtype* data) {
  if (own_sys_data_) {
    spdlog::error("Cannot reset_sys_data when own_sys_data is true.");
  } else {
    sys_data_ = data;
  }
}

template <typename Dtype>
void Tensor<Dtype>::reset_dev_data(bm_device_mem_t* data) {
  if (own_dev_data_) {
    spdlog::error("Cannot reset_dev_data when own_dev_data is true.");
  } else {
    dev_data_ping_ = data;
  }
}

template <typename Dtype>
void Tensor<Dtype>::reset_dev_data(
    bm_device_mem_t* data_ping,
    bm_device_mem_t* data_pong) {
  if (own_dev_data_) {
    spdlog::error("Cannot reset_dev_data when own_dev_data is true.");
  } else {
    dev_data_ping_ = data_ping;
    dev_data_pong_ = data_pong;
  }
}

template <typename Dtype>
bool Tensor<Dtype>::own_sys_data() {
  return own_sys_data_;
}

template <typename Dtype>
bool Tensor<Dtype>::own_dev_data() {
  return own_dev_data_;
}

template <typename Dtype>
bm_device_mem_t* Tensor<Dtype>::dev_data(TensorMode mode) {
  if (!async_) {
    spdlog::error("Can not specify mode when async_ is false.");
    return NULL;
  } else {
    return (mode == PING) ? dev_data_ping_ : dev_data_pong_;
  }
}

template <typename Dtype>
bm_device_mem_t* Tensor<Dtype>::dev_data() {
  if (async_) {
    spdlog::error("Please call dev_data(mode) when async_ is true.");
    return NULL;
  } else {
    return dev_data_ping_;
  }
}

template <typename Dtype>
Dtype* Tensor<Dtype>::sys_data() {
  return sys_data_;
}

template <typename Dtype>
void Tensor<Dtype>::sync_s2d() {
  if (async_) {
    spdlog::error("Please call sync_s2d(mode) when async_ is true.");
    return;
  }
  if (dev_data_ping_ && sys_data_) {
    bmrt_memcpy_s2d(p_bmrt_, *dev_data_ping_, sys_data_);
  }
}

template <typename Dtype>
void Tensor<Dtype>::sync_d2s() {
  if (async_) {
    spdlog::error("Please call sync_d2s(mode) when async_ is true.");
    return;
  }
  if (dev_data_ping_ != NULL && sys_data_ != NULL) {
    bmrt_memcpy_d2s(p_bmrt_, sys_data_, *dev_data_ping_);
  }
}

template <typename Dtype>
void Tensor<Dtype>::sync_d2s(int size) {
  if (async_) {
    spdlog::error("Please call sync_d2s(mode) when async_ is true.");
    return;
  }
  if (dev_data_ping_ != NULL && sys_data_ != NULL) {
    bmrt_memcpy_d2s_withsize(p_bmrt_, sys_data_, *dev_data_ping_, size);
  }
}

template <typename Dtype>
void Tensor<Dtype>::sync_s2d(TensorMode mode) {
  if (!async_) {
    spdlog::error("Please call sync_s2d() when async_ is false.");
    return;
  }
  switch (mode) {
    case PING:
      if (dev_data_ping_ != NULL && sys_data_!= NULL) {
        bmrt_memcpy_s2d(p_bmrt_, *dev_data_ping_, sys_data_);
      }
      break;
    case PONG:
      if (dev_data_pong_ != NULL && sys_data_ != NULL) {
        bmrt_memcpy_s2d(p_bmrt_, *dev_data_pong_, sys_data_);
      }
      break;
  }
}

template <typename Dtype>
void Tensor<Dtype>::sync_d2s(TensorMode mode) {
  if (!async_) {
    spdlog::error("Please call sync_d2s() when async_ is false.");
    return;
  }
  switch (mode) {
    case PING:
      if (dev_data_ping_ != NULL && sys_data_!= NULL) {
        bmrt_memcpy_d2s(p_bmrt_, sys_data_, *dev_data_ping_);
      }
      break;
    case PONG:
      if (dev_data_pong_ != NULL && sys_data_ != NULL) {
        bmrt_memcpy_d2s(p_bmrt_, sys_data_, *dev_data_pong_);
      }
      break;
  }
}

template <typename Dtype>
void Tensor<Dtype>::sync_d2s(TensorMode mode, int size) {
  if (!async_) {
    spdlog::error("Please call sync_d2s() when async_ is false.");
    return;
  }
  switch (mode) {
    case PING:
      if (dev_data_ping_ != NULL && sys_data_!= NULL) {
        bmrt_memcpy_d2s_withsize(p_bmrt_, sys_data_, *dev_data_ping_, size);
      }
      break;
    case PONG:
      if (dev_data_pong_ != NULL && sys_data_ != NULL) {
        bmrt_memcpy_d2s_withsize(p_bmrt_, sys_data_, *dev_data_pong_, size);
      }
      break;
  }
}

template <typename Dtype>
bool Tensor<Dtype>::copy_to(Tensor<Dtype>& dst) {
  if (!shapes_are_equal(shape_, dst.shape())) {
    spdlog::error("Error: ensor shapes are not equal!");
    return false;
  }
  bool ret = true;
  int size = get_size_of_shape(shape_) * sizeof(Dtype);
  Dtype* dst_sys_data = dst.sys_data();
  bm_device_mem_t* dst_dev_data_ping = dst.dev_data(PING);
  bm_device_mem_t* dst_dev_data_pong = dst.dev_data(PONG);
  if (dst_sys_data != NULL) {
    if (sys_data_ != NULL) {
      memcpy(dst_sys_data, sys_data_, size);
    } else if (dev_data_ping_ != NULL) {
      bmrt_memcpy_d2s(p_bmrt_, dst_sys_data, *dev_data_ping_);
    }
  }
  if (dst_dev_data_ping != NULL) {
    if (dst_sys_data != NULL) {
      bmrt_memcpy_s2d(p_bmrt_, *dst_dev_data_ping, dst_sys_data);
    } else {
      ret = false;
    }
  }
  if (dst_dev_data_pong != NULL) {
    if (dst_sys_data != NULL) {
      bmrt_memcpy_s2d(p_bmrt_, *dst_dev_data_pong, dst_sys_data);
    } else {
      ret = false;
    }
  }
  return ret;
}

template <typename Dtype>
bool Tensor<Dtype>::copy_from(Tensor<Dtype>& src) {
  if (!shapes_are_equal(shape_, src.shape())) {
    spdlog::error("Error: ensor shapes are not equal!");
    return false;
  }
  bool ret = true;
  int size = get_size_of_shape(shape_) * sizeof(Dtype);
  Dtype* src_sys_data = src.sys_data();
  bm_device_mem_t* src_dev_data_ping = src.dev_data(PING);
  bm_device_mem_t* src_dev_data_pong = src.dev_data(PONG);
  if (sys_data_ != NULL) {
    if (src_sys_data != NULL) {
      memcpy(sys_data_, src_sys_data, size);
    } else if (src_dev_data_ping != NULL) {
      bmrt_memcpy_d2s(p_bmrt_, sys_data_, *src_dev_data_ping);
    }
  }
  if (dev_data_ping_ != NULL) {
    if (src_sys_data != NULL) {
      bmrt_memcpy_s2d(p_bmrt_, *dev_data_ping_, src_sys_data);
    } else {
      ret = false;
    }
  }
  if (dev_data_pong_ != NULL) {
    if (src_sys_data != NULL) {
      bmrt_memcpy_s2d(p_bmrt_, *dev_data_pong_, src_sys_data);
    } else {
      ret = false;
    }
  }
  return ret;
}

template class Tensor<float>;

}  // namespace sail
