/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include <sys/stat.h>
#include <dirent.h>
#include <numeric>
#include <cstring>
#include <algorithm>
#include <fstream>
#include <iostream>
#include "core/common.hpp"
#include "core/utils.hpp"

namespace sail {

bool shape_is_valid(const vector<int>& shape) {
  if (shape.empty()) { return false; }
  if (std::any_of(shape.begin(), shape.end(), 
                  [](int i){return i<=0;})) {
    return false;
  }
  return true;
}

bool shapes_are_equal(
    const vector<int>& shape0,
    const vector<int>& shape1) {
  return shape0 == shape1;
}

int get_size_of_shape(const vector<int>& shape) {
  if (shape.empty()) { return 0; }
  return std::accumulate(shape.begin(), shape.end(),
                         1, std::multiplies<int>());
}

// TODO C++17 std::filesystem
vector<string> get_files_in_dir(const string& dir_path) {
  vector<string> files;
  struct stat s;
  lstat(dir_path.c_str(), &s);

  struct dirent* filename;    // return value for readdir()
  DIR* dir;                   // return value for opendir()
  dir = opendir(dir_path.c_str());

  /* read all the files in the dir ~ */
  while ((filename = readdir(dir)) != NULL) {
    // get rid of "." and ".."
    if (strcmp(filename->d_name , ".") == 0 ||
        strcmp(filename->d_name , "..") == 0) {
      continue;
    }
    files.push_back(filename -> d_name);
  }
  return files;
}

#ifdef PYTHON
vector<int> extract_shape(bpn::ndarray& shape) {
  int dimension = bp::len(shape);
  vector<int> ret(dimension);
  for (int i = 0; i < dimension; ++i) {
    ret[i] = bp::extract<int>(shape[i]);
  }
  return ret;
}

bpn::ndarray convert_shape(const vector<int>& shape) {
  bpn::ndarray ret = bpn::zeros(bp::make_tuple(shape.size()),
                                bpn::dtype::get_builtin<int>());
  for (size_t i = 0; i < shape.size(); ++i) {
    ret[i] = shape[i];
  }
  return ret;
}

map<string, vector<int>> extract_shapes(bp::dict& shapes) {
  map<string, vector<int>> ret;
  bp::list names = shapes.keys();
  int size = bp::len(names);
  for (int i = 0; i < size; ++i) {
    string key = static_cast<string>(bp::extract<string>(names[i]));
    bpn::ndarray raw_shape = bp::extract<bpn::ndarray>(shapes.get(key));
    ret[key] = extract_shape(raw_shape);
  }
  return ret;
}
#endif

}  // namespace sail
