/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include "process/postprocess.hpp"
#include <numeric>
#include <algorithm>
#include <iostream>
#include <cmath>
#include "core/common.hpp"

namespace sail {

MtcnnPostprocess::MtcnnPostprocess(double* threshold) {
  threshold_.assign(threshold, threshold + 3);
}

static bool CompareBBox(const FaceInfo& a, const FaceInfo& b) {
  return a.bbox.score > b.bbox.score;
}

void MtcnnPostprocess::GenerateBoundingBox(float* conf, float* coord,
    std::vector<FaceInfo>& condidate_rects,
    int width, int height,
    float scale, float thresh,
    int image_width, int image_height, int imgid) {
  int stride = 2;
  int cellSize = 12;
  // cout << scale << " " << thresh << " " << image_width << " "
  // << image_height << " " << width << " " << height << endl;
  int offset = height * width;
  const float* confidence_data = conf + offset;
  const float* reg_data = coord;
  condidate_rects.clear();
  for (int y = 0; y < height; ++y) {
    for (int x = 0; x < width; ++x) {
      int index = y * width + x;
      if (confidence_data[index] >= thresh) {
        float xTop = std::round((y * stride) / scale);
        float yTop = std::round((x * stride) / scale);
        float xBot = std::round((y * stride + cellSize - 1) / scale);
        float yBot = std::round((x * stride + cellSize - 1) / scale);
        FaceRect faceRect;
        faceRect.x1 = yTop;
        faceRect.y1 = xTop;
        faceRect.x2 = yBot;
        faceRect.y2 = xBot;
        // cout << "genera_data: " << x <<" " << y<<" "<<
        // faceRect.x1 << " " << faceRect.y1 << " " << faceRect.x2
        // << " " << faceRect.y2 << endl;
        faceRect.score  = confidence_data[index];
        FaceInfo faceInfo;
        faceInfo.bbox = faceRect;
        faceInfo.regression = cv::Vec4f(
            reg_data[offset + index],
            reg_data[index],
            reg_data[3 * offset + index],
            reg_data[2 * offset + index]);
        faceInfo.imgid = imgid;
        condidate_rects.push_back(faceInfo);
      }
    }
  }
}

std::vector<FaceInfo> MtcnnPostprocess::BoxRegress(
    std::vector<FaceInfo>& faceInfo, int stage) {
  std::vector<FaceInfo> bboxes;
  for (size_t bboxId = 0; bboxId < faceInfo.size(); ++bboxId) {
    FaceRect faceRect;
    FaceInfo tempFaceInfo;
    float regw = faceInfo[bboxId].bbox.x2 - faceInfo[bboxId].bbox.x1 + 1;
    float regh = faceInfo[bboxId].bbox.y2 - faceInfo[bboxId].bbox.y1 + 1;
    faceRect.x1 = faceInfo[bboxId].bbox.x1 + regw *
            faceInfo[bboxId].regression[0] - 1;
    faceRect.y1 = faceInfo[bboxId].bbox.y1 + regh *
            faceInfo[bboxId].regression[1] - 1;
    faceRect.x2 = faceInfo[bboxId].bbox.x2 + regw *
            faceInfo[bboxId].regression[2] - 1;
    faceRect.y2 = faceInfo[bboxId].bbox.y2 + regh *
            faceInfo[bboxId].regression[3] - 1;
    faceRect.score = faceInfo[bboxId].bbox.score;
    tempFaceInfo.bbox = faceRect;
    tempFaceInfo.regression = faceInfo[bboxId].regression;
    if (stage == 3)
        tempFaceInfo.facePts = faceInfo[bboxId].facePts;
    tempFaceInfo.imgid = faceInfo[bboxId].imgid;
    bboxes.push_back(tempFaceInfo);
  }
  return bboxes;
}

void MtcnnPostprocess::Bbox2Square(std::vector<FaceInfo>& bboxes) {
  for (size_t i = 0; i < bboxes.size(); ++i) {
    float w = bboxes[i].bbox.x2 - bboxes[i].bbox.x1 + 1;
    float h = bboxes[i].bbox.y2 - bboxes[i].bbox.y1 + 1;
    float side = std::max<float>(w, h);
    bboxes[i].bbox.x1 += (w - side) * 0.5;
    bboxes[i].bbox.y1 += (h - side) * 0.5;
    bboxes[i].bbox.x2 = std::round(bboxes[i].bbox.x1 + side - 1);
    bboxes[i].bbox.y2 = std::round(bboxes[i].bbox.y1 + side - 1);
    bboxes[i].bbox.x1 = std::round(bboxes[i].bbox.x1);
    bboxes[i].bbox.y1 = std::round(bboxes[i].bbox.y1);
  }
}

std::vector<FaceInfo> MtcnnPostprocess::MultiNMS(
    std::vector<FaceInfo>& BBoxes, float thresh, char methodType) {
  int imgsize = 1;
  std::vector<FaceInfo> bboxes_nms;
  for (int imgid = 0; imgid < imgsize; ++imgid) {
    std::vector<FaceInfo> bboxes;
    if (imgsize == 1) {
      bboxes = BBoxes;
    } else {
      for (size_t i = 0; i < BBoxes.size(); i++) {
        if (BBoxes[i].imgid == imgid)
          bboxes.push_back(BBoxes[i]);
      }
    }
    std::sort(bboxes.begin(), bboxes.end(), CompareBBox);
    int32_t select_idx = 0;
    int32_t num_bbox = static_cast<int32_t>(bboxes.size());
    std::vector<int32_t> mask_merged(num_bbox, 0);
    bool all_merged = false;
    while (!all_merged) {
      while (select_idx < num_bbox && mask_merged[select_idx] == 1)
        ++select_idx;
      if (select_idx == num_bbox) {
        all_merged = true;
        continue;
      }
      bboxes_nms.push_back(bboxes[select_idx]);
      mask_merged[select_idx] = 1;
      FaceRect select_bbox = bboxes[select_idx].bbox;
      float area1 = static_cast<float>(
          (select_bbox.x2 - select_bbox.x1 + 1) *
          (select_bbox.y2 - select_bbox.y1 + 1));
      float x1 = static_cast<float>(select_bbox.x1);
      float y1 = static_cast<float>(select_bbox.y1);
      float x2 = static_cast<float>(select_bbox.x2);
      float y2 = static_cast<float>(select_bbox.y2);
      ++select_idx;
      for (int32_t i = select_idx; i < num_bbox; i++) {
        if (mask_merged[i] == 1)
          continue;
        FaceRect &bbox_i = bboxes[i].bbox;
        float x = std::max<float>(x1, static_cast<float>(bbox_i.x1));
        float y = std::max<float>(y1, static_cast<float>(bbox_i.y1));
        float w = std::min<float>(x2, static_cast<float>(bbox_i.x2))
                  - x + 1;
        float h = std::min<float>(y2, static_cast<float>(bbox_i.y2))
                  - y + 1;
        if (w <= 0 || h <= 0)
          continue;
        float area2 = static_cast<float>(
                (bbox_i.x2 - bbox_i.x1 + 1) *
                (bbox_i.y2 - bbox_i.y1 + 1));
        float area_intersect = w * h;
        switch (methodType) {
          case 'u':
            if (static_cast<float>(area_intersect) /
                  (area1 + area2 - area_intersect) > thresh)
              mask_merged[i] = 1;
            break;
          case 'm':
            if (static_cast<float>(area_intersect) /
                  std::min(area1, area2) > thresh)
              mask_merged[i] = 1;
            break;
          default:
            break;
        }
      }
    }
  }
  return bboxes_nms;
}

void MtcnnPostprocess::Padding(
    cv::Mat& frame,
    std::vector<FaceInfo>& regressed_rects,
    std::vector<FaceInfo>& regressed_pading) {
  std::vector<cv::Mat> sample_singles;
  sample_singles.push_back(frame);
  for (size_t i = 0; i < regressed_rects.size(); ++i) {
    FaceInfo tempFaceInfo;
    tempFaceInfo = regressed_rects[i];
    int imgid = tempFaceInfo.imgid;
    int img_w = sample_singles[imgid].cols;
    int img_h = sample_singles[imgid].rows;
    tempFaceInfo.bbox.x1 = (regressed_rects[i].bbox.x1 < 0) ?
                0 : regressed_rects[i].bbox.x1;
    tempFaceInfo.bbox.y1 = (regressed_rects[i].bbox.y1 < 0) ?
                0 : regressed_rects[i].bbox.y1;
    tempFaceInfo.bbox.x2 = (regressed_rects[i].bbox.x2 > img_w - 1) ?
                img_w - 1: regressed_rects[i].bbox.x2;
    tempFaceInfo.bbox.y2 = (regressed_rects[i].bbox.y2 > img_h - 1) ?
                img_h - 1: regressed_rects[i].bbox.y2;
    regressed_pading.push_back(tempFaceInfo);
  }
}

std::vector<FaceInfo> MtcnnPostprocess::Pnet_process(
    float* conf, float* coord, float scale,
    int net_output_width, int net_output_height,
    int input_width, int input_height) {
  std::vector<FaceInfo> condidate_rects;
  GenerateBoundingBox(conf, coord, condidate_rects,
    net_output_width, net_output_height, scale,
    threshold_[0], input_width, input_height, 0);
  std::vector<FaceInfo> bboxes_nms = MultiNMS(condidate_rects,
    0.5, 'u');
  return bboxes_nms;
}

std::vector<FaceInfo> MtcnnPostprocess::Rnet_process(
    float* conf, float* coord,
    cv::Mat& frame,
    std::vector<FaceInfo>& regressed_rects) {
  const float* confidence_data = conf;
  const float* reg_data = coord;
  const float* points_data = NULL;
  std::vector<FaceInfo> condidate_rects;
  int numBox  = regressed_rects.size();
  double thresh = threshold_[1];
  int netidx = 1;
  for (int i = 0; i < numBox; i++) {
    if (confidence_data[i * 2 + 1] > thresh) {
      FaceRect faceRect;
      faceRect.x1 = regressed_rects[i].bbox.x1;
      faceRect.y1 = regressed_rects[i].bbox.y1;
      faceRect.x2 = regressed_rects[i].bbox.x2;
      faceRect.y2 = regressed_rects[i].bbox.y2;
      faceRect.score  = confidence_data[i * 2 + 1];
      FaceInfo faceInfo;
      faceInfo.bbox = faceRect;
      faceInfo.regression = cv::Vec4f(reg_data[4 * i + 1],
              reg_data[4 * i + 0], reg_data[4 * i + 3], reg_data[4 * i + 2]);
      faceInfo.imgid = regressed_rects[i].imgid;
      if (netidx == 2) {
        FacePts face_pts;
        float w = faceRect.x2 - faceRect.x1 + 1;
        float h = faceRect.y2 - faceRect.y1 + 1;
        for (int j = 0; j < 5; j++) {
            face_pts.x[j] = faceRect.x1
                + points_data[2 * j + 10 * i + 1] * w - 1;
            face_pts.y[j] = faceRect.y1
                + points_data[2 * j + 10 * i] * h - 1;
        }
        faceInfo.facePts = face_pts;
      }
      condidate_rects.push_back(faceInfo);
    }
  }
  condidate_rects = MultiNMS(condidate_rects, 0.7, 'u');
  regressed_rects = BoxRegress(condidate_rects, 2);
  Bbox2Square(regressed_rects);
  return regressed_rects;
}

std::vector<FaceInfo> MtcnnPostprocess::Onet_process(
    float* conf, float* coord, float* landmark,
    cv::Mat& frame,
    std::vector<FaceInfo>& regressed_rects) {
  const float* confidence_data = conf;
  const float* reg_data = coord;
  const float* points_data = landmark;
  std::vector<FaceInfo> condidate_rects;
  std::vector<FaceInfo> faceinfos;
  int numBox  = regressed_rects.size();
  double thresh = threshold_[2];
  int netidx = 2;
  for (int i = 0; i < numBox; i++) {
    if (confidence_data[i * 2 + 1] > thresh) {
      FaceRect faceRect;
      faceRect.x1 = regressed_rects[i].bbox.x1;
      faceRect.y1 = regressed_rects[i].bbox.y1;
      faceRect.x2 = regressed_rects[i].bbox.x2;
      faceRect.y2 = regressed_rects[i].bbox.y2;
      faceRect.score  = confidence_data[i * 2 + 1];
      FaceInfo faceInfo;
      faceInfo.bbox = faceRect;
      faceInfo.regression = cv::Vec4f(reg_data[4 * i + 1],
              reg_data[4 * i + 0], reg_data[4 * i + 3], reg_data[4 * i + 2]);
      faceInfo.imgid = regressed_rects[i].imgid;
      if (netidx == 2) {
        FacePts face_pts;
        float w = faceRect.x2 - faceRect.x1 + 1;
        float h = faceRect.y2 - faceRect.y1 + 1;
        for (int j = 0; j < 5; j++) {
            // face_pts.x[j] = faceRect.x1
            //     + points_data[2 * j + 10 * i + 1] * w - 1;
            // face_pts.y[j] = faceRect.y1
            //     + points_data[2 * j + 10 * i] * h - 1;
            face_pts.x[j] = faceRect.x1
                + points_data[j + 10 * i + 5] * w - 1;
            face_pts.y[j] = faceRect.y1
                + points_data[j + 10 * i] * h - 1;
        }
        faceInfo.facePts = face_pts;
      }
      condidate_rects.push_back(faceInfo);
    }
  }
  regressed_rects = BoxRegress(condidate_rects, 3);
  faceinfos = MultiNMS(regressed_rects, 0.7, 'm');
  return faceinfos;
}

MtcnnPostprocess::~MtcnnPostprocess() {}

}  // namespace sail
