/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include "process/postprocess.hpp"
#include <assert.h>
#include <stdlib.h>
#include <malloc.h>
#include <numeric>
#include <algorithm>
#include <iostream>
#include <cmath>
#include <string>
#include "core/common.hpp"
#include "core/statistics.hpp"

namespace sail {

void *traversal_confs(void *arg) {
  float threshold = (reinterpret_cast<TraversalArgs *>(arg))->threshold;
  int num_input_prior = (reinterpret_cast<TraversalArgs *>(arg))
      ->num_input_prior;
  int num_class = (reinterpret_cast<TraversalArgs *>(arg))->num_class;
  int *num_output_prior = &((reinterpret_cast<TraversalArgs *>(arg))
      ->num_output_prior);
  assert((*num_output_prior) == 0);
  float *input_conf = (reinterpret_cast<TraversalArgs *>(arg))->input_conf;
  float *input_loc = (reinterpret_cast<TraversalArgs *>(arg))->input_loc;
  int *output_index = (reinterpret_cast<TraversalArgs *>(arg))->output_index;
  float *output_conf = (reinterpret_cast<TraversalArgs *>(arg))->output_conf;
  float *output_box = (reinterpret_cast<TraversalArgs *>(arg))->output_box;
  int anchor_offset = (reinterpret_cast<TraversalArgs *>(arg))->anchor_offset;
  std::vector<std::vector<float>> *anchor_pointer =
      (reinterpret_cast<TraversalArgs *>(arg))->anchor_pointer;

  float anchor_x = 0.0f;
  float anchor_y = 0.0f;
  float anchor_w = 0.0f;
  float anchor_h = 0.0f;
  float delta_x = 0.0f;
  float delta_y = 0.0f;
  float delta_w = 0.0f;
  float delta_h = 0.0f;
  float new_x = 0.0f;
  float new_y = 0.0f;
  float new_w = 0.0f;
  float new_h = 0.0f;

  for (int i = 0; i < num_input_prior; i++) {
    float max_prob = 0.f;
    int max_index = -1;
    for (int j = 0; j < num_class; j++) {
      if (input_conf[i * num_class + j] > max_prob) {
        max_prob = input_conf[i * num_class + j];
        max_index = j;
      }
    }
    if (max_index > -1) {
      output_index[*num_output_prior] = max_index;
      output_conf[*num_output_prior] = max_prob;

      anchor_x = (*anchor_pointer)[i + anchor_offset][0];
      delta_y = input_loc[i * 4 + 0];
      anchor_y = (*anchor_pointer)[i + anchor_offset][1];
      delta_x = input_loc[i * 4 + 1];
      anchor_w = (*anchor_pointer)[i + anchor_offset][2];
      delta_h = input_loc[i * 4 + 2];
      anchor_h = (*anchor_pointer)[i + anchor_offset][3];
      delta_w = input_loc[i * 4 + 3];

      new_x = anchor_x + 0.1 * delta_x * anchor_w;
      new_y = anchor_y + 0.1 * delta_y * anchor_h;
      new_w = anchor_w * exp(0.2 * delta_w);
      new_h = anchor_h * exp(0.2 * delta_h);

      output_box[4 * (*num_output_prior) + 0] =
          min(max((new_x - new_w / 2.0f), 0.0f), 1.0f);
      output_box[4 * (*num_output_prior) + 1] =
          min(max((new_y - new_h / 2.0f), 0.0f), 1.0f);
      output_box[4 * (*num_output_prior) + 2] =
          min(max((new_x + new_w / 2.0f), 0.0f), 1.0f);
      output_box[4 * (*num_output_prior) + 3] =
          min(max((new_y + new_h / 2.0f), 0.0f), 1.0f);

      *num_output_prior = (*num_output_prior) + 1;
    }
  }
}

void *traversal_confs_v2(void *arg) {
  float threshold = (reinterpret_cast<TraversalArgsV2 *>(arg))->threshold;
  int num_thread = (reinterpret_cast<TraversalArgsV2 *>(arg))->num_thread;
  int id = (reinterpret_cast<TraversalArgsV2 *>(arg))->id_this_thread;
  int num_input_prior = (reinterpret_cast<TraversalArgsV2 *>(arg))
      ->num_input_prior;
  int num_class = (reinterpret_cast<TraversalArgsV2 *>(arg))->num_class;
  int *num_output_prior = &((reinterpret_cast<TraversalArgsV2 *>(arg))
      ->num_output_prior);
  assert((*num_output_prior) == 0);
  float *input_conf = (reinterpret_cast<TraversalArgsV2 *>(arg))->input_conf;
  float *input_loc = (reinterpret_cast<TraversalArgsV2 *>(arg))->input_loc;
  int *output_index = (reinterpret_cast<TraversalArgsV2 *>(arg))->output_index;
  float *output_conf = (reinterpret_cast<TraversalArgsV2 *>(arg))->output_conf;
  float *output_box = (reinterpret_cast<TraversalArgsV2 *>(arg))->output_box;
  std::vector<std::vector<float>> *anchor_pointer =
      (reinterpret_cast<TraversalArgsV2 *>(arg))->anchor_pointer;

  float anchor_x = 0.0f;
  float anchor_y = 0.0f;
  float anchor_w = 0.0f;
  float anchor_h = 0.0f;
  float delta_x = 0.0f;
  float delta_y = 0.0f;
  float delta_w = 0.0f;
  float delta_h = 0.0f;
  float new_x = 0.0f;
  float new_y = 0.0f;
  float new_w = 0.0f;
  float new_h = 0.0f;

  for (int i = 0; i < num_input_prior; i++) {
    float max_prob = 0.f;
    int max_index = -1;
    for (int j = 0; j < num_class; j++) {
      if (input_conf[i*num_thread*num_class + id*num_class + j] > max_prob) {
        max_prob = input_conf[i*num_thread*num_class + id*num_class + j];
        max_index = j;
      }
    }
    if (max_index > -1) {
      output_index[*num_output_prior] = max_index;
      output_conf[*num_output_prior] = max_prob;

      anchor_x = (*anchor_pointer)[i * num_thread + id][0];
      delta_y = input_loc[i * num_thread * 4 + id * 4 + 0];
      anchor_y = (*anchor_pointer)[i * num_thread + id][1];
      delta_x = input_loc[i * num_thread * 4 + id * 4 + 1];
      anchor_w = (*anchor_pointer)[i * num_thread + id][2];
      delta_h = input_loc[i * num_thread * 4 + id * 4 + 2];
      anchor_h = (*anchor_pointer)[i * num_thread + id][3];
      delta_w = input_loc[i * num_thread * 4 + id * 4 + 3];

      new_x = anchor_x + 0.1 * delta_x * anchor_w;
      new_y = anchor_y + 0.1 * delta_y * anchor_h;
      new_w = anchor_w * exp(0.2 * delta_w);
      new_h = anchor_h * exp(0.2 * delta_h);

      output_box[4 * (*num_output_prior) + 0] =
          min(max((new_x - new_w / 2.0f), 0.0f), 1.0f);
      output_box[4 * (*num_output_prior) + 1] =
          min(max((new_y - new_h / 2.0f), 0.0f), 1.0f);
      output_box[4 * (*num_output_prior) + 2] =
          min(max((new_x + new_w / 2.0f), 0.0f), 1.0f);
      output_box[4 * (*num_output_prior) + 3] =
          min(max((new_y + new_h / 2.0f), 0.0f), 1.0f);

      *num_output_prior = (*num_output_prior) + 1;
    }
  }
}

ImageNetPostprocess::ImageNetPostprocess(
    size_t batch_size,
    size_t class_num,
    size_t top_k)
    : batch_size_(batch_size), class_num_(class_num), top_k_(top_k) {
  if (top_k > class_num) {
    cout << "Error: top_k > class_num" << endl;
    throw;
  }
}

vector<vector<size_t>> ImageNetPostprocess::process(float *input) {
  vector<vector<size_t>> result;
  float *data = input;
  for (size_t i = 0; i < batch_size_; ++i) {
    // initialize original index locations
    vector<size_t> idx(class_num_);
    iota(idx.begin(), idx.end(), 0);
    // sort indexes based on comparing values in data
    sort(idx.begin(), idx.end(),
         [&data](size_t i1, size_t i2) { return data[i1] > data[i2]; });
    idx.resize(top_k_);
    result.push_back(idx);
    data += class_num_;
  }
  return result;
}

/*
ssd bmnetc
*/
SsdPostprocess::SsdPostprocess(
    size_t class_num,
    size_t prior_num,
    float threshold_conf,
    float threshold_nms)
    : class_num_(class_num), prior_num_(prior_num),
      threshold_conf_(threshold_conf), threshold_nms_(threshold_nms) {
}

float SsdPostprocess::lap(
    float x1_min,
    float x1_max,
    float x2_min,
    float x2_max) {
  return max(0.0f, min(x1_max, x2_max) - max(x1_min, x2_min));
}

float SsdPostprocess::jaccard(vector<float> x1, vector<float> x2) {
  float lap1 = lap(x1[0], x1[2], x2[0], x2[2]);
  float lap2 = lap(x1[1], x1[3], x2[1], x2[3]);
  if (lap1 < 1e-8 || lap2 < 1e-8) {
    return 0.0f;
  } else {
    return lap1 * lap2 / ((x1[2] - x1[0]) * (x1[3] - x1[1])
        + (x2[2] - x2[0]) * (x2[3] - x2[1]) - lap1 * lap2);
  }
}

vector<DetectRet> SsdPostprocess::process(
    float *conf_data,
    float *loc_data,
    float *prior_data) {
  int max_idx_ = 0;
  vector<DetectRet> result;
  vector<vector<float>> bboxes(prior_num_, vector<float>(4, 0));
  vector < pair < pair < int, int >, float > > idx_class_conf;
  for (int prior_idx = 0; prior_idx < prior_num_; ++prior_idx) {
    int idx = prior_idx * class_num_;
    int max_idx = -1;
    float max = 0.0;
    // class_idx = 0 is background
    for (int class_idx = 1; class_idx < class_num_; ++class_idx) {
      if (conf_data[idx + class_idx] > max) {
        max = conf_data[idx + class_idx];
        max_idx = class_idx;
      }
    }
    if (max > threshold_conf_) {
      idx_class_conf.push_back(make_pair(make_pair(prior_idx, max_idx),
                                         conf_data[idx + max_idx]));
    }
  }
  for (auto item : idx_class_conf) {
    int i = item.first.first;
    vector<float> temp;
    int iter_0 = i * 4;
    int iter_1 = iter_0 + 1;
    int iter_2 = iter_0 + 2;
    int iter_3 = iter_0 + 3;
    float prior_w = prior_data[iter_2] - prior_data[iter_0];
    float prior_h = prior_data[iter_3] - prior_data[iter_1];
    float prior_x = (prior_data[iter_2] + prior_data[iter_0]) / 2;
    float prior_y = (prior_data[iter_1] + prior_data[iter_3]) / 2;
    float bb_x = 0.1 * loc_data[iter_0] * prior_w + prior_x;
    float bb_y = 0.1 * loc_data[iter_1] * prior_h + prior_y;
    float bb_w = exp(0.2 * loc_data[iter_2]) * prior_w;
    float bb_h = exp(0.2 * loc_data[iter_3]) * prior_h;
    temp.push_back(min(1.0f, max((bb_x - bb_w / 2), 0.0f)));
    temp.push_back(min(1.0f, max((bb_y - bb_h / 2), 0.0f)));
    temp.push_back(min(1.0f, max((bb_x + bb_w / 2), 0.0f)));
    temp.push_back(min(1.0f, max((bb_y + bb_h / 2), 0.0f)));
    bboxes[i] = temp;
  }
  for (int i = 0; i < idx_class_conf.size(); ++i) {
    for (int j = idx_class_conf.size() - 1; j > i; --j) {
      if (idx_class_conf[i].first.second == idx_class_conf[j].first.second) {
        if (jaccard(bboxes[idx_class_conf[i].first.first],
                    bboxes[idx_class_conf[j].first.first]) > threshold_nms_) {
          if (idx_class_conf[i].second < idx_class_conf[j].second) {
            idx_class_conf.erase(idx_class_conf.begin() + i);
            --i;
            break;
          } else {
            idx_class_conf.erase(idx_class_conf.begin() + j);
          }
        }
      }
    }
  }
  for (int i = 0; i < idx_class_conf.size(); ++i) {
    vector<float> bbox = bboxes[idx_class_conf[i].first.first];
    DetectRet ret = {bbox[0], bbox[1], bbox[2], bbox[3],
                     idx_class_conf[i].first.second, idx_class_conf[i].second};
    result.push_back(ret);
  }
  return result;
}

/*
ssd tf
*/
SsdTfPostprocess::SsdTfPostprocess(size_t class_num,
    size_t prior_num,
    float threshold_conf,
    float threshold_nms,
    size_t max_box_num) : class_num_(class_num), prior_num_(prior_num),
      threshold_conf_(threshold_conf), threshold_nms_(threshold_nms),
      max_box_num_(max_box_num) {
  priors_.clear();
}

void SsdTfPostprocess::generate_multiscale_priors(std::vector<float>& levels,
                                              float img_side,
                                              std::vector<float>& scales,
                                              std::vector<float>& aspect_ratios,
                                              float offset_ratio,
                                              float anchor_scale) {
  assert(scales.size() == aspect_ratios.size());
  priors_.clear();
  for (size_t i = 0; i < levels.size(); i++) {
    float stride = pow(2.f, levels[i]);
    float offset = offset_ratio * stride;
    float base_anchor_side = anchor_scale * stride;
    std::vector<std::vector<float>> base_anchors;
    base_anchors.clear();
    for (size_t j = 0; j < scales.size(); j++) {
      float aspect = sqrt(aspect_ratios[j]);
      float scale = scales[j];
      std::vector<float> tmp_box;
      tmp_box.clear();
      tmp_box.push_back(offset);
      tmp_box.push_back(offset);
      tmp_box.push_back(base_anchor_side * scale * aspect);
      tmp_box.push_back(base_anchor_side * scale / aspect);
      base_anchors.push_back(tmp_box);
    }
    int feat_side = static_cast<int>(img_side / stride);
    for (int y = 0; y < feat_side; y++) {
      for (int x = 0; x < feat_side; x++) {
        for (size_t k = 0; k < base_anchors.size(); k++) {
          std::vector<float> tmp_box;
          tmp_box.clear();
          tmp_box.push_back((base_anchors[k][0] + x * stride) / img_side);
          tmp_box.push_back((base_anchors[k][1] + y * stride) / img_side);
          tmp_box.push_back(base_anchors[k][2] / img_side);
          tmp_box.push_back(base_anchors[k][3] / img_side);
          priors_.push_back(tmp_box);
        }
      }
    }
  }
}

// input_value will be changed
void SsdTfPostprocess::argsort_down(
    float *data,
    int *res,
    int start,
    int end) {
  if (start >= end) {
    return;
  }
  float center = data[start];
  int center_id = res[start];
  int i = start;
  int j = end;
  while (i < j) {
    while (i < j && data[j] <= center) j--;
    data[i] = data[j];
    res[i] = res[j];
    while (i < j && data[i] >= center) i++;
    data[j] = data[i];
    res[j] = res[i];
  }
  data[i] = center;
  res[i] = center_id;
  argsort_down(data, res, start, i - 1);
  argsort_down(data, res, i + 1, end);
}

float SsdTfPostprocess::area(float *box) {
  return (box[2] - box[0]) * (box[3] - box[1]);
}

// y1 x1 y2 x2
float SsdTfPostprocess::box_overlap(float *box1, float *box2) {
  float h = min(box1[2], box2[2]) - max(box1[0], box2[0]);
  float w = min(box1[3], box2[3]) - max(box1[1], box2[1]);
  if (w <= 0 || h <= 0) {
    return 0.0f;
  }
  return (w * h) / (area(box1) + area(box2) - w * h);
}

vector<DetectRet> SsdTfPostprocess::process(float *conf_data, float *loc_data) {
  if (priors_.size() == 0) {
    std::cout << "priors is empty, please "
              << "\"generate_multiscale_priors.\"" << std::endl;
    exit(-1);
  }

  float *boxes_after_get = reinterpret_cast<float *>(
      malloc(6 * prior_num_ * sizeof(float)));
  memset(boxes_after_get, 0, 6 * prior_num_ * sizeof(float));
  float *probs_after_get_for_sort = reinterpret_cast<float *>(
      malloc(prior_num_ * sizeof(float)));
  memset(probs_after_get_for_sort, 0, prior_num_ * sizeof(float));
  int *sorted_index = reinterpret_cast<int *>(
      malloc(prior_num_ * sizeof(int)));
  memset(sorted_index, 0, prior_num_ * sizeof(int));
  bool *if_retain_for_nms = reinterpret_cast<bool *>(
      malloc(prior_num_ * sizeof(bool)));
  memset(if_retain_for_nms, 1, prior_num_ * sizeof(bool));
  float *boxes_after_nms = reinterpret_cast<float *>(
      malloc(6 * prior_num_ * sizeof(float)));
  memset(boxes_after_nms, 0, 6 * prior_num_ * sizeof(float));

  int num_after_get = 0;
  int num_after_nms = 0;
  int i = 0;
  int j = 0;
  float conf_thresh = -log((1.0f - threshold_conf_) / threshold_conf_);

  for (i = 0; i < prior_num_; i++) {
    float max_prob = 0.0f;
    float max_label = -1.0f;

    float anchor_x = 0.0f;
    float anchor_y = 0.0f;
    float anchor_w = 0.0f;
    float anchor_h = 0.0f;
    float delta_x = 0.0f;
    float delta_y = 0.0f;
    float delta_w = 0.0f;
    float delta_h = 0.0f;

    for (j = 0; j < class_num_; j++) {
      float tmp_prob = conf_data[i * class_num_ + j];
      if (tmp_prob > max_prob) {
        max_prob = tmp_prob;
        max_label = static_cast<float>(j);
      }
    }

    if (max_prob > conf_thresh) {
      anchor_x = priors_[i][0];
      delta_y = loc_data[i * 4 + 0];
      anchor_y = priors_[i][1];
      delta_x = loc_data[i * 4 + 1];
      anchor_w = priors_[i][2];
      delta_h = loc_data[i * 4 + 2];
      anchor_h = priors_[i][3];
      delta_w = loc_data[i * 4 + 3];

      // prob label x1 y1 x2 y2
      boxes_after_get[6 * num_after_get + 0] = max_prob;
      boxes_after_get[6 * num_after_get + 1] = max_label;
      float new_x = anchor_x + 0.1 * delta_x * anchor_w;
      float new_y = anchor_y + 0.1 * delta_y * anchor_h;
      float new_w = anchor_w * exp(0.2 * delta_w);
      float new_h = anchor_h * exp(0.2 * delta_h);

      boxes_after_get[6 * num_after_get + 3] = min(max((new_x
          - new_w / 2.0f), 0.0f), 1.0f);
      boxes_after_get[6 * num_after_get + 2] = min(max((new_y
          - new_h / 2.0f), 0.0f), 1.0f);
      boxes_after_get[6 * num_after_get + 5] = min(max((new_x
          + new_w / 2.0f), 0.0f), 1.0f);
      boxes_after_get[6 * num_after_get + 4] = min(max((new_y
          + new_h / 2.0f), 0.0f), 1.0f);

      probs_after_get_for_sort[num_after_get] = max_prob;
      sorted_index[num_after_get] = num_after_get;

      num_after_get++;
    }
  }

  argsort_down(probs_after_get_for_sort, sorted_index, 0, num_after_get - 1);

  // do nms
  for (i = 0; i < num_after_get && num_after_nms < max_box_num_; i++) {
    if (if_retain_for_nms[sorted_index[i]]) {
      // prob label y1 x1 y2 x2
      memcpy(boxes_after_nms + 6 * num_after_nms,
             boxes_after_get + 6 * sorted_index[i], 6 * sizeof(float));
      for (j = i + 1; j < num_after_get; j++) {
        if (box_overlap(boxes_after_get + 6 * sorted_index[i] + 2,
                        boxes_after_get + 6 * sorted_index[j] + 2)
            > threshold_nms_
            && abs(boxes_after_get[6 * sorted_index[i] + 1] -
                boxes_after_get[6 * sorted_index[j] + 1]) < 0.001f) {
          if_retain_for_nms[sorted_index[j]] = 0;
        }
      }
      num_after_nms++;
    }
  }

  vector<DetectRet> result;
  result.clear();
  for (i = 0; i < num_after_nms; i++) {
    DetectRet tmp_ret = {boxes_after_nms[i * 6 + 3],
                         boxes_after_nms[i * 6 + 2],
                         boxes_after_nms[i * 6 + 5],
                         boxes_after_nms[i * 6 + 4],
                         static_cast<int>(boxes_after_nms[i * 6 + 1]),
                         (1.f / (1.f + exp(-boxes_after_nms[i * 6 + 0])))};
    result.push_back(tmp_ret);
  }

  free(boxes_after_nms);
  free(if_retain_for_nms);
  free(sorted_index);
  free(probs_after_get_for_sort);
  free(boxes_after_get);

  return result;
}

vector<DetectRet> SsdTfPostprocess::process_mt(
    float *conf_data,
    float *loc_data,
    int num_thread) {
  assert(num_thread > 1);
  if (priors_.size() == 0) {
    std::cout << "priors is empty, please "
              << "\"generate_multiscale_priors.\"" << std::endl;
    exit(-1);
  }
  pthread_t *threads = reinterpret_cast<pthread_t *>(
      malloc(num_thread * sizeof(pthread_t)));
  memset(threads, 0, num_thread * sizeof(pthread_t));
  int *index_after_get = reinterpret_cast<int *>(
      malloc(prior_num_ * sizeof(int)));
  memset(index_after_get, 0, prior_num_ * sizeof(int));
  float *conf_after_get = reinterpret_cast<float *>(
      malloc(prior_num_ * sizeof(float)));
  memset(conf_after_get, 0, prior_num_ * sizeof(float));
  float *box_after_get = reinterpret_cast<float *>(
      malloc(4 * prior_num_ * sizeof(float)));
  memset(box_after_get, 0, 4 * prior_num_ * sizeof(float));
  TraversalArgs *tras = reinterpret_cast<TraversalArgs *>(
      malloc(num_thread * sizeof(TraversalArgs)));
  memset(tras, 0, num_thread * sizeof(TraversalArgs));

  int num_after_get = 0;

  float conf_thresh = -log((1.0f - threshold_conf_) / threshold_conf_);
  int num_prior_per_thread = static_cast<int>(prior_num_) / num_thread;
  if (static_cast<int>(prior_num_) % num_thread == 0) {
    for (int i = 0; i < num_thread; i++) {
      tras[i].threshold = conf_thresh;
      tras[i].num_input_prior = num_prior_per_thread;
      tras[i].num_class = class_num_;
      tras[i].input_conf = conf_data + i * (num_prior_per_thread * class_num_);
      tras[i].input_loc = loc_data + i * (num_prior_per_thread * 4);
      tras[i].output_index = index_after_get + i * num_prior_per_thread;
      tras[i].output_conf = conf_after_get + i * num_prior_per_thread;
      tras[i].output_box = box_after_get + i * num_prior_per_thread * 4;
      tras[i].anchor_pointer = &priors_;
      tras[i].anchor_offset = i * num_prior_per_thread;
      pthread_create(threads + i, 0, traversal_confs, tras + i);
    }
  } else {
    int index_last_thread = (num_thread - 1) * num_prior_per_thread;
    int num_prior_last_thread = static_cast<int>(prior_num_)
        - index_last_thread;
    for (int i = 0; i < num_thread - 1; i++) {
      tras[i].threshold = conf_thresh;
      tras[i].num_input_prior = num_prior_per_thread;
      tras[i].num_class = class_num_;
      tras[i].input_conf = conf_data + i * (num_prior_per_thread * class_num_);
      tras[i].input_loc = loc_data + i * (num_prior_per_thread * 4);
      tras[i].output_index = index_after_get + i * num_prior_per_thread;
      tras[i].output_conf = conf_after_get + i * num_prior_per_thread;
      tras[i].output_box = box_after_get + i * num_prior_per_thread * 4;
      tras[i].anchor_pointer = &priors_;
      tras[i].anchor_offset = i * num_prior_per_thread;
      pthread_create(threads + i, 0, traversal_confs, tras + i);
    }
    tras[num_thread - 1].threshold = conf_thresh;
    tras[num_thread - 1].num_input_prior = num_prior_last_thread;
    tras[num_thread - 1].num_class = class_num_;
    tras[num_thread - 1].input_conf = conf_data + index_last_thread
        * class_num_;
    tras[num_thread - 1].input_loc = loc_data + index_last_thread * 4;
    tras[num_thread - 1].output_index = index_after_get + index_last_thread;
    tras[num_thread - 1].output_conf = conf_after_get + index_last_thread;
    tras[num_thread - 1].output_box = box_after_get + index_last_thread * 4;
    tras[num_thread - 1].anchor_pointer = &priors_;
    tras[num_thread - 1].anchor_offset = index_last_thread;
    pthread_create(threads + num_thread - 1, 0,
                   traversal_confs, tras + num_thread - 1);
  }

  for (int i = 0; i < num_thread; i++) {
    pthread_join(threads[i], 0);
    num_after_get += tras[i].num_output_prior;
  }

  int current_prior_offset = 0;
  for (int i = 1; i < num_thread; i++) {
    current_prior_offset += tras[i - 1].num_output_prior;
    memcpy(index_after_get + current_prior_offset, tras[i].output_index,
           tras[i].num_output_prior * sizeof(int));
    memcpy(conf_after_get + current_prior_offset, tras[i].output_conf,
           tras[i].num_output_prior * sizeof(float));
    memcpy(box_after_get + 4 * current_prior_offset, tras[i].output_box,
           4 * tras[i].num_output_prior * sizeof(float));
  }
  int *index_sorted = reinterpret_cast<int *>(
      malloc(num_after_get * sizeof(int)));
  memset(index_sorted, 0, num_after_get * sizeof(int));
  bool *if_retain_for_nms = reinterpret_cast<bool *>(
      malloc(num_after_get * sizeof(bool)));
  memset(if_retain_for_nms, 1, num_after_get * sizeof(bool));
  for (int i = 0; i < num_after_get; i++) {
    index_sorted[i] = i;
  }

  argsort_down(conf_after_get, index_sorted, 0, num_after_get - 1);

  vector<DetectRet> result;
  result.clear();
  int num_after_nms = 0;
  // do nms
  for (int i = 0; i < num_after_get && num_after_nms < max_box_num_; i++) {
    if (if_retain_for_nms[index_sorted[i]]) {
      DetectRet tmp = {box_after_get[4 * index_sorted[i]],
                       box_after_get[4 * index_sorted[i] + 1],
                       box_after_get[4 * index_sorted[i] + 2],
                       box_after_get[4 * index_sorted[i] + 3],
                       index_after_get[index_sorted[i]],
                       1.f / (1.f + exp(-conf_after_get[i]))};
      result.push_back(tmp);
      num_after_nms++;
      for (int j = i + 1; j < num_after_get; j++) {
        if (index_after_get[index_sorted[i]] ==
            index_after_get[index_sorted[j]]
            && box_overlap(box_after_get + 4 * index_sorted[i],
                box_after_get + 4 * index_sorted[j]) > threshold_nms_) {
          if_retain_for_nms[index_sorted[j]] = 0;
        }
      }
    }
  }

  free(box_after_get);
  free(if_retain_for_nms);
  free(index_sorted);
  free(tras);
  free(conf_after_get);
  free(index_after_get);
  free(threads);

  return result;
/**

struct TraversalArgs{
  float threshold;
  int num_input_prior;
  int num_class;
  float * input_conf;
  float * input_loc;
  int num_output_prior;
  int * output_index;
  float * output_conf;
  float * output_box;
  std::vector<std::vector<float> > * anchor_piointer;
};
void * traversal_confs(void *arg)
*struct TraversalArgs{
  float threshold;
  int num_input_prior;
  int num_class;
  float * input_data;
  int num_output_prior;
  int * output_index;
  float * output_conf;
};
*/
}

vector<DetectRet> SsdTfPostprocess::process_mt_v2(
    float *conf_data,
    float *loc_data,
    int num_thread) {
  assert(num_thread > 1);
  if (priors_.size() == 0) {
    std::cout << "priors is empty, please "
              << "\"generate_multiscale_priors.\"" << std::endl;
    exit(-1);
  }
  pthread_t *threads = reinterpret_cast<pthread_t *>(
      malloc(num_thread * sizeof(pthread_t)));
  memset(threads, 0, num_thread * sizeof(pthread_t));
  int *index_after_get = reinterpret_cast<int *>(
      malloc(prior_num_ * sizeof(int)));
  memset(index_after_get, 0, prior_num_ * sizeof(int));
  float *conf_after_get = reinterpret_cast<float *>(
      malloc(prior_num_ * sizeof(float)));
  memset(conf_after_get, 0, prior_num_ * sizeof(float));
  float *box_after_get = reinterpret_cast<float *>(
      malloc(4 * prior_num_ * sizeof(float)));
  memset(box_after_get, 0, 4 * prior_num_ * sizeof(float));
  TraversalArgsV2 *tras = reinterpret_cast<TraversalArgsV2 *>(
      malloc(num_thread * sizeof(TraversalArgsV2)));
  memset(tras, 0, num_thread * sizeof(TraversalArgsV2));

  int num_after_get = 0;
  /*
  struct TraversalArgsV2{
  float threshold;
  int num_thread;
  int id_this_thread;
  int num_input_prior;
  int num_class;
  float * input_conf;
  float * input_loc;
  int num_output_prior;
  int * output_index;
  float * output_conf;
  float * output_box;
  std::vector<std::vector<float> > * anchor_pointer;
};
  */
  float conf_thresh = -log((1.0f - threshold_conf_) / threshold_conf_);
  int num_prior_per_thread = static_cast<int>(prior_num_) / num_thread;
  if (static_cast<int>(prior_num_) % num_thread == 0) {
    for (int i = 0; i < num_thread; i++) {
      tras[i].threshold = conf_thresh;
      tras[i].num_thread = num_thread;
      tras[i].id_this_thread = i;
      tras[i].num_input_prior = num_prior_per_thread;
      tras[i].num_class = class_num_;
      tras[i].input_conf = conf_data;
      tras[i].input_loc = loc_data;
      tras[i].output_index = index_after_get + i * num_prior_per_thread;
      tras[i].output_conf = conf_after_get + i * num_prior_per_thread;
      tras[i].output_box = box_after_get + i * num_prior_per_thread * 4;
      tras[i].anchor_pointer = &priors_;
      pthread_create(threads + i, 0, traversal_confs_v2, tras + i);
    }
  } else {
    // little different from v1, this last thread
    // will process less priors than before.
    num_prior_per_thread++;
    int index_last_thread = (num_thread - 1) * num_prior_per_thread;
    int num_prior_last_thread = static_cast<int>(prior_num_)
        - index_last_thread;
    for (int i = 0; i < num_thread - 1; i++) {
      tras[i].threshold = conf_thresh;
      tras[i].num_thread = num_thread;
      tras[i].id_this_thread = i;
      tras[i].num_input_prior = num_prior_per_thread;
      tras[i].num_class = class_num_;
      tras[i].input_conf = conf_data;
      tras[i].input_loc = loc_data;
      tras[i].output_index = index_after_get + i * num_prior_per_thread;
      tras[i].output_conf = conf_after_get + i * num_prior_per_thread;
      tras[i].output_box = box_after_get + i * num_prior_per_thread * 4;
      tras[i].anchor_pointer = &priors_;
      pthread_create(threads + i, 0, traversal_confs_v2, tras + i);
    }
    tras[num_thread - 1].threshold = conf_thresh;
    tras[num_thread - 1].num_thread = num_thread;
    tras[num_thread - 1].id_this_thread = num_thread - 1;
    tras[num_thread - 1].num_input_prior = num_prior_last_thread;
    tras[num_thread - 1].num_class = class_num_;
    tras[num_thread - 1].input_conf = conf_data;
    tras[num_thread - 1].input_loc = loc_data;
    tras[num_thread - 1].output_index = index_after_get + index_last_thread;
    tras[num_thread - 1].output_conf = conf_after_get + index_last_thread;
    tras[num_thread - 1].output_box = box_after_get + index_last_thread * 4;
    tras[num_thread - 1].anchor_pointer = &priors_;
    pthread_create(threads + num_thread - 1, 0,
                   traversal_confs_v2, tras + num_thread - 1);
  }

  for (int i = 0; i < num_thread; i++) {
    pthread_join(threads[i], 0);
    num_after_get += tras[i].num_output_prior;
  }

  int current_prior_offset = 0;
  for (int i = 1; i < num_thread; i++) {
    current_prior_offset += tras[i - 1].num_output_prior;
    memcpy(index_after_get + current_prior_offset, tras[i].output_index,
           tras[i].num_output_prior * sizeof(int));
    memcpy(conf_after_get + current_prior_offset, tras[i].output_conf,
           tras[i].num_output_prior * sizeof(float));
    memcpy(box_after_get + 4 * current_prior_offset, tras[i].output_box,
           4 * tras[i].num_output_prior * sizeof(float));
  }
  int *index_sorted = reinterpret_cast<int *>(
      malloc(num_after_get * sizeof(int)));
  memset(index_sorted, 0, num_after_get * sizeof(int));
  bool *if_retain_for_nms = reinterpret_cast<bool *>(
      malloc(num_after_get * sizeof(bool)));
  memset(if_retain_for_nms, 1, num_after_get * sizeof(bool));
  for (int i = 0; i < num_after_get; i++) {
    index_sorted[i] = i;
  }

  argsort_down(conf_after_get, index_sorted, 0, num_after_get - 1);

  vector<DetectRet> result;
  result.clear();
  int num_after_nms = 0;
  // do nms
  for (int i = 0; i < num_after_get && num_after_nms < max_box_num_; i++) {
    if (if_retain_for_nms[index_sorted[i]]) {
      DetectRet tmp = {box_after_get[4 * index_sorted[i]],
                       box_after_get[4 * index_sorted[i] + 1],
                       box_after_get[4 * index_sorted[i] + 2],
                       box_after_get[4 * index_sorted[i] + 3],
                       index_after_get[index_sorted[i]],
                       1.f / (1.f + exp(-conf_after_get[i]))};
      result.push_back(tmp);
      num_after_nms++;
      for (int j = i + 1; j < num_after_get; j++) {
        if (index_after_get[index_sorted[i]] ==
            index_after_get[index_sorted[j]]
            && box_overlap(box_after_get + 4 * index_sorted[i],
                box_after_get + 4 * index_sorted[j]) > threshold_nms_) {
          if_retain_for_nms[index_sorted[j]] = 0;
        }
      }
    }
  }

  free(box_after_get);
  free(if_retain_for_nms);
  free(index_sorted);
  free(tras);
  free(conf_after_get);
  free(index_after_get);
  free(threads);

  return result;
}

/*
faster rcnn
*/
FasterRcnnPostprocess::FasterRcnnPostprocess(
    size_t class_num,
    float threshold_conf,
    float threshold_nms) : class_num_(class_num),
  threshold_conf_(threshold_conf), threshold_nms_(threshold_nms) {
}

vector<DetectRet> FasterRcnnPostprocess::process(
    int roi_num,
    float *rois_data,
    float *bbox_delt_data,
    float *score) {
}

/*****************************************
 *  * new retail solution post-process
 *   * Added by Kui Ye
 *    *****************************************/

NRSSDPostprocess_v2::NRSSDPostprocess_v2(
    size_t class_num,
    size_t prior_num,
    float threshold_conf,
    float threshold_nms)
    : class_num_(class_num), prior_num_(prior_num),
      threshold_conf_(threshold_conf), threshold_nms_(threshold_nms) {}

double NRSSDPostprocess_v2::jaccard(const Dets &a, const Dets &b) {
  if (a.class_id != b.class_id)
    return 0;
  auto iw = (std::min(a.xmax, b.xmax) - std::max(a.xmin, b.xmin) + 1);
  if (iw <= 0)
    return 0;
  auto ih = (std::min(a.ymax, b.ymax) - std::max(a.ymin, b.ymin) + 1);
  if (ih <= 0)
    return 0;
  auto area = (b.xmax - b.xmin + 1) * (b.ymax - b.ymin + 1);
  auto ua = static_cast<float>((a.xmax - a.xmin + 1) * (a.ymax - a.ymin + 1)
      + area - iw * ih);
  // iou between max box and detection box
  return iw * ih / ua;
}

std::vector<Dets> NRSSDPostprocess_v2::process(
    float *conf_data,
    float *loc_data,
    float *prior_data) {
  // get score filtered results
  vector<Dets> dets;
  Statistics timer;
  timer.record_start("before nms");
  for (int prior_idx = 0; prior_idx < prior_num_; ++prior_idx) {
    float *p_conf = std::max_element(conf_data + prior_idx * class_num_ + 1,
                                     conf_data + (prior_idx + 1) * class_num_);
    if (*p_conf < threshold_conf_)
      continue;
    size_t class_id = std::distance(conf_data + prior_idx * class_num_,
                                    p_conf);
    float score = *p_conf;

    int iter_x = prior_idx * 4 + 0;
    int iter_y = prior_idx * 4 + 1;
    int iter_w = prior_idx * 4 + 2;
    int iter_h = prior_idx * 4 + 3;

    float cx = 0.1 * loc_data[iter_x] * prior_data[iter_w]
        + prior_data[iter_x];
    float cy = 0.1 * loc_data[iter_y] * prior_data[iter_h]
        + prior_data[iter_y];
    float w = exp(0.2 * loc_data[iter_w]) * prior_data[iter_w];
    float h = exp(0.2 * loc_data[iter_h]) * prior_data[iter_h];

    int xmin = static_cast<int>(std::max(cx - 0.5 * w, 0.));
    int ymin = static_cast<int>(std::max(cy - 0.5 * h, 0.));
    int xmax = static_cast<int>(std::min(cx + 0.5 * w, 300.));
    int ymax = static_cast<int>(std::min(cy + 0.5 * h, 300.));

    dets.emplace_back(class_id, score, xmin, ymin, xmax, ymax);
  }
  timer.record_end("before nms");

  if (dets.empty()) {
    cout << "NO Detections meet condition... ..." << endl;
    return dets;
  } else {
    cout << "Num of detections: " << dets.size() << endl;
  }

  // do nms
  //  1. sort the vector by score
  timer.record_start("sort score");
  std::sort(dets.begin(), dets.end(), [](const Dets &lhs, const Dets &rhs) {
    return lhs.score > rhs.score;
  });
  timer.record_end("sort score");

  // 2. nms for each class
  timer.record_start("doing nms");
  vector<Dets> ret;
  /* dets.resize(9); */
  cout << "num bboxes before nums: " << dets.size() << endl;
  while (!dets.empty()) {
    Dets ele = dets[0];
    ret.emplace_back(dets[0]);
    dets.erase(dets.begin());
    for (auto it = dets.begin(); it != dets.end();) {
      cout << "IOU: " << jaccard(ele, *it) << endl;
      if (jaccard(ele, *it) > 0.5) {
        it = dets.erase(it);
      } else {
        ++it;
      }
    }
    cout << endl;
  }
  cout << "num bboxes after nums: " << ret.size() << ", or: "
       << dets.size() << endl;

  timer.record_end("doing nms");
  cout << "before nms: " << timer.get_time("before nms") << endl;
  cout << "sort score: " << timer.get_time("sort score") << endl;
  cout << "doing  nms: " << timer.get_time("doing nms") << endl;

  return ret;
}

/*****************************************
 *  new retail solution post-process
 *  Added by Yu
 ******************************************/
NRSSDPostprocess::NRSSDPostprocess(
    size_t class_num,
    size_t prior_num,
    float threshold_conf,
    float threshold_nms)
    : nb_classes(class_num), nb_priors(prior_num),
      threshold_conf(threshold_conf), threshold_nms(threshold_nms) {
  cols = nb_classes + 1 + 12;
  length_of_dets = nb_priors * cols;
}

double NRSSDPostprocess::jaccard(const Dets &a, const Dets &b) {
  if (a.class_id != b.class_id)
    return 0;
  int iw = (std::min(a.xmax, b.xmax) - std::max(a.xmin, b.xmin) + 1);
  if (iw <= 0)
    return 0;
  int ih = (std::min(a.ymax, b.ymax) - std::max(a.ymin, b.ymin) + 1);
  if (ih <= 0)
    return 0;
  int area = (b.xmax - b.xmin + 1) * (b.ymax - b.ymin + 1);
  auto ua = static_cast<float>((a.xmax - a.xmin + 1) * (a.ymax - a.ymin + 1)
      + area - iw * ih);
  // iou between max box and detection box
  return iw * ih / ua;
}

std::vector<Dets> NRSSDPostprocess::process(float *concated_output) {
  // get score filtered results
  vector<Dets> dets;
  Statistics timer;
  timer.record_start("before nms");
  for (size_t i = nb_classes + 1; i < length_of_dets; i += cols) {
    // find max score [first, last)
    auto iter = std::max_element(concated_output + i - nb_classes,
                                 concated_output + i);
    if (*iter < threshold_conf)
      continue;
    // get index
    auto index = std::distance(concated_output, iter);
    double score = concated_output[index];
    auto class_id = static_cast<size_t>(index % cols);

    float cx = concated_output[i] * 0.1 * concated_output[i + 6]
        + concated_output[i + 4];
    float cy = concated_output[i + 1] * 0.1 * concated_output[i + 7]
        + concated_output[i + 5];
    float w = exp(concated_output[i + 2] * 0.2) * concated_output[i + 6];
    float h = exp(concated_output[i + 3] * 0.2) * concated_output[i + 7];

    auto xmin = static_cast<int>((cx - 0.5 * w) * 300);
    auto ymin = static_cast<int>((cy - 0.5 * h) * 300);
    auto xmax = static_cast<int>((cx + 0.5 * w) * 300);
    auto ymax = static_cast<int>((cy + 0.5 * h) * 300);

    xmin = std::max(xmin, 0);
    ymin = std::max(ymin, 0);
    xmax = std::min(xmax, 300);
    ymax = std::min(ymax, 300);

    dets.emplace_back(class_id, score, xmin, ymin, xmax, ymax);
  }
  if (dets.empty()) {
    cout << "NO Detections meet condition... ..." << endl;
    return dets;
  } else {
    cout << "Num of detections: " << dets.size() << endl;
  }
  timer.record_end("before nms");

  // do nms
  //  1. sort the vector by score
  timer.record_start("sort score");
  std::sort(dets.begin(), dets.end(), [](const Dets &lhs, const Dets &rhs) {
    return lhs.score > rhs.score;
  });
  timer.record_end("sort score");

  // 2. nms for each class
  timer.record_start("doing nms");
  vector<Dets> ret;
  /* dets.resize(9); */
  cout << "num bboxes before nums: " << dets.size() << endl;
  while (!dets.empty()) {
    Dets ele = dets[0];
    ret.emplace_back(dets[0]);
    dets.erase(dets.begin());
    for (auto it = dets.begin(); it != dets.end();) {
      cout << "IOU: " << jaccard(ele, *it) << endl;
      if (jaccard(ele, *it) > 0.5) {
        it = dets.erase(it);
      } else {
        ++it;
      }
    }
    cout << endl;
  }
  cout << "num bboxes after nums: " << ret.size() << ", or: "
       << dets.size() << endl;

  timer.record_end("doing nms");
  cout << "before nms: " << timer.get_time("before nms") << endl;
  cout << "sort score: " << timer.get_time("sort score") << endl;
  cout << "doing  nms: " << timer.get_time("doing nms") << endl;

  return ret;
}

}  // namespace sail
