/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#ifdef USING_OPENCV3
#include <opencv2/opencv.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/core.hpp>
#endif
#include <bmruntime_interface.h>
#include <vector>
#include <algorithm>
#include "core/common.hpp"
#include "core/utils.hpp"
#include "process/preprocess.hpp"

namespace sail {

ImageNetPreprocess::ImageNetPreprocess()
    : height_(224), width_(224), interpolation_(cv::INTER_LINEAR),
      scale_factor_(0.0), convert_color_(false) {
  float negative_mean[] = {-103.94, -116.78, -123.68};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

ImageNetPreprocess::ImageNetPreprocess(Bmcv* processor)
    : Preprocess(processor), height_(224), width_(224),
      interpolation_(cv::INTER_NEAREST),
      scale_factor_(0.0), convert_color_(false) {
  float negative_mean[] = {-103.94, -116.78, -123.68};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

ImageNetPreprocess::ImageNetPreprocess(
    Bmcv*        processor,
    int          height,
    int          width,
    int          interpolation,
    float        mean_0,
    float        mean_1,
    float        mean_2,
    float        scale_factor,
    bool         convert_color)
    : Preprocess(processor), height_(height), width_(width),
      interpolation_(interpolation), scale_factor_(scale_factor),
      convert_color_(convert_color) {
  float negative_mean[] = {-mean_0, -mean_1, -mean_2};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

ImageNetPreprocess::ImageNetPreprocess(
    int   height,
    int   width,
    int   interpolation,
    float mean_0,
    float mean_1,
    float mean_2,
    float scale_factor,
    bool  convert_color)
    : height_(height), width_(width),
      interpolation_(interpolation), scale_factor_(scale_factor),
      convert_color_(convert_color) {
  float negative_mean[] = {-mean_0, -mean_1, -mean_2};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

void ImageNetPreprocess::process_cpu(float* input, cv::Mat& frame) {
  frame.convertTo(frame, CV_32FC3);
  std::vector<cv::Mat> input_channels;
  int size_per_channel = height_ * width_;
  for (int i = 0; i < 3; ++i) {
    cv::Mat channel(height_, width_, CV_32FC1, input);
    input_channels.push_back(channel);
    input += size_per_channel;
  }
  cv::resize(frame, frame, cv::Size(width_, height_), 0, 0, interpolation_);
  frame += cv::Scalar(negative_mean_[0], negative_mean_[1], negative_mean_[2]);
  if (std::abs(scale_factor_) > 1e-6) {
    frame *= scale_factor_;
  }
  if (convert_color_) {
    cv::cvtColor(frame, frame, cv::COLOR_BGR2RGB);
  }
  cv::split(frame, input_channels);
}

void ImageNetPreprocess::process_tpu(
    Tensor<float>&       input,
    Frame&               frame) {
  if (bm_cv_ == NULL) {
    cout << "Error: bm_cv_ is NULL" << endl;
    return;
  }
  vector<int> shape(4);
  shape[0] = 1;
  shape[1] = 3;
  shape[2] = frame.get_height();
  shape[3] = frame.get_width();
  int bgr_bytes = get_size_of_shape(shape);
  Tensor<float> bgr(bm_cv_->get_bmrt(), shape, bgr_bytes);
  bm_cv_->get_bgr_from_frame(frame, bgr);
  if (convert_color_) {
    void* p_bmrt = bm_cv_->get_bmrt();
    vector<int> out_shape(4);
    out_shape[0] = 1;
    out_shape[1] = 3;
    out_shape[2] = height_;
    out_shape[3] = width_;
    Tensor<float> rgb(p_bmrt, out_shape);
    if (std::abs(scale_factor_) > 1e-6) {
      bm_cv_->resize_norm(bgr, rgb,
                          scale_factor_, negative_mean_[0] * scale_factor_,
                          scale_factor_, negative_mean_[1] * scale_factor_,
                          scale_factor_, negative_mean_[2] * scale_factor_);
    } else {
      bm_cv_->resize_norm(bgr, rgb,
                          1.0, negative_mean_[0],
                          1.0, negative_mean_[1],
                          1.0, negative_mean_[2]);
    }
    bm_cv_->bgrsplit(rgb, input);
  } else {
    if (std::abs(scale_factor_) > 1e-6) {
      bm_cv_->resize_norm(bgr, input,
                          scale_factor_, negative_mean_[0] * scale_factor_,
                          scale_factor_, negative_mean_[1] * scale_factor_,
                          scale_factor_, negative_mean_[2] * scale_factor_);
    } else {
      bm_cv_->resize_norm(bgr, input,
                          1.0, negative_mean_[0],
                          1.0, negative_mean_[1],
                          1.0, negative_mean_[2]);
    }
  }
}

/***********************
for yolov2/v3 preprocess
************************/

Yolov23Preprocess::Yolov23Preprocess() : height_(416), width_(416) {}

Yolov23Preprocess::Yolov23Preprocess(int height, int width)
    : height_(height), width_(width) {}

#ifdef BMCV
Yolov23Preprocess::Yolov23Preprocess(Bmcv* processor)
    : Preprocess(processor), height_(416), width_(416) {}

Yolov23Preprocess::Yolov23Preprocess(
    Bmcv* processor,
    int           height,
    int           width)
    : Preprocess(processor), height_(height), width_(width) {}

void Yolov23Preprocess::process_tpu(
    bm_device_mem_t* input,
    AVFrame* frame) {
  if (bm_cv_ == NULL) {
    cout << "Error: bm_cv_ is NULL" << endl;
    return;
  }
  void* p_bmrt = bm_cv_->get_bmrt();
  bm_device_mem_t dmem_bgr, dmem_resize;
  int input_size = 1 * 3 * frame->height * frame->width * sizeof(float);
  // cout << "yolo input_size: " << input_size << endl;
  bmrt_malloc_device_byte(p_bmrt, &dmem_bgr, input_size);
  bmrt_malloc_device_byte(p_bmrt, &dmem_resize, input_size);
  bm_cv_->get_bgr_from_frame(frame, &dmem_bgr, FLOAT);
  Shape input_shape(1, 3, frame->height, frame->width);
  Shape output_shape(1, 3, height_, width_);
  bm_cv_->resize_fit(dmem_bgr,
                                dmem_resize,
                                input_shape,
                                output_shape,
                                FLOAT);
  bm_cv_->normalize(dmem_resize, *input, output_shape,
                               1.0f / 255, 0.0f, 1.0f / 255, 0.0f,
                               1.0f / 255, 0.0f);
  bm_cv_->bgr2rgb(*input, *input, output_shape);
  bmrt_free_device(p_bmrt, dmem_bgr);
  bmrt_free_device(p_bmrt, dmem_resize);
}
#endif

void Yolov23Preprocess::process_cpu(float* input, cv::Mat& frame) {
  frame.convertTo(frame, CV_32FC3);
  cv::Mat resize_img;
  cv::Mat img_roi;
  cv::Mat back = cv::Mat(cv::Size(width_, height_),
                 CV_32FC3, cv::Scalar(127.5, 127.5, 127.5));
  int w = frame.cols;
  int h = frame.rows;
  if (w > h) {
    float ratio = (h * 1.0) / w;
    cv::resize(frame, resize_img, cv::Size(width_, width_ * ratio),
               0, 0, CV_INTER_LINEAR);
    img_roi = back(cv::Rect(0, ((height_ - width_ * ratio) / 2),
                   resize_img.cols, resize_img.rows));
  } else {
    float ratio = (w * 1.0) / h;
    cv::resize(frame, resize_img, cv::Size(height_ * ratio, height_),
               0, 0, CV_INTER_LINEAR);
    img_roi = back(cv::Rect(((width_ - height_ * ratio) / 2), 0,
              resize_img.cols, resize_img.rows));
  }
  resize_img.copyTo(img_roi);
  back /= 255;
  cv::cvtColor(back, back, cv::COLOR_BGR2RGB);
  std::vector<cv::Mat> input_channels;
  int size_per_channel = height_ * width_;
  for (int i = 0; i < 3; ++i) {
    cv::Mat channel(height_, width_, CV_32FC1, input);
    input_channels.push_back(channel);
    input += size_per_channel;
  }
  cv::split(back, input_channels);
  // memcpy(input, (float*)back.data, (3 * width_ * height_) * sizeof(float));
}

FasterRcnnPreprocess::FasterRcnnPreprocess() : height_(608), width_(608) {
  im_info_.assign(3, 0.0);
  hw_scale_.assign(2, 0.0);
}

#ifdef BMCV
FasterRcnnPreprocess::FasterRcnnPreprocess(Bmcv* processor)
    : Preprocess(processor), height_(608), width_(608) {
  im_info_.assign(3, 0.0);
  hw_scale_.assign(2, 0.0);
}

FasterRcnnPreprocess::FasterRcnnPreprocess(
    Bmcv* processor,
    int           height,
    int           width)
    : Preprocess(processor), height_(height), width_(width) {
  im_info_.assign(3, 0.0);
  hw_scale_.assign(2, 0.0);
}
#endif

FasterRcnnPreprocess::FasterRcnnPreprocess(int height, int width)
    : height_(height), width_(width) {
  im_info_.assign(3, 0.0);
  hw_scale_.assign(2, 0.0);
}

vector<float> FasterRcnnPreprocess::get_im_info() {
  return im_info_;
}

vector<float> FasterRcnnPreprocess::get_hw_scale() {
  return hw_scale_;
}

void FasterRcnnPreprocess::process_cpu(float* input, cv::Mat& frame) {
  float height_scale = static_cast<float>(frame.rows) / height_;
  float width_scale = static_cast<float>(frame.cols) / width_;
  float max_scale = max(height_scale, width_scale);
  float img_scale = 1.0 / max_scale;
  hw_scale_[0] = 1.0 / height_scale;
  hw_scale_[1] = 1.0 / width_scale;
  im_info_[0] = height_;
  im_info_[1] = width_;
  im_info_[2] = img_scale;
  cv::Mat cv_resized;
  frame.convertTo(cv_resized, CV_32FC3);
  cv::resize(cv_resized, cv_resized, cv::Size(width_, height_));
  cv::Mat mean(height_, width_, cv_resized.type(),
      cv::Scalar(102.9801, 115.9465, 122.7717));
  cv::Mat normalized;
  subtract(cv_resized, mean, normalized);
  vector<cv::Mat> input_channels;
  for (int i = 0; i < 3; ++i) {
    cv::Mat channel(height_, width_, CV_32FC1, input);
    input_channels.push_back(channel);
    input += height_ * width_;
  }
}

#ifdef BMCV
// to do
void FasterRcnnPreprocess::process_tpu(
    bm_device_mem_t* input,
    AVFrame* frame) {
}
#endif

MtcnnPreprocess::MtcnnPreprocess()
    : scale_factor_(0.0078125), face_factor_(0.89), minSize_(40) {
  float negative_mean[] = {-127.5, -127.5, -127.5};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

MtcnnPreprocess::MtcnnPreprocess(
    float mean_0,
    float mean_1,
    float mean_2,
    float scale_factor)
    : scale_factor_(scale_factor), face_factor_(0.89), minSize_(40) {
  float negative_mean[] = {-mean_0, -mean_1, -mean_2};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

#ifdef BMCV
MtcnnPreprocess::MtcnnPreprocess(Bmcv* processor)
    : Preprocess(processor), scale_factor_(0.0078125) {
  float negative_mean[] = {-127.5, -127.5, -127.5};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

MtcnnPreprocess::MtcnnPreprocess(
    Bmcv*        processor,
    float        mean_0,
    float        mean_1,
    float        mean_2,
    float        scale_factor)
    : Preprocess(processor), scale_factor_(scale_factor) {
  float negative_mean[] = {-mean_0, -mean_1, -mean_2};
  negative_mean_.assign(negative_mean, negative_mean + 3);
}

void MtcnnPreprocess::process_tpu(
    bm_device_mem_t* input,
    AVFrame*             frame,
    int                  height,
    int                  width) {
  if (bm_cv_ == NULL) {
    cout << "Error: bm_cv_ is NULL" << endl;
    return;
  }
  void* p_bmrt = bm_cv_->get_bmrt();
  bm_device_mem_t dmem_bgr;
  int img_size = 3 * frame->height * frame->width * sizeof(float);
  bmrt_malloc_device_byte(p_bmrt, &dmem_bgr, img_size);

  bm_cv_->get_bgr_from_frame(frame, &dmem_bgr, FLOAT);
  Shape input_shape(1, 3, frame->height, frame->width);
  bm_cv_->transpose(dmem_bgr, dmem_bgr, input_shape, INT);
  input_shape.h = frame->width;
  input_shape.w = frame->height;

float* data = new float[input_shape.size()];
bmrt_memcpy_d2s(p_bmrt, data, dmem_bgr);
std::cout << "Raw image data: " << data[1] << ", " << data[100] << ", " <<
          data[1000] << std::endl;
delete [] data;

  Shape output_shape(1, 3, width, height);
cout << "input_size: [" << frame->height << ", " << frame->width
     << "]" << endl;
cout << "output_size: [" << height << ", " << width << "]" << endl;

  if (std::abs(scale_factor_) > 1e-6) {
    bm_cv_->resize_norm(dmem_bgr, *input,
                    input_shape, output_shape, FLOAT,
                    scale_factor_, negative_mean_[0] * scale_factor_,
                    scale_factor_, negative_mean_[1] * scale_factor_,
                    scale_factor_, negative_mean_[2] * scale_factor_);
  } else {
    bm_cv_->resize_norm(dmem_bgr, *input,
                    input_shape, output_shape, FLOAT,
                    1.0, negative_mean_[0],
                    1.0, negative_mean_[1],
                    1.0, negative_mean_[2]);
  }
  bmrt_free_device(p_bmrt, dmem_bgr);
}

#endif

void MtcnnPreprocess::process_cpu(
    float*   input,
    cv::Mat& frame,
    int      height,
    int      width) {
  std::vector<cv::Mat> input_channels;
  int size_per_channel = height * width;
  for (int i = 0; i < 3; ++i) {
    cv::Mat channel(height, width, CV_32FC1, input);
    input_channels.push_back(channel);
    input += size_per_channel;
  }
  cv::Mat resized;
  cv::resize(frame, resized, cv::Size(width, height), 0, 0, cv::INTER_LINEAR);
  resized.convertTo(resized, CV_32FC3, scale_factor_,
                      negative_mean_[0] * scale_factor_);
  cv::split(resized, input_channels);
}

std::vector<double> MtcnnPreprocess::face_scale_gen(  // gen face factor
    cv::Mat& frame,
    double   factor,
    int      minSize) {
  face_factor_ = factor;
  minSize_ = minSize;
  int width = frame.cols;
  int height = frame.rows;
  int minWH = std::min(height, width);
  int factor_count = 0;
  double m = 12.0 / minSize;
  minWH *= m;
  std::vector<double> scales;
  while (minWH >= 12) {
    scales.push_back(m * std::pow(face_factor_, factor_count));
    minWH *= factor;
    factor_count++;
  }
  return scales;
}

void MtcnnPreprocess::ronet_pre_img(float* input_data,  // for img joint
    cv::Mat& frame, int input_height, int input_width,
    std::vector<FaceInfo>& regressed_pading,
    std::vector<FaceInfo>& regressed_rects) {
  int numBox = regressed_rects.size();  // input img num
  std::vector<cv::Mat> img_group;
  for (int i = 0; i < numBox; i++) {
    int pad_left = std::abs(regressed_pading[i].bbox.x1
                  - regressed_rects[i].bbox.x1);
    int pad_top = std::abs(regressed_pading[i].bbox.y1
                  - regressed_rects[i].bbox.y1);
    int pad_right = std::abs(regressed_pading[i].bbox.x2
                  - regressed_rects[i].bbox.x2);
    int pad_bottom = std::abs(regressed_pading[i].bbox.y2
                  - regressed_rects[i].bbox.y2);
    cv::Mat crop_img = frame(cv::Range(regressed_pading[i].bbox.y1,
                             regressed_pading[i].bbox.y2 + 1),
                             cv::Range(regressed_pading[i].bbox.x1,
                             regressed_pading[i].bbox.x2 + 1));
    cv::copyMakeBorder(crop_img, crop_img, pad_top, pad_bottom,
        pad_left, pad_right, cv::BORDER_CONSTANT, cv::Scalar(0));
    cv::Mat resized, ended;
    cv::resize(crop_img, resized, cv::Size(input_width, input_height),
               0, 0, cv::INTER_LINEAR);
    resized.convertTo(ended, CV_32FC3, scale_factor_,
                      negative_mean_[0] * scale_factor_);
    img_group.push_back(ended);
  }
  regressed_pading.clear();
  for (size_t i = 0; i < img_group.size(); ++i) {
    if (img_group[i].channels() != 3)
      throw;
    const float* src = img_group[i].ptr<float>(0);
    float* dst[3];
    dst[0] = input_data + i * 3 * input_height * input_width;
    dst[1] = dst[0] + input_height * input_width;
    dst[2] = dst[1] + input_height * input_width;
    for (int j = 0; j < input_height * input_width; ++j) {
        *dst[0] = *src; ++src; ++dst[0];
        *dst[1] = *src; ++src; ++dst[1];
        *dst[2] = *src; ++src; ++dst[2];
    }
  }
  cout << "img joint finished" << endl;
}

void MtcnnPreprocess::ronet_preprocess_cpu(float* input,  // for input offset
    float* all_data, vector<int>& input_shape, int num_done, int num_work) {
  int input_stride = input_shape[1] * input_shape[2] * input_shape[3];
  // cpy data to input sys data
  memcpy(input, all_data + num_done * input_stride,
      num_work * input_stride * sizeof(float));
}

}  // namespace sail
