/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#include "process/postprocess.hpp"
#include "core/bm_dnn.hpp"
#include <numeric>
#include <vector>
#include <algorithm>
#include <iostream>
#include <cmath>
#include "core/common.hpp"

namespace sail {

Yolov3Postprocess::Yolov3Postprocess(
    size_t       class_num,
    float        threshold_prob,
    float        threshold_nms,
    vector<int>& input_shape)
    : classes_num_(class_num), threshold_prob_(threshold_prob),
      threshold_nms_(threshold_nms), net_w_(input_shape[3]),
      net_h_(input_shape[2]) {
  out_s_ = net_w_ / 32;
  out_m_ = net_w_ / 16;
  out_l_ = net_w_ / 8;
  //  cout << "check yolov3 post process output size:
  //  " << "out_s_: " << out_s_ << " out_m_: "
  //  << out_m_ << " out_l_: " << out_l_ << endl;
}

static int nms_comparator(const void *pa, const void *pb) {
  detection a = *reinterpret_cast<const detection *>(pa);
  detection b = *reinterpret_cast<const detection *>(pb);
  float diff = 0;
  if (b.sort_class >= 0)
    diff = a.prob[b.sort_class] - b.prob[b.sort_class];
  else
    diff = a.objectness - b.objectness;
  return diff < 0 ? 1 : -1;
}

float Yolov3Postprocess::box_iou(yolov3_box a, yolov3_box b) {
  float area1 = a.w * a.h;
  float area2 = b.w * b.h;
  float wi = std::min((a.x + a.w / 2), (b.x + b.w / 2))
              - std::max((a.x - a.w / 2), (b.x - b.w / 2));
  float hi = std::min((a.y + a.h / 2), (b.y + b.h / 2))
              - std::max((a.y - a.h / 2), (b.y - b.h / 2));
  float area_i = std::max(wi, 0.0f) * std::max(hi, 0.0f);
  return area_i / (area1 + area2 - area_i);
}

void Yolov3Postprocess::do_nms_sort(
    detection* dets,
    int        total,
    int        classes,
    float      thresh) {
  int i, j, k;
  k = total - 1;
  for (i = 0; i <= k; ++i) {
    if (dets[i].objectness == 0) {
      detection swap = dets[i];
      dets[i] = dets[k];
      dets[k] = swap;
      --k;
      --i;
    }
  }
  total = k + 1;
  for (k = 0; k < classes; ++k) {
    for (i = 0; i < total; ++i) {
      dets[i].sort_class = k;
    }
    qsort(dets, total, sizeof(detection), nms_comparator);
    for (i = 0; i < total; ++i) {
      if (dets[i].prob[k] == 0) continue;
      yolov3_box a = dets[i].bbox;
      for (j = i + 1; j < total; ++j) {
        yolov3_box b = dets[j].bbox;
        if (box_iou(a, b) > thresh) {
          dets[j].prob[k] = 0;
        }
      }
    }
  }
}

layer Yolov3Postprocess::make_yolo_layer(
    int    batch,
    int    w,
    int    h,
    int    n,
    int    total,
    int    classes) {
  layer l = {0};
  l.n = n;
  l.total = total;
  l.batch = batch;
  l.h = h;
  l.w = w;
  l.c = n * (classes + 4 + 1);
  l.out_w = l.w;
  l.out_h = l.h;
  l.out_c = l.c;
  l.classes = classes;
  l.inputs = l.w * l.h * l.c;
  l.biases = reinterpret_cast<float *>(calloc(total * 2, sizeof(float)));
  for (int i = 0; i < total * 2; ++i) {
    l.biases[i] = biases_[i];  // init the anchor size with pre value
  }
  l.mask = reinterpret_cast<int *>(calloc(n, sizeof(int)));
  if (l.w == 13) {
    int j = 6;
    for (int i = 0; i < l.n; ++i)
      l.mask[i] = j++;
  }
  if (l.w == 26) {
    int j = 3;
    for (int i = 0; i < l.n; ++i)
      l.mask[i] = j++;
  }
  if (l.w == 52) {
    int j = 0;
    for (int i = 0; i < l.n; ++i)
      l.mask[i] = j++;
  }
  l.outputs = l.inputs;
  l.output =
      reinterpret_cast<float *>(calloc(batch * l.outputs, sizeof(float)));
  return l;
}

void Yolov3Postprocess::free_yolo_layer(layer l) {
  if (NULL != l.biases) {
    free(l.biases);
    l.biases = NULL;
  }
  if (NULL != l.mask) {
    free(l.mask);
    l.mask = NULL;
  }
  if (NULL != l.output) {
    free(l.output);
    l.output = NULL;
  }
}

int Yolov3Postprocess::entry_index(
    layer    l,
    int      batch,
    int      location,
    int      entry) {
  int n = location / (l.w * l.h);
  int loc = location % (l.w * l.h);
  return batch * l.outputs + n * l.w * l.h * (4 + l.classes + 1)
         + entry * l.w * l.h + loc;
}

void Yolov3Postprocess::forward_yolo_layer(
    const float*    input,
    layer           l) {
  memcpy(l.output, reinterpret_cast<const float *>(input),
          l.outputs * l.batch * sizeof(float));
}

int Yolov3Postprocess::yolo_num_detections(
    layer     l,
    float     thresh) {
  int i, n;
  int count = 0;
  for (i = 0; i < l.w * l.h; ++i) {
    for (n = 0; n < l.n; ++n) {
      int obj_index = entry_index(l, 0, n * l.w * l.h + i, 4);
      if (l.output[obj_index] > thresh)
        ++count;
    }
  }
  return count;
}

int Yolov3Postprocess::num_detections(
    std::vector<layer> layers_params,
    float              thresh) {
  unsigned int i;
  int s = 0;
  for (i = 0; i < layers_params.size(); ++i) {
    layer l = layers_params[i];
    s += yolo_num_detections(l, thresh);
  }
  return s;
}

detection* Yolov3Postprocess::make_network_boxes(
    std::vector<layer>    layers_params,
    float                 thresh,
    int*                  num) {
  layer l = layers_params[0];
  int i;
  int nboxes = num_detections(layers_params, thresh);
  if (num) *num = nboxes;
  detection *dets =
      reinterpret_cast<detection *>(calloc(nboxes, sizeof(detection)));
  for (i = 0; i < nboxes; ++i) {
    dets[i].prob = reinterpret_cast<float *>(calloc(l.classes, sizeof(float)));
  }
  return dets;
}

void Yolov3Postprocess::correct_yolo_boxes(
    detection*    dets,
    int           n,
    int           w,
    int           h,
    int           netw,
    int           neth,
    int           relative) {
  int i;
  int new_w = 0;
  int new_h = 0;
  if (((float)netw / w) < ((float)neth / h)) {
    new_w = netw;
    new_h = (h * netw) / w;
  } else {
    new_h = neth;
    new_w = (w * neth) / h;
  }
  for (i = 0; i < n; ++i) {
    yolov3_box b = dets[i].bbox;
    b.x = (b.x - (netw - new_w) / 2. / netw)
        / ((float)new_w / netw);
    b.y = (b.y - (neth - new_h) / 2. / neth)
        / ((float)new_h / neth);
    b.w *= (float)netw / new_w;
    b.h *= (float)neth / new_h;
    if (!relative) {
      b.x *= w;
      b.w *= w;
      b.y *= h;
      b.h *= h;
    }
    dets[i].bbox = b;
  }
}

yolov3_box Yolov3Postprocess::get_yolo_box(
     float*    x,
     float*    biases,
     int       n,
     int       index,
     int       i,
     int       j,
     int       lw,
     int       lh,
     int       w,
     int       h,
     int       stride) {
  yolov3_box b;
  b.x = (i + x[index + 0 * stride]) / lw;
  b.y = (j + x[index + 1 * stride]) / lh;
  b.w = exp(x[index + 2 * stride]) * biases[2 * n] / w;
  b.h = exp(x[index + 3 * stride]) * biases[2 * n + 1] / h;
  return b;
}

int Yolov3Postprocess::get_yolo_detections(
    layer         l,
    int           w,
    int           h,
    int           netw,
    int           neth,
    float         thresh,
    int*          map,
    int           relative,
    detection*    dets) {
  int i, j, n;
  float *predictions = l.output;
  int count = 0;
  for (i = 0; i < l.w * l.h; ++i) {
    int row = i / l.w;
    int col = i % l.w;
    for (n = 0; n < l.n; ++n) {
      int obj_index = entry_index(l, 0, n * l.w * l.h + i, 4);
      float objectness = predictions[obj_index];
      if (objectness <= thresh) {
        continue;
      }
      int box_index = entry_index(l, 0, n * l.w * l.h + i, 0);
      dets[count].bbox =
          get_yolo_box(predictions, l.biases, l.mask[n], box_index,
                          col, row, l.w, l.h, netw, neth, l.w * l.h);
      dets[count].objectness = objectness;
      dets[count].classes = l.classes;
      for (j = 0; j < l.classes; ++j) {
        int class_index = entry_index(l, 0, n * l.w * l.h + i, 4 + 1 + j);
        float prob = objectness * predictions[class_index];
        dets[count].prob[j] = (prob > thresh) ? prob : 0;
      }
      ++count;
    }
  }
  correct_yolo_boxes(dets, count, w, h, netw, neth, relative);
  return count;
}

void Yolov3Postprocess::fill_network_boxes(
    std::vector<layer>    layers_params,
    int                   w,
    int                   h,
    float                 thresh,
    float                 hier,
    int*                  map,
    int                   relative,
    detection*            dets) {
  unsigned int j;
  for (j = 0; j < layers_params.size(); ++j) {
    layer l = layers_params[j];
    int count = get_yolo_detections(l, w, h, net_w_, net_h_,
                                    thresh, map, relative, dets);
    dets += count;
  }
}

detection* Yolov3Postprocess::get_network_boxes(
    std::vector<layer>    layers_params,
    int                   img_w,
    int                   img_h,
    float                 thresh,
    float                 hier,
    int*                  map,
    int                   relative,
    int*                  num) {
  detection* dets = make_network_boxes(layers_params, thresh, num);
  fill_network_boxes(layers_params, img_w, img_h,
                     thresh, hier, map, relative, dets);
  return dets;
}

detection* Yolov3Postprocess::get_detections(
    std::vector<float*>    blobs,
    int                    img_w,
    int                    img_h,
    int*                   nboxes) {
  int classes = classes_num_;
  float thresh = threshold_prob_;
  float hier_thresh = threshold_prob_;
  float nms = threshold_nms_;
  int fm_size[6] = {out_s_, out_s_, out_m_, out_m_, out_l_, out_l_};
  std::vector<layer> layers_params;
  layers_params.clear();
  for (unsigned int i = 0; i < blobs.size(); ++i) {
    layer l_params = make_yolo_layer(1, fm_size[2 * i],
                                     fm_size[2 * i + 1], 3, 9, classes);
    layers_params.push_back(l_params);
    forward_yolo_layer(blobs[i], l_params);  // blobs[i] host_mem data
  }
  detection *dets = get_network_boxes(layers_params, img_w, img_h,
                                      thresh, hier_thresh, 0, 1, nboxes);
  // release layer memory
  for (unsigned int index = 0; index < layers_params.size(); ++index) {
    free_yolo_layer(layers_params[index]);
  }
  if (nms) do_nms_sort(dets, (*nboxes), classes, nms);
  return dets;
}

void Yolov3Postprocess::free_detections(detection *dets, int nboxes) {
  for (int i = 0; i < nboxes; ++i)
    free(dets[i].prob);
  free(dets);
}

int Yolov3Postprocess::set_index(
    int    w,
    int    h,
    int    num,
    int    classes,
    int    batch,
    int    location,
    int    entry) {
  int n = location / (w * h);
  int loc = location % (w * h);
  int c = num * (classes + 4 + 1);
  int output = w * h * c;
  return batch * output + n * w * h * (4 + classes + 1)
         + entry * w * h + loc;
}

void Yolov3Postprocess::bmrt_return(std::vector<float *> data_i) {
  int anchor_num = 3;
  int total_num = 9;
  int classes = classes_num_;
  bmrt_return_internal(data_i[0], out_s_, out_s_,
          anchor_num, total_num, classes);
  bmrt_return_internal(data_i[1], out_m_, out_m_,
          anchor_num, total_num, classes);
  bmrt_return_internal(data_i[2], out_l_, out_l_,
          anchor_num, total_num, classes);
}

void Yolov3Postprocess::bmrt_return_internal(
    float*    data_i,
    int       w,
    int       h,
    int       num,
    int       total,
    int       classes) {
  int b, n;
  for (b = 0; b < 1; b++) {
    for (n = 0; n < num; n++) {
      int index = set_index(w, h, num, classes, b, n * w * h, 2);
      for (int i = 0; i < 2 * w * h; i++) {
        data_i[index + i] = -log(1.0f / data_i[index + i] - 1);
      }
    }
  }
}

int Yolov3Postprocess::max_index(float* a, int n) {
  if (n <= 0) return -1;
  int i, max_i = 0;
  float max = a[0];
  for (i = 1; i < n; ++i) {
    if (a[i] > max) {
      max = a[i];
      max_i = i;
    }
  }
  return max_i;
}

std::vector<yolov3_DetectRect> Yolov3Postprocess::detection_yolov3_process(
    detection*    dets,
    int           nboxes,
    int           cols,
    int           rows) {
  int classes = classes_num_;
  float thresh = threshold_prob_;
  std::vector<yolov3_DetectRect> dets_all;
  for (int k = 0; k < nboxes; k++) {
    yolov3_box b = dets[k].bbox;
    int left = (b.x - b.w / 2.) * cols;
    int right = (b.x + b.w / 2.) * cols;
    int top = (b.y - b.h / 2.) * rows;
    int bot = (b.y + b.h / 2.) * rows;
    if (left < 0) left = 0;
    if (right > cols - 1) right = cols - 1;
    if (top < 0) top = 0;
    if (bot > rows - 1) bot = rows - 1;
    int category = max_index(dets[k].prob, classes);
    if (dets[k].prob[category] > thresh) {
      yolov3_DetectRect det_k;
      det_k.left = left;
      det_k.right = right;
      det_k.top = top;
      det_k.bot = bot;
      det_k.category = category;
      det_k.score = dets[k].prob[category];
      dets_all.push_back(det_k);
    }
  }
  return dets_all;
}

std::vector<yolov3_DetectRect> Yolov3Postprocess::process(
    BmRuntime         bmrt,
    Tensor<float>&    output_s,
    Tensor<float>&    output_m,
    Tensor<float>&    output_l,
    int               img_w,
    int               img_h) {
  // logistics calculation for tx ty to tclasses
  int s_size = out_s_ * out_s_ * (3 * (5 + classes_num_));
  int m_size = out_m_ * out_m_ * (3 * (5 + classes_num_));
  int l_size = out_l_ * out_l_ * (3 * (5 + classes_num_));
  // cout << "size before yolov3 sigmoid cal: " << s_size <<
  // " " << m_size << " " << l_size << endl;
  Bmdnn bmdnn(bmrt);
  Tensor<float> ret_s(bmrt.get_ptr(), output_s.shape(), true, true, false,
                      NULL, NULL, NULL);
  bmdnn.sigmoid(output_s, ret_s);
  ret_s.sync_d2s();
  Tensor<float> ret_m(bmrt.get_ptr(), output_m.shape(), true, true, false,
                      NULL, NULL, NULL);
  bmdnn.sigmoid(output_m, ret_m);
  ret_m.sync_d2s();
  Tensor<float> ret_l(bmrt.get_ptr(), output_l.shape(), true, true, false,
                      NULL, NULL, NULL);
  bmdnn.sigmoid(output_l, ret_l);
  ret_l.sync_d2s();
  int nboxes = 0;
  vector<float*> blobs;
  blobs.push_back(ret_s.sys_data());
  blobs.push_back(ret_m.sys_data());
  blobs.push_back(ret_l.sys_data());
  bmrt_return(blobs);
  detection* dets = get_detections(blobs, img_w, img_h, &nboxes);
  std::vector<yolov3_DetectRect> det_results =
                   detection_yolov3_process(dets, nboxes, img_w, img_h);
  free_detections(dets, nboxes);
  return det_results;
}

Yolov3Postprocess::~Yolov3Postprocess() {}

}  // namespace sail
