/* Copyright 2016-2022 by Bitmain Technologies Inc. All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.

==============================================================================*/

#define PY_ARRAY_UNIQUE_SYMBOL pbcvt_ARRAY_API
#include <boost/python.hpp>
#include <boost/python/enum.hpp>
#include <boost/python/def.hpp>
#include <boost/python/register_ptr_to_python.hpp>
#include <boost/python/numpy.hpp>
#include "core/inference.hpp"
#include "core/bm_cv.hpp"
#include "core/bm_dnn.hpp"
using std::vector;
using std::map;
using std::string;
using boost::python::override;
using boost::python::pure_virtual;
using boost::python::bases;
using boost::python::init;
using boost::python::def;
using boost::python::to_python_converter;
using boost::python::class_;
using boost::python::enum_;
namespace bp = boost::python;
namespace bpn = boost::python::numpy;

namespace sail {

// For Python, for now, we'll just always use float as the type.
typedef float Dtype;

BOOST_PYTHON_MODULE(sail) {
  bp::scope().attr("__version__") = "2.0.0";
  Py_Initialize();
  bpn::initialize();

  enum_<TensorMode>("TensorMode")
    .value("PING", PING)
    .value("PONG", PONG)
    .export_values();

  enum_<IOMode>("IOMode")
    .value("SYSI", SYSI)
    .value("SYSO", SYSO)
    .value("SYSIO", SYSIO)
    .value("DEVIO", DEVIO)
    .export_values();

  enum_<AsyncStatus>("AsyncStatus")
    .value("BEGIN", BEGIN)
    .value("PROCESS", PROCESS)
    .value("END", END)
    .export_values();

  class_<Tensor<Dtype>>("Tensor", init<bpn::ndarray&>())
    .def(init<BmRuntime, bpn::ndarray&>())
    .def("shape", &Tensor<Dtype>::shape_py)
    .def("sys_data", &Tensor<Dtype>::sys_data_py);

  bpn::ndarray (Inference<Dtype>::*f0)(const std::string&, const std::string&)
      = &Inference<Dtype>::get_input_tensor_sys_py;
  Tensor<Dtype> (Inference<Dtype>::*f1)(const std::string&, const std::string&)
      = &Inference<Dtype>::get_input_tensor_dev_py;
  Tensor<Dtype> (Inference<Dtype>::*f2)(const std::string&, const std::string&,
      TensorMode) = &Inference<Dtype>::get_input_tensor_dev_py;
  bpn::ndarray (Inference<Dtype>::*f3)(const std::string&, const std::string&)
      = &Inference<Dtype>::get_output_tensor_sys_py;
  Tensor<Dtype> (Inference<Dtype>::*f4)(const std::string&, const std::string&)
      = &Inference<Dtype>::get_output_tensor_dev_py;
  Tensor<Dtype> (Inference<Dtype>::*f5)(const std::string&, const std::string&,
      TensorMode) = &Inference<Dtype>::get_output_tensor_dev_py;

  class_<Inference<Dtype>>("Inference", init<const std::string&,
       const std::string&, const std::string&, IOMode, bool, bool>())
    .def("get_bmrt", &Inference<Dtype>::get_bmrt)
    .def("get_all_graph_names", &Inference<Dtype>::get_all_graph_names_py)
    .def("get_input_tensor_names", &Inference<Dtype>::get_input_tensor_names_py)
    .def("get_output_tensor_names",
        &Inference<Dtype>::get_output_tensor_names_py)
    .def("get_max_input_shapes", &Inference<Dtype>::get_max_input_shapes_py)
    .def("get_input_shape", &Inference<Dtype>::get_input_shape_py)
    .def("get_output_shape", &Inference<Dtype>::get_output_shape_py)
    .def("set_input_shape_sync", &Inference<Dtype>::set_input_shape_sync_py)
    .def("set_input_shape_async", &Inference<Dtype>::set_input_shape_async_py)
    .def("set_input_tensor_sys", &Inference<Dtype>::set_input_tensor_sys_py)
    .def("get_input_tensor_sys", f0)
    .def("get_input_tensor_dev", f1)
    .def("get_input_tensor_dev", f2)
    .def("get_output_tensor_sys", f3)
    .def("get_output_tensor_dev", f4)
    .def("get_output_tensor_dev", f5)
    .def("process_sync", &Inference<Dtype>::process_sync)
    .def("process_async", &Inference<Dtype>::process_async);

  class_<Frame>("Frame")
    .def("is_null", &Frame::is_null)
    .def("get_height", &Frame::get_height)
    .def("get_width", &Frame::get_width);

  class_<VideoDecoder>("VideoDecoder", init<const std::string&>())
    .def("is_opened", &VideoDecoder::is_opened)
    .def("read", &VideoDecoder::read);

  class_<Bmcv>("Bmcv", init<BmRuntime>())
    .def("get_bgr_from_frame", &Bmcv::get_bgr_from_frame)
    .def("resize", &Bmcv::resize)
    .def("crop", &Bmcv::crop)
    .def("yuv2bgr", &Bmcv::yuv2bgr)
    .def("crop_resize_norm", &Bmcv::crop_resize_norm)
    .def("resize_norm", &Bmcv::resize_norm)
    .def("transpose", &Bmcv::transpose)
    .def("bgrsplit", &Bmcv::bgrsplit);

  class_<Bmdnn>("Bmdnn", init<BmRuntime>())
    .def("sigmoid", &Bmdnn::sigmoid);
}

}  // namespace sail
