# Copyright 2016-2022 Bitmain Technologies Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Run a object detection model of MobileNet_SSD with input of a video
"""

from __future__ import print_function
import sys
import os
import argparse
import cv2
import unittest
#WORK_DIR = os.getcwd()
#sys.path.append(os.path.join(WORK_DIR, "build/lib/"))
#import sail

def print_result(index, bbox):
  """Print the classification results.

    Args:
      index: Index of frame.
      bbox: Bounding boxes of detection reuslt.

    Returns:
      None.
  """
  for i in range(bbox.size()):
    print("Frame[{:}]: {:.4f}, {:.4f}, {:.4f}, {:.4f}, {:}, {:.4f}".format(\
          index, bbox[i].x_min_ratio, bbox[i].y_min_ratio, \
          bbox[i].x_max_ratio, bbox[i].y_max_ratio, \
          bbox[i].class_id, bbox[i].score))
  print("")


def check_bbboxes(bboxes,true_result):
  if (bboxes.size() != len(true_result)):
    return False
  for i in range(bboxes.size()):
    if (abs(bboxes[i].x_min_ratio-true_result[i].x_min_ratio) > 1.0e-6 or \
        abs(bboxes[i].y_min_ratio-true_result[i].y_min_ratio) > 1.0e-6 or \
        abs(bboxes[i].x_max_ratio-true_result[i].x_max_ratio) > 1.0e-6 or \
        abs(bboxes[i].y_max_ratio-true_result[i].y_max_ratio) > 1.0e-6 or \
        abs(bboxes[i].class_id-true_result[i].class_id) > 1.0e-6 or \
        abs(bboxes[i].score-true_result[i].score) > 1.0e-6 ) :
      return False;
  return True;

class TestSSD(unittest.TestCase):
  def __init__(self,testname,param):
    self.sail_path = param
    self.context_dir = os.path.join(self.sail_path,"BM1682bmodel_c/mobilenetssd_ir/")
    true_result_file = os.path.join(self.sail_path,"tests/data/ssd_result.txt")
    self.input_path = os.path.join(self.sail_path,"tests/data/car.mp4")
    self.true_result_m = dict()
    mfile = open(true_result_file,"r")
    for eachline in mfile.readlines():
       t = eachline.rstrip().split()
       frameid = int(t[0])
       tmp_v = []
       for i in range(1,len(t),6):
         tmp_detectret = sail.DetectRet()
         tmp_detectret.x_min_ratio = float(t[i]);
         tmp_detectret.y_min_ratio = float(t[i+1]);
         tmp_detectret.x_max_ratio = float(t[i+2]);
         tmp_detectret.y_max_ratio = float(t[i+3]);
         tmp_detectret.class_id = int(t[i+4]);
         tmp_detectret.score = float(t[i+5]);
         tmp_v.append(tmp_detectret)
       self.true_result_m[frameid] = tmp_v
    super(TestSSD, self).__init__(testname)

  def SSD_sync(self):
    inference = sail.Inference(self.context_dir, "BM1682_PCIE", "0", \
                               sail.IOMode.SYSIO, False, False)
    graph_name = inference.get_all_graph_names()[0]
    input_names = inference.get_input_tensor_names(graph_name)
    input_tensor = inference.get_input_tensor_sys(graph_name, input_names[0])

    preprocessor = sail.ImageNetPreprocess(300, 300, cv2.INTER_NEAREST, \
                                           127.5, 127.5, 127.5, 0.007843, False)
    postprocessor = sail.SsdPostprocess(21, 1917, 0.5, 0.5)
    print("Frame[index]: x_min_ratio, y_min_ratio, x_max_ratio, " \
          + "y_max_ratio, class_id, score")
    cap = cv2.VideoCapture(self.input_path)
    ret, frame = cap.read()
    index = 0
    # Run synchronous mode of inference
    while ret:
      preprocessor.process_cpu(input_tensor, 0, frame)
      inference.process(graph_name)
      conf_data = inference.get_output_tensor_sys( \
          graph_name, "mbox_conf_flatten")
      loc_data = inference.get_output_tensor_sys(graph_name, "mbox_loc")
      prior_data = inference.get_output_tensor_sys(graph_name, "mbox_priorbox")
      bboxes = postprocessor.process(conf_data, loc_data, prior_data)
      if bboxes.size() > 0 :
        self.assertTrue(check_bbboxes(bboxes,self.true_result_m[index]))
      index += 1
      ret, frame = cap.read()
    cap.release()

  def SSD_async(self):
    inference = sail.Inference(self.context_dir, "BM1682_PCIE", "0", \
                               sail.IOMode.SYSIO, False, True)
    graph_name = inference.get_all_graph_names()[0]
    input_names = inference.get_input_tensor_names(graph_name)
    input_tensor = inference.get_input_tensor_sys(graph_name, input_names[0])

    preprocessor = sail.ImageNetPreprocess(300, 300, cv2.INTER_NEAREST, \
                                           127.5, 127.5, 127.5, 0.007843, False)
    postprocessor = sail.SsdPostprocess(21, 1917, 0.5, 0.5)
    print("Frame[index]: x_min_ratio, y_min_ratio, x_max_ratio, " \
          + "y_max_ratio, class_id, score")
    cap = cv2.VideoCapture(self.input_path)
    ret, frame = cap.read()
    index = 0
    mode = sail.TensorMode.PING
    status = sail.AsyncStatus.BEGIN
    while ret:
      preprocessor.process_cpu(input_tensor, 0, frame)
      inference.process(graph_name, mode, status)
      if status != sail.AsyncStatus.BEGIN:
        conf_data = inference.get_output_tensor_sys( \
            graph_name, "mbox_conf_flatten")
        loc_data = inference.get_output_tensor_sys(graph_name, "mbox_loc")
        prior_data = inference.get_output_tensor_sys( \
            graph_name, "mbox_priorbox")
        bboxes = postprocessor.process(conf_data, loc_data, prior_data)
        if bboxes.size() > 0 :
           self.assertTrue(check_bbboxes(bboxes,self.true_result_m[index]))
#        print_result(index, bboxes)
        index += 1
      mode = sail.TensorMode.PONG if mode == sail.TensorMode.PING \
             else sail.TensorMode.PING
      status = sail.AsyncStatus.PROCESS
      ret, frame = cap.read()
    inference.process(graph_name, mode, sail.AsyncStatus.END)
    conf_data = inference.get_output_tensor_sys( \
        graph_name, "mbox_conf_flatten")
    loc_data = inference.get_output_tensor_sys(graph_name, "mbox_loc")
    prior_data = inference.get_output_tensor_sys(graph_name, "mbox_priorbox")
    bboxes = postprocessor.process(conf_data, loc_data, prior_data)
#    print_result(index, bboxes)
    cap.release()



if __name__ == "__main__":
  if len(sys.argv) != 2:
    print("Usage: python pytest_img_cls.py sail_root")
    sys.exit(-1)

  root_path = sys.argv[1]
  sys.path.append(os.path.join(root_path, "./build/lib/"))
  import sail
  suite = unittest.TestSuite()
  suite.addTest(TestSSD("SSD_sync",root_path ))
  suite.addTest(TestSSD("SSD_async",root_path ))

  unittest.TextTestRunner().run(suite)

